/*****************************************************************************
 * Effectors.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Effectors
 *
 * requires:
 *
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors: cyco
 *
 * $Id: Effectors.cc,v 1.11 2004/05/13 15:54:02 harlekin Exp $
 *
 *****************************************************************************/

#include <signal.h>

#include "Effectors.hh"
#include "Configuration.hh"
#include "Player.hh"
#include "debug.hh"
#include "WorldModel.hh"

static Effectors *eff = NULL;
static Logger    **logger = NULL;


Effectors::Effectors(Communicator *c, ServerCommandHistory &sch, 
                     ServerCommandCounters &scc, Logger **logg)
  :comm(c), sentCommandsWM(&sch), queuedCommands(&scc){
  pthread_mutex_init(&mutex_SendAllCommands, NULL);
  
  // Unused parameters:
  logg = logg;
  logger = logger;

  // init SIGALRM Handler
  struct sigaction sigact;
  // BSD-like signal handling: primitives (recvfrom) should not be unblocked
  sigact.sa_flags = SA_RESTART;
  sigact.sa_handler = &sigalrmHandler;
  sigaction( SIGALRM, &sigact, NULL );

  eff = this;
  //logger = logg;
}

void Effectors::sigalrmHandler(int){
  DEBUG1_SYNCHRO("[SIGARLM] \tsigalrmHandler(): called" );

  //send all queued commands
  eff->sendAllCommands(true);
  //(*logger)->writeWorldModel();
}


void Effectors::sendCommandDirect(const char *cmd){
  DEBUG3("sending direct: \"", cmd, "\"");

  //UNQUEUED DIRECT WRITE TO SOCKET
  comm->sendMessage(cmd);
}



void Effectors::sendAllCommands(bool byHandler){
  static char str[64];

  pthread_mutex_lock(&mutex_SendAllCommands);
  //avoid multiple sending of the same commands
  
  // Already sent all commands
  if( !newCycle ){
    // this happens when the handler already fired and
    // now the see-behaviour wants to send its commands
    // we need to recover the backup so we have the
    // correct values for next cycle estimation
    /*
    if ( !byHandler && !Configuration::synchMode ){
      commands      = commands_BackUp;
      *sentCommands = *sentCommands_BackUp;
      DEBUG1_HARLEKIN("Effectors: Restored old commands");
    }
    */
    DEBUG1_SYNCHRO("Effectors: already sent message; dont sent");
    if (!byHandler){
      DEBUG1_HARLEKIN("Effectors: Would have been clash!");
    }
    pthread_mutex_unlock(&mutex_SendAllCommands);
    return;
  }

  // the see-behaviour is currently running and we want to 
  // send the old commands from estimated wm-behaviour
  if (Configuration::synchMode){
    commands = &commands_Full;
    *sentCommandsWM = sentCommands_Full;
  }else{
    if (byHandler){
      commands = &commands_Estimate;
      *sentCommandsWM = sentCommands_Estimate;
    }else{
      commands = &commands_See;
      *sentCommandsWM = sentCommands_See;
    }
  }
  /*
    if (byHandler && haveBackUp){
    if ( !Configuration::synchMode ){
    commands     = commands_BackUp;
    *sentCommands = *sentCommands_BackUp;
    }
  */
  
  if(commands->isMainSet()){
    comm->sendMessage(commands->getMain());
    incrementCounters(commands->getMainType());
  }else{
    comm->sendMessage("(turn 0)");
    incrementCounters(CMD_TURN);
  }
  while( commands->getNextOther(str) ){
    comm->sendMessage(str);
    incrementCounters(commands->getOthersType());
  }  
  newCycle = false;
  pthread_mutex_unlock(&mutex_SendAllCommands);
}


void Effectors::incrementCounters(Command command){
  switch(command){
  case CMD_MOVE:         { queuedCommands->moves++;       break; }
  case CMD_KICK:         { queuedCommands->kicks++;       break; }
  case CMD_DASH:         { queuedCommands->dashes++;      break; }
  case CMD_TURN:         { queuedCommands->turns++;       break; }
  case CMD_GOALIE_CATCH: { queuedCommands->catches++;     break; }
  case CMD_TURN_NECK:    { queuedCommands->turnNecks++;   break; }
  case CMD_CHANGE_VIEW:  { queuedCommands->changeViews++; break; }
  case CMD_SAY:          { queuedCommands->says++;        break; }
  case CMD_TACKLE:       { queuedCommands->tackles++;     break; }
  case CMD_POINT_TO:     { queuedCommands->attentionTos++;break; }
  case CMD_ATTENTION_TO: { queuedCommands->pointTos++;    break; }
  case CMD_SCORE:        { /* not counted by the server */       }
  case CMD_NONE:         { /* do nothing */                      }
  }//switch
}

void Effectors::init(const char *teamname, bool goalie){
  if(goalie) sprintf(command, "(init %s %s)", teamname, "(version 8)(goalie)");
  else       sprintf(command, "(init %s %s)", teamname, "(version 8)");

  //UNQUEUED DIRECT WRITE TO SOCKET
  comm->sendMessage(command);
}

void Effectors::queueTackle(float power){
  if (!newCycle) return;

  sprintf(command, "(tackle %f)", power);

  commands->setMain(command, CMD_TACKLE);
  sentCommands->setFst(CMD_TACKLE);
  sentCommands->setFstParam1(power);

  sentCommandsWM->setFst(CMD_TACKLE);
  sentCommandsWM->setFstParam1(power);
}

void Effectors::queuePointTo(float distance, float direction, bool turnOff){
  if (!newCycle) return;

  if(turnOff)
    sprintf(command, "(pointto off)");
  else
    sprintf(command, "(pointto %f %f)", distance, direction);

  commands->addOther(command, CMD_POINT_TO);
  sentCommands->setSnd(CMD_POINT_TO);
  sentCommands->setSndParam1(distance);
  sentCommands->setSndParam2(direction);
  sentCommandsWM->setSnd(CMD_POINT_TO);
  sentCommandsWM->setSndParam1(distance);
  sentCommandsWM->setSndParam2(direction);
}

void Effectors::queueAttentionTo(bool ourTeam, int unum, bool turnOff){
  if (!newCycle) return;

  if(turnOff)
    sprintf(command, "(attentionto off)");
  else
    if(ourTeam)
      sprintf(command, "(attentionto our %d)", unum);
    else
      sprintf(command, "(attentionto other %d)", unum);
  
  commands->addOther(command, CMD_ATTENTION_TO);
  sentCommands->setSnd(CMD_ATTENTION_TO);
  sentCommands->setSndParam1(ourTeam);
  sentCommands->setSndParam2(unum);
  sentCommandsWM->setSnd(CMD_ATTENTION_TO);
  sentCommandsWM->setSndParam1(ourTeam);
  sentCommandsWM->setSndParam2(unum);
}


void Effectors::queueKick(float power, float direction){
  if (!newCycle) return;

  if (power > 100.0f || power < 0.0f) {
    DEBUG2("Effectors::queueKick(...): tried to kick too hard: ", power);
    power = max(min(100.0f, power), 0.0f);
  } 

  sprintf(command, "(kick %f %f)", power, direction);

  commands->setMain(command, CMD_KICK);
  sentCommands->setFst(CMD_KICK);
  sentCommands->setFstParam1(power);
  sentCommands->setFstParam2(direction);
  sentCommandsWM->setFst(CMD_KICK);
  sentCommandsWM->setFstParam1(power);
  sentCommandsWM->setFstParam2(direction);
}

void  Effectors::queueDash(float power){
  if (!newCycle) return;

  if (fabs(power) > 100.0f) {
    DEBUG2("Effectors::queueDash(...): tried to dash too hard: ", power);
    power = max(min(100.0f, power), -100.0f);
  } 

  sprintf(command, "(dash %f)", power);

  commands->setMain(command, CMD_DASH);
  sentCommands->setFst(CMD_DASH);
  sentCommands->setFstParam1(power);
  sentCommandsWM->setFst(CMD_DASH);
  sentCommandsWM->setFstParam1(power);
}

void Effectors::queueTurn(float angle, float myspeed){
  if (!newCycle) return;

  // turn so that the actual turn is the desired turn
  float corrAngle = angle;
  float inertiaMoment = Configuration::getInertiaMoment(myself->getType());
  float maxAngle = 180.0 / ( 1.0 + inertiaMoment * myspeed );
  if (angle < 0) angle = max(angle, -maxAngle);
  else           angle = min(angle,  maxAngle);

  if ( myspeed != 0 ){
    corrAngle = angle * (1.0 + inertiaMoment * myspeed );

    //if the angle cant be turned, turn maximal and correct 'angle' in order
    //to have the right value in 'sentCommands'
    if (corrAngle < Configuration::minmoment){
      corrAngle = Configuration::minmoment;
      angle = corrAngle / (1.0 + inertiaMoment * myspeed );
    }
    if (corrAngle > Configuration::maxmoment){
      corrAngle = Configuration::maxmoment;
      angle = corrAngle / (1.0 + inertiaMoment * myspeed );
    }
  }

  //correct the angle if too big
  if ( corrAngle > Configuration::maxmoment ) corrAngle = Configuration::maxmoment;
  if ( corrAngle < Configuration::minmoment ) corrAngle = Configuration::minmoment;

  //dont turn if the turn angle not significant
  if (corrAngle > POSITIVE_TO_SMALL_ANGLE
   || corrAngle < NEGATIVE_TO_SMALL_ANGLE){
    sprintf(command, "(turn %f)", corrAngle);
    commands->setMain(command, CMD_TURN);
    sentCommands->setFst(CMD_TURN);
    sentCommands->setFstParam1(angle);  //angle is the effect of turn(corrAngle)
                                        //at myspeed
    sentCommandsWM->setFst(CMD_TURN);
    sentCommandsWM->setFstParam1(angle);  //angle is the effect of turn(corrAngle)  
  }else{
    DEBUG2("Effectors::turn(): tried to turn an angle which is too small: ", corrAngle);
  }
}

void Effectors::queueTurnNeck(float angle){
  if (!newCycle) return;

  sprintf(command, "(turn_neck %f)", angle);

  commands->addOther(command, CMD_TURN_NECK);
  sentCommands->setSnd(CMD_TURN_NECK);
  sentCommands->setSndParam1(angle);
  sentCommandsWM->setSnd(CMD_TURN_NECK);
  sentCommandsWM->setSndParam1(angle);
}

void Effectors::queueGoalieCatch(float direction){
  if (!newCycle) return;

  sprintf(command, "(catch %f)", direction);

  commands->setMain(command, CMD_GOALIE_CATCH);
  sentCommands->setFst(CMD_GOALIE_CATCH);
  sentCommands->setFstParam1(direction);
  sentCommandsWM->setFst(CMD_GOALIE_CATCH);
  sentCommandsWM->setFstParam1(direction);
}

void Effectors::queueSay(const char *message){
  if (!newCycle) return;

  sprintf(command, "(say \"%s\")", message);

  commands->addOther(command, CMD_SAY);
  sentCommands->setTrd(CMD_SAY);
  sentCommands->setTrdParam1(message);
  sentCommandsWM->setTrd(CMD_SAY);
  sentCommandsWM->setTrdParam1(message);
}

void Effectors::queueMove(float x, float y){
  if (!newCycle) return;

  sprintf(command, "(move %f %f)", x, y);

  commands->setMain(command, CMD_MOVE);
  sentCommands->setFst(CMD_MOVE);
  sentCommands->setFstParam1(x);
  sentCommands->setFstParam2(y);
  sentCommandsWM->setFst(CMD_MOVE);
  sentCommandsWM->setFstParam1(x);
  sentCommandsWM->setFstParam2(y);
}

void Effectors::queueChangeView(const ViewWidth vw, const ViewQuality vq){
  if (!newCycle) return;

  char width[16], quality[16];
  if(vw == VW_NARROW) strcpy(width, "narrow");
  if(vw == VW_NORMAL) strcpy(width, "normal");
  if(vw == VW_WIDE)   strcpy(width, "wide");
  (vq == VQ_HIGH) ?   strcpy(quality, "high") : strcpy(quality, "low");

  sprintf(command, "(change_view %s %s)", width, quality);

  commands->addOther(command, CMD_CHANGE_VIEW);
  sentCommands->setSnd(CMD_CHANGE_VIEW);
  sentCommands->setSndParam1(vw);
  sentCommands->setSndParam2(vq);
  sentCommandsWM->setSnd(CMD_CHANGE_VIEW);
  sentCommandsWM->setSndParam1(vw);
  sentCommandsWM->setSndParam2(vq);
}

void Effectors::sendChangeView(const ViewWidth vw, const ViewQuality vq){
  char width[16], quality[16];
  if(vw == VW_NARROW) strcpy(width, "narrow");
  if(vw == VW_NORMAL) strcpy(width, "normal");
  if(vw == VW_WIDE)   strcpy(width, "wide");
  (vq == VQ_HIGH) ?   strcpy(quality, "high") : strcpy(quality, "low");

  sprintf(command, "(change_view %s %s)", width, quality);

  sendCommandDirect(command);
  sentCommands->setSnd(CMD_CHANGE_VIEW);
  sentCommands->setSndParam1(vw);
  sentCommands->setSndParam2(vq);
}

void Effectors::queueScore(){
  if (!newCycle) return;

  strcpy(command, "(score)");

  commands->addOther(command, CMD_SCORE);
  sentCommands->setSnd(CMD_SCORE);
  sentCommandsWM->setSnd(CMD_SCORE);
}

