/*****************************************************************************
 * Effectors.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Effectors
 *
 * requires:
 *
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors: cyco
 *
 * $Id: Effectors.hh,v 1.9 2004/05/07 12:04:41 harlekin Exp $
 *
 *****************************************************************************/

#ifndef EFFECTORS_HH
#define EFFECTORS_HH

#include <stdio.h>
#include <pthread.h>

#include "Myself.hh"
#include "Communicator.hh"
#include "ServerCommands.hh"
#include "Logger.hh"
/**
 * The (only) interface of the agent to the environment, which can be used
 * to influence the environment. There are a fixed number of commands, which
 * can be sent to the server. It is possible to send many of them in one cycle
 * but there are commands, that exclude one other (namely: kick, turn, dash,
 * move, tackle and catch). This commands are often called primary commands. Effectors
 * contains a queue, which can store one primary and many non-primary commands.
 * When the signal SIGALRM from the system timer arrives, the commands that
 * are currently in the queue are sent to the server. The Worldmodel is
 * informed about every command that was sent in the last cycle and about the
 * total number of commands in order to observe the reliability of the
 * communinication and in order to be able to simulate the effects of the
 * commands.
 *
 * There are two different queues. One which is filled by the behaviour that
 * is based on the estimated state and one that is filled by the behaviour 
 * that is called after an new see message arrives. This is necessary to prevent 
 * those commands from overwriting each other. Even if the see-behaviour
 * queued some commands the agent sometimes wants to send his commands before
 * that behaviour is done. So the commands from the estimated queue are sent.
 * Those commands don't rely on an up to date worldmodel but at least are
 * complete. If the see-behaviour is done its queued commands will be send.
 */
class Effectors {
private:
  //constants
  static const float MAX_TURN_ANGLE          = 180;
  static const float MIN_TURN_ANGLE          = -180;
  static const float NEGATIVE_TO_SMALL_ANGLE = -.1;
  static const float POSITIVE_TO_SMALL_ANGLE = +.1;

  /**
   * Schtzt die SendAllCommands-Methode vor Nebenlufigkeit.
   * Nebenlufigkeitsprobleme knnen auftreten, wenn der Sigalarm-Handler 
   * in die Abarbeitung der Methode (ausgelst durch Agent->run()) einspringt.
   */
  pthread_mutex_t mutex_SendAllCommands;

  /**
   * Is set to "false" after sending of the commands to the server and indicates
   * a succeded "acting". At the beginning of a new cycle is set back to true (by
   * Agent). Is not protected by everything because the practice showed no need
   * on it.
   */
  volatile bool          newCycle;
  bool                   haveBackUp;

  EffectorState          state;

  /** the "queue" to store the commands into */
  ServerCommands         *commands;
  ServerCommands         commands_Estimate;
  ServerCommands         commands_See;
  ServerCommands         commands_Full;
  char                   command[SAY_MSG_MAX_SIZE];
  Communicator          *comm;
  /** stores the kind and params of the command sent in the last cycle. resides in worldmodel */
  ServerCommandHistory  *sentCommandsWM;
  ServerCommandHistory  *sentCommands;
  /* This backup is used if stored commands cant be sent because some commands
     were already sent at that cycle */
  ServerCommandHistory  sentCommands_Estimate;
  ServerCommandHistory  sentCommands_See;
  ServerCommandHistory  sentCommands_Full;
  /** stores the total number of commands sent for each command type. resides in worldmodel */
  ServerCommandCounters *queuedCommands;

  /**
   * increments the counter for a certain command
   * @param c the counter for this command is to increase
   */
  void incrementCounters(Command c);

  /**
   * This function is a signal handler and is called from the OS when a SIGALRM
   * signal arrives. The time this signal comes is set by
   * SenseThread::setTimeSignal. If the signal arrives, all commands currently
   * stored in the command queue are sent to the server using the Communicator.
   * @param i is ignored
   */
  static void sigalrmHandler(int i);
public:
  Effectors(Communicator *, ServerCommandHistory &, ServerCommandCounters&,
	    Logger**);


  Myself*                 myself; // Wenn das mal kein fieser Hack ist...


  void setState(EffectorState es){
    state = es;
    switch(state){
    case ES_See:
      commands_See.deleteAll();
      sentCommands_See.clear();
      commands = &commands_See;
      sentCommands = &sentCommands_See;
      //sentCommandsWM = &sentCommands_See;
      break;
    case ES_Estimate:
      commands_See.deleteAll();
      commands_Estimate.deleteAll();
      sentCommands_Estimate.clear();
      commands = &commands_Estimate;
      sentCommands = &sentCommands_Estimate;
      //*sentCommandsWM = sentCommands_Estimate;
      break;
    default:
      commands_Full.deleteAll();
      sentCommands_Full.clear();
      commands = &commands_Full;
      sentCommands = &sentCommands_Full;
      //*sentCommandsWM = sentCommands_Full;
      break;
    }
  };

  /**
   * \name The effectors of the agent
   *
   * the following methodes assebles the strings that correspond to a certain
   * command with certain parameters. These strings are then stored in the command
   * queue. In the case of a primary command the previous primary command will be
   * overwritten (if there was one)
   */
  //@{
  void init(const char *teamname, bool goalie = false);
  void queueMove       (float x, float y);
  void queueKick       (float power, float direction);
  void queueDash       (float power);
  void queueGoalieCatch(float direction);
  void queueTurnNeck   (float angle);
  void queueChangeView (const ViewWidth width, const ViewQuality quality);
  void sendChangeView(const ViewWidth vw, const ViewQuality vq);


  void queueScore      ();
  void queueSay        (const char *message);

  /**
     Queues a tackle message
     @param power the power with which to tackle
  */
  void queueTackle(float power);

  /**
     Queues an attentionTo Message
     @param ourTeam the Agents Team, or the other
     @param unum the uniform number of the Player to pay attention to
     @param turnOff standard value is off. If one wants toi send an attention to off, send true. The other values will be ignored then
  */
  void queueAttentionTo(bool ourTeam, int unum, bool turnOff = false);

  /**
     Queues a Pointto Message. This messages lasts at least 5 cycles (point_to_ban). It is turned off, either if another pointTo
     command is set, or (pointTo off) is set or 20 Cycles pass (point_to_duration).
     @param distance the distance where to point
     @param the angle in which to point
     @param turnOff standard value is off. If one wants toi send an attention to off, send true. The other values will be ignored then
  */
  void queuePointTo(float distance, float direction, bool turnOff = false);

  /**
   * In contrast to the other queue... methods this one performs some
   * calculations on angle, before it is queued. The aspect of inertia in
   * the turn model of the soccer server is considered here. Also it prevents
   * queueing of illegal or very small (not significant) turns.
   * @param angle is desired to be turned. Turn with such an angle x is queued,
   * so that the effect of executing "turn x" by the speed "myspeed" is turning
   * of angle "angle"
   * @param myspeed The speed that is considered by the calculations ot the
   * effects of inertia.
   */
  void queueTurn       (float angle, float myspeed = 0);
  //@}


  /**
   * deletes all commands currently stored in the command queue.
   */
  void deleteAllCommands() { 
    //commands.deleteAll(); 
  }

  void createBackUp(){
    // create BackUp
    /*
    if ( !Configuration::synchMode ){
      commands_BackUp      = commands;
      *sentCommands_BackUp = *sentCommands;
      haveBackUp = true;
    }
    */
  }

  /**
   * sends all queued commands to the soccerserver if it wasn't executed in the
   * current cycle already. Sets the flag "newCycle" to "false" after sending of
   * all messages.
   */
  void sendAllCommands(bool byHandler = false);

  /**
   * sends the command to the server direct, does not queue it
   * @param cmd the command to send
   */
  void sendCommandDirect(const char *cmd);

  void setNewCycle()       {     
    newCycle = true;  
    haveBackUp = false;
  }
};

// $Source: /home/robocup/virtualwerder/CVSRepository/VirtualWerder2004/Effectors.hh,v $
#endif //EFFECTORS_HH
