/*****************************************************************************
 * SenseObjects.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Sensors
 *
 * requires:
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: cyco
 * Group:
 * Co-authors: wenns
 *
 * $Id: SenseObjects.hh,v 1.6 2004/05/13 16:28:24 aw Exp $
 *****************************************************************************/

#ifndef SENSEOBJECTS_HH
#define SENSEOBJECTS_HH

#include <vector>
#include "debug.hh"
using namespace std;

#include "Geometry.hh"

#define TEAMNAME_LENGTH 32


namespace Sense {

  class StatObject {
    public:
      StatObject ();
      virtual ~StatObject();

      Vector2D getRelPos() const { return relPos; }
      const Vector2D *getRelPosPt() const { return &relPos; }
      void setRelPos ( const Vector2D &v) { relPos = v; }

    private:
      Vector2D relPos;		// objects relative position to player
  };


  class Line : public StatObject {
    public:
      Line ();
      virtual ~Line();

      LineId getId () const              { return id; }
      void   setId ( const LineId lid )  { id = lid;  }

      //virtual ostream& print ( ostream& o) const;

    private:
      LineId id;
  };

  //=======================================================================
  //	MobObject
  //=======================================================================

  class MobObject : public StatObject {
    public:
      static Vector2D calcVel( const float dist, const float dir,
                               const float distChg, const float dirChg );

      MobObject ( );
      virtual ~MobObject();

      Vector2D getRelVel() const              { return relVel; }
    const Vector2D *getRelVelPt() const              { return &relVel; }
      void setRelVel( const Vector2D &v ) {
        relVel = v;
      }

      bool isClose  () const { return close; }
      void setClose ( const bool c ) { close = c; }

    private:
      bool close;
      Vector2D relVel;
  };

  //=======================================================================
  //	Flag
  //=======================================================================

  /**
   * since the flag can habe distChg and dirChg from the player own movement,
   * it should be a MobObject
   */
  class Flag : public MobObject {
    public:
      Flag ();
      virtual ~Flag();

      MarkerId  getId () const          { return id; }
      void setId ( const MarkerId fid ) { id = fid; }

      //virtual ostream& print ( ostream& o, Vector2D::mode_t = Vector2D::POLAR, Angle::mode_t = Angle::DEG ) const;
    private:
      MarkerId id;
  };


  class Goal : public Flag {
    public:
      Goal ();
      virtual ~Goal();
  };


  class Ball : public MobObject {
    public:
      Ball();
      virtual ~Ball();

      //virtual ostream& print ( ostream& o, Vector2D::mode_t = Vector2D::POLAR, Angle::mode_t = Angle::DEG ) const;
  };

  //=======================================================================
  //	Player
  //=======================================================================

  class Player : public MobObject {
    public:
      Player();
      ~Player();

      float getBodyDir ()  const { return  bodyDir; }
      float getHeadDir ()  const { return headDir; }
      const char* getTeamname () const { return teamname; }
      int   getUnum ()     const { return unum; }
    // HACK: Somehow the goalie-variable is never set
    //       now we should be able to recognize our own goalie
    //      bool  isGoalie ()    const { return goalie; }
      bool  isGoalie ()    const { return goalie || unum == 1; }
      bool  isTackling()   const { return tackling; }
      float getPointDir()  const { return pointDir; }
      int   getType()      const { return type; }
      float getStamina()   const { return stamina; }
      float getEffort()    const { return effort;} 
      float getRecovery()  const { return recovery; }

      void  setBodyDir  ( const float dir )  { bodyDir = dir; }
      void  setHeadDir  ( const float dir )  { headDir = dir; }
      void  setTeamname ( const char *name ) { strcpy(teamname, name); }
      void  setUnum     ( const int un )     { unum = un; }
      void  setGoalie   ( const bool g )     { goalie = g; }
      void  setTackling ( const bool t )     { tackling = t; }
      void  setPointDir ( const float dir )  { pointDir = dir; }
      void  setType     ( const int t )      { type = t; }
      void  setStaminaInfos( const float s, const float e, const float r)
      { stamina = s; effort = e; recovery = r; }
      //virtual ostream& print ( ostream& o, Vector2D::mode_t = Vector2D::POLAR, Angle::mode_t = Angle::DEG ) const;

    private:
      float bodyDir;
      float headDir;
      char  teamname[TEAMNAME_LENGTH];
      int   unum;
      bool  goalie;
      bool  tackling;
      float pointDir;

      //only sensed through fullstate msg
      int   type;     /**< 0: normal, heteros: 1-6 */
      float stamina;  /**< max 4000 by default */
      float effort;   /**< [0.6 .. 1.0] by default */
      float recovery; /**< [0.5 .. 1.0] by default */
  };


  //=======================================================================
  //	Score
  //=======================================================================

  class Score {
    public:
      Score();
      virtual ~Score ();

      //ostream& print ( ostream& o ) const
      //{
      //  return o << "(score " << time_ << " " << our_score << " " << opp_score << ")";
      //}
      //ostream& print ( ostream& o, const Side& ) const { return print ( o ); }

      void setTime (const int t ) { time = t; }
      void setOurScore (const int ourSc) { ourScore = ourSc; }
      void setOppScore (const int oppSc) { oppScore = oppSc; }

      int getTime () const { return time; }
      int getOurScore () const { return ourScore; }
      int getOppScore () const { return oppScore; }

    private:
      int time;
      int ourScore;
      int oppScore;
  };



  //=======================================================================
  //	SenseBody
  //=======================================================================

  struct TackleInfo{
    public:
      TackleInfo();
      int getExpires () const {return expires; }
      int getCount ()   const {return count;   }

      void setExpires (const int e) { expires = e; }
      void setCount (const int c) { count = c; }
    private:
      int expires; /**< amount of cycles until the player can move
                        after a tackle, 0 indicates the player isnt tackling */
      int count;   /**< number of tackles performed by player */
  };

  struct FocusInfo{
    public:
      FocusInfo();
      bool isTargetSet () const        { return targetSet; }
      FieldSide getTargetSide () const { return targetSide; }
      int getTargetUnum () const       { return targetUnum; }
      int getCount () const            { return count; }

      void setTargetSet (const bool s) { targetSet = s; }
      void setTargetSide (const FieldSide ts) { targetSide = ts; }
      void setTargetUnum (const int unum) { targetUnum = unum; }
      void setCount (const int c) { count = c; }
    private:
      bool      targetSet;  /**< true, if the focus is currently set */
      FieldSide targetSide; /**< team affiliation of the focused player */
      int       targetUnum; /**< unum of the focused player */
      int       count;      /**< number of attentionto actions performed
                                  by player */
  };

  struct ArmInfo{
    public:
      ArmInfo();
      void setMovable (const int m) { movable = m; }
      void setExpires (const int e) { expires = e; }
      void setTarget (const float dist, const float dir)
      { target = Vector2D::getVector2DFromPolar( dist, dir ); }
      void setCount (int c) { count = c; }

      int getMovable() const { return movable; }
      int getExpires() const { return expires; }
      Vector2D getTarget() const { return target; }
      int getCount() const { return count; }
    private:
      int movable;     /**< number of cycles, till the arm is movable
                            0 means the arm is movable now */
      int expires;     /**< numver of cycles till the arm stops
                            pointing 0 means the arm isnt pointing */
      Vector2D target; /**< the absolute field position the arm
                            is pointing */
      int count;       /**< number of pointto actions performed by
                            player */
  };


  class SenseBody {
    public:

      SenseBody();
      virtual ~SenseBody();

      /** \name Accessors */
      //@{
      int getTime ()                const {return time;}
      ViewQuality getViewQuality () const {return viewQuality;}
      ViewWidth getViewWidth ()     const {return viewWidth;}
      float getStamina ()           const {return stamina;}
      float getEffort ()            const {return effort;}
      float getSpeedAmount ()       const {return speedAmount;}
      float getSpeedDirection ()    const {return speedDirection;}
      float getHeadAngle ()         const {return headAngle;}
      int getKickCount ()           const {return kickCount;}
      int getDashCount ()           const {return dashCount;}
      int getTurnCount ()           const {return turnCount;}
      int getSayCount ()            const {return sayCount;}
      int getTurnNeckCount ()       const {return turnNeckCount;}
      int getCatchCount ()          const {return catchCount;}
      int getMoveCount ()           const {return moveCount;}
      int getChangeViewCount ()     const {return changeViewCount;}

      void setAll(const int newTime, const ViewQuality newViewQuality,
                  const ViewWidth newViewWidth, const float newStamina,
                  const float newEffort, const float newSpeedAmount,
                  const float newSpeedDirection, const float newHeadAngle,
                  const int newKickCount, const int newDashCount,
                  const int newTurnCount, const int newSayCount,
                  const int newTurnNeckCount, int newCatchCount,
                  const int newMoveCount, const int newChangeViewCount)
      {
        time            = newTime;
        viewQuality     = newViewQuality;
        viewWidth       = newViewWidth;
        stamina         = newStamina;
        effort          = newEffort;
        speedAmount     = newSpeedAmount;
        speedDirection  = newSpeedDirection;
        headAngle       = newHeadAngle;
        kickCount       = newKickCount;
        dashCount       = newDashCount;
        turnCount       = newTurnCount;
        sayCount        = newSayCount;
        turnNeckCount   = newTurnNeckCount;
        catchCount      = newCatchCount;
        moveCount       = newMoveCount;
        changeViewCount = newChangeViewCount;
      }

      /** output on a string stream */
      //friend ostream& operator<< (ostream &os, const SenseBody &bi){
      //  os << "time: " << bi.time << " viewQuality: " << bi.viewQuality
    //    << " viewWidth: " << bi.viewWidth << " stamina: " << bi.stamina
    //    << " effort: " << bi.effort << " speedAmount: " << bi.speedAmount
    //    << " speedDirection: " << bi.speedDirection
    //    << " headAngle: " << bi.headAngle
    //    << " kickCount: " << bi.kickCount << " dashCount: " << bi.dashCount
    //    << " turnCount: " << bi.turnCount << " sayCount: " << bi.sayCount
    //    << " turnNeckCount: " << bi.turnNeckCount
    //    << " catchCount: " << bi.catchCount
    //    << " moveCount: " << bi.moveCount
    //    << " changeViewCount: " << bi.changeViewCount
    //    << endl;
    //}

      ArmInfo armInfo;
      FocusInfo focusInfo;
      TackleInfo tackleInfo;
  private:
    /** The soccer server cycle the information comes from */
    int time;
    ViewQuality viewQuality;   /**< HIGH or LOW*/
    ViewWidth   viewWidth;     /**< WIDE, NORMAL or NARROW*/
    float stamina;             /**< max 4000 by default */
    float effort;              /**< [0.6 .. 1] by default */
    float speedAmount;         /**< */
    float speedDirection;      /**< */
    float headAngle;           /**< relative to the body */

    /**
     * \name Command counter
     *
     * The counter serve primarily as a debugging goody. The server increments
     * a counter, if an according command was received, and sends all the
     * counter values to the client with the next sense body message.
     */
    //@{
    int kickCount;
    int dashCount;
    int turnCount;
    int sayCount;
    int turnNeckCount;
    int catchCount;
    int moveCount;
    int changeViewCount;
    //@}
};


class Fullstate {
  public:
    Fullstate();
    virtual ~Fullstate();

    /** \name Accessors */
    //@{
    int getTime ()                const {return time;}
    ViewQuality getViewQuality () const {return viewQuality;}
    ViewWidth getViewWidth ()     const {return viewWidth;}
    Playmode getPlaymode()        const {return playmode;}
    int getOurScore()             const {return ourScore;}
    int getOppScore()             const {return oppScore;}

    int getKickCount ()           const {return kickCount;}
    int getDashCount ()           const {return dashCount;}
    int getTurnCount ()           const {return turnCount;}
    int getSayCount ()            const {return sayCount;}
    int getTurnNeckCount ()       const {return turnNeckCount;}
    int getCatchCount ()          const {return catchCount;}
    int getMoveCount ()           const {return moveCount;}
    int getChangeViewCount ()     const {return changeViewCount;}


    void setTime( const int t ) { time = t; }
    void setScores( const int our, const int opp )
    { ourScore = our; oppScore = opp; }
    void setViewMode( const ViewQuality vq, const ViewWidth vw )
    { viewQuality = vq; viewWidth = vw; }
    void setPlayMode( const Playmode p ) { playmode = p; };

    void setCounters( const int kc, const int dc, const int tc, const int sc,
                      const int tnc, int cc, const int mc, const int cvc)
    {
      kickCount       = kc;
      dashCount       = dc;
      turnCount       = tc;
      sayCount        = sc;
      turnNeckCount   = tnc;
      catchCount      = cc;
      moveCount       = mc;
      changeViewCount = cvc;
    }

    /** output on a string stream */
    //friend ostream& operator<< (ostream &os, const Fullstate &fs){
    //  os << "time: " << bi.time << " viewQuality: " << bi.viewQuality
    //    << " viewWidth: " << bi.viewWidth << " stamina: " << bi.stamina
    //    << " effort: " << bi.effort << " speedAmount: " << bi.speedAmount
    //    << " speedDirection: " << bi.speedDirection
    //    << " headAngle: " << bi.headAngle
    //    << " kickCount: " << bi.kickCount << " dashCount: " << bi.dashCount
    //    << " turnCount: " << bi.turnCount << " sayCount: " << bi.sayCount
    //    << " turnNeckCount: " << bi.turnNeckCount
    //    << " catchCount: " << bi.catchCount
    //    << " moveCount: " << bi.moveCount
    //    << " changeViewCount: " << bi.changeViewCount
    //    << endl;
    //}

    ArmInfo armInfo;
    //FocusInfo focusInfo;      //whats about this?
    //TackleInfo tackleInfo;    //... and this?
  private:
    int time;                  /**< The soccer server cycle the information
                                    comes from */
    ViewQuality viewQuality;   /**< HIGH or LOW*/
    ViewWidth   viewWidth;     /**< WIDE, NORMAL or NARROW*/
    Playmode playmode;         /**< current playmode */
    int ourScore;              /**< goals scored by our team */
    int oppScore;              /**< goals scored by the opponent team */

    /**
     * \name Command counter
     *
     * The counter serve primarily as a debugging goody. The server increments
     * a counter, if an according command was received, and sends all the
     * counter values to the client with the next sense body message.
     */
    //@{
    int kickCount;
    int dashCount;
    int turnCount;
    int sayCount;
    int turnNeckCount;
    int catchCount;
    int moveCount;
    int changeViewCount;
    //@}
};



//=======================================================================
//	ServerParam
//=======================================================================
class ServerParam{
public:
  ServerParam ( );
  virtual ~ServerParam ();

  void setAll(
              const int catch_ban_cycle,
              const int clang_advice_win,
              const int clang_define_win,
              const int clang_del_win,
              const int clang_info_win,
              const int clang_mess_delay,
              const int clang_mess_per_cycle,
              const int clang_meta_win,
              const int clang_rule_win,
              const int clang_win_size,
              const int coach_port,
              const int drop_ball_time,
              const int freeform_send_period,
              const int freeform_wait_period,
              const int game_log_compression,
              const int game_log_version,
              const int goalie_max_moves,
              const int half_time,
              const int hear_decay,
              const int hear_inc,
              const int hear_max,
              const int max_goal_kicks,
              const int olcoach_port,
              const int point_to_ban,
              const int point_to_duration,
              const int port,
              const int recv_step,
              const int say_coach_cnt_max,
              const int say_coach_msg_size,
              const int say_msg_size,
              const int send_step,
              const int send_vi_step,
              const int sense_body_step,
              const int simulator_step,
              const int slow_down_factor,
              const int start_goal_l,
              const int start_goal_r,
              const int synch_micro_sleep,
              const int synch_offset,
              const int tackle_cycles,
              const int text_log_compression,
              const char *game_log_dir,
              const char *game_log_fixed_name,
              const char *landmark_file,
              const char *log_date_format,
              const char *text_log_dir,
              const char *text_log_fixed_name,
              const bool coach,
              const bool coach_w_referee,
              const bool old_coach_hear,
              const bool wind_none,
              const float wind_random,
              const bool back_passes,
              const bool forbid_kick_off_offside,
              const bool free_kick_faults,
              const bool fullstate_l,
              const bool fullstate_r,
              const bool game_log_dated,
              const bool game_log_fixed,
              const bool game_logging,
              const bool log_times,
              const bool profile,
              const bool proper_goal_kicks,
              const bool record_messages,
              const bool send_comms,
              const bool synch_mode,
              const bool team_actuator_noise,
              const bool text_log_dated,
              const bool text_log_fixed,
              const bool text_logging,
              const bool use_offside,
              const bool verbose,
              const float audio_cut_dist,
              const float ball_accel_max,
              const float ball_decay,
              const float ball_rand,
              const float ball_size,
              const float ball_speed_max,
              const float ball_weight,
              const float catch_probability,
              const float catchable_area_l,
              const float catchable_area_w,
              const float ckick_margin,
              const float control_radius,
              const float dash_power_rate,
              const float effort_dec,
              const float effort_dec_thr,
              const float effort_inc,
              const float effort_inc_thr,
              const float effort_init,
              const float effort_min,
              const float goal_width,
              const float inertia_moment,
              const float kick_power_rate,
              const float kick_rand,
              const float kick_rand_factor_l,
              const float kick_rand_factor_r,
              const float kickable_margin,
              const float maxmoment,
              const float maxneckang,
              const float maxneckmoment,
              const float maxpower,
              const float minmoment,
              const float minneckang,
              const float minneckmoment,
              const float minpower,
              const float offside_active_area_size,
              const float offside_kick_margin,
              const float player_accel_max,
              const float player_decay,
              const float player_rand,
              const float player_size,
              const float player_speed_max,
              const float player_weight,
              const float prand_factor_l,
              const float prand_factor_r,
              const float quantize_step,
              const float quantize_step_l,
              const float recover_dec,
              const float recover_dec_thr,
              const float recover_min,
              const float slowness_on_top_for_left_team,
              const float slowness_on_top_for_right_team,
              const float stamina_inc_max,
              const float stamina_max,
              const float stopped_ball_vel,
              const float tackle_back_dist,
              const float tackle_dist,
              const float tackle_exponent,
              const float tackle_power_rate,
              const float tackle_width,
              const float visible_angle,
              const float visible_distance,
              const float wind_ang,
              const float wind_dir,
              const float wind_force,
              const float wind_rand
             );

  //friend ostream& operator<< (ostream& o, const ServerParam& sp)
  //{
  //  return sp.print ( o );
  //}

  int getCatchBanCycle () const { return catchBanCycle; }
  void setCatchBanCycle (const int val) { catchBanCycle = val; }
  int getClangAdviceWin () const { return clangAdviceWin; }  
  void setCLangAdviceWin (const int val) { clangAdviceWin = val; }
  int getClangDefineWin () const { return clangDefineWin; }
  void setCLangDefineWin (const int val) { clangDefineWin = val; }
  int getClangDelWin () const { return clangDelWin; }
  void setCLangDelWin (const int val) { clangDelWin = val; }
  int getClangInfoWin () const { return clangInfoWin; }
  void setCLangInfoWin (const int val) { clangInfoWin = val; }
  int getClangMessDelay () const { return clangMessDelay; }
  void setCLangMessDelay (const int val) { clangMessDelay = val; }
  int getClangMessPerCycle () const { return clangMessPerCycle; }
  void setCLangMessPerCycle (const int val) { clangMessPerCycle = val; }
  int getClangMetaWin () const { return clangMetaWin; }
  void setCLangMetaWin (const int val) { clangMetaWin = val; }
  int getClangRuleWin () const { return clangRuleWin; }
  void setCLangRuleWin (const int val) { clangRuleWin = val; }
  int getClangWinSize () const { return clangWinSize; }
  void setCLangWinSize (const int val) { clangWinSize = val; }
  int getCoachPort () const { return coachPort; }
  void setCoachPort (const int val) { coachPort = val; }
  int getDropBallTime () const { return dropBallTime; }
  void setDropBallTime (const int val) { dropBallTime = val; }
  int getFreeformSendPeriod () const { return freeformSendPeriod; }
  void setFreeformSendPeriod (const int val) { freeformSendPeriod = val; }
  int getFreeformWaitPeriod () const { return freeformWaitPeriod; }
  void setFreeformWaitPeriod (const int val) { freeformWaitPeriod = val; }
  int getGameLogCompression () const { return gameLogCompression; }
  void setGameLogCompression (const int val) { gameLogCompression = val; }

  void setGameLogVersion (const int val) { gameLogVersion = val; }
  void setGoalieMaxMoves (const int val) { goalieMaxMoves = val; }
  void setHalfTime (const int val) { halfTime = val; }
  void setHearDecay (const int val) { hearDecay = val; }
  void setHearInc (const int val) { hearInc = val; }
  void setHearMax (const int val) { hearMax = val; }
  void setMaxGoalKicks (const int val) { maxGoalKicks = val; }
  void setOLCoachPort (const int val) { olcoachPort = val; }
  void setPointToBan (const int val) { pointToBan = val; }
  void setPointToDuration (const int val) { pointToDuration = val; }
  void setPort (const int val) { this->port = val; }
  void setRecvStep (const int val) { recvStep = val; }
  void setSayCoachCntMax (const int val) { sayCoachCntMax = val; }
  void setSayCoachMsgSize (const int val) { sayCoachMsgSize = val; }
  void setSayMsgSize (const int val) { sayMsgSize = val; }
  void setSendStep (const int val) { sendStep = val; }
  void setSendVIStep (const int val) { sendViStep = val; }
  void setSenseBodyStep (const int val) { senseBodyStep = val; }
  void setSimulatorStep (const int val) { simulatorStep = val; }
  void setSlowDownFactor (const int val) { slowDownFactor = val; }
  void setStartGoalL (const int val) { startGoalL = val; }
  void setStartGoalR (const int val) { startGoalR = val; }
  void setSynchMicroSleep (const int val) { synchMicroSleep = val; }
  void setSynchOffset (const int val) { synchOffset = val; }
  void setTackleCycles (const int val) { tackleCycles = val; }
  void setTextLogCompression (const int val) { textLogCompression = val; }
  void setGameLogDir (const char* val) { gameLogDir = strdup(val); }
  void setGameLogFixedName (const char* val) { gameLogFixedName = strdup(val); }
  void setLandmarkFile (const char* val) { landmarkFile = strdup(val); }
  void setLogDateFormat (const char* val) { logDateFormat = strdup(val); }
  void setTextLogDir (const char* val) { textLogDir = strdup(val); }
  void setTextLogFixedName (const char* val) { textLogFixedName = strdup(val); }
  void setCoach (const bool val) { this->coach = val; }
  void setCoachWReferee (const bool val) { coachWReferee = val; }
  void setOldCoachHear (const bool val) { oldCoachHear = val; }
  void setWindNone (const bool val) { windNone = val; }
  void setWindRandom (const float val) { windRandom = val; }
  void setBackPasses (const bool val) { backPasses = val; }
  void setForbidKickOffOffside (const bool val) { forbidKickOffOffside = val; }
  void setFreeKickFaults (const bool val) { freeKickFaults = val; }
  void setFullstateL (const bool val) { fullstateL = val; }
  void setFullstateR (const bool val) { fullstateR = val; }
  void setGameLogDated (const bool val) { gameLogDated = val; }
  void setGameLogFixed (const bool val) { gameLogFixed = val; }
  void setGameLogging (const bool val) { gameLogging = val; }
  void setLogTimes (const bool val) { logTimes = val; }
  void setProfile (const bool val) { this->profile = val; }
  void setProperGoalKicks (const bool val) { properGoalKicks = val; }
  void setRecordMessages (const bool val) { recordMessages = val; }
  void setSendComms (const bool val) { sendComms = val; }
  void setSynchMode (const bool val) { synchMode = val; }
  void setTeamActuatorNoise (const bool val) { teamActuatorNoise = val; }
  void setTextLogDated (const bool val) { textLogDated = val; }
  void setTextLogFixed (const bool val) { textLogFixed = val; }
  void setTextLogging (const bool val) { textLogging = val; }
  void setUseOffside (const bool val) { useOffside = val; }
  void setVerbose (const bool val) { this->verbose = val; }
  void setAudioCutDist (const float val) { audioCutDist = val; }
  void setBallAccelMax (const float val) { ballAccelMax = val; }
  void setBallDecay (const float val) { ballDecay = val; }
  void setBallRand (const float val) { ballRand = val; }
  void setBallSize (const float val) { ballSize = val; }
  void setBallSpeedMax (const float val) { ballSpeedMax = val; }
  void setBallWeight (const float val) { ballWeight = val; }
  void setCatchProbability (const float val) { catchProbability = val; }
  void setCatchableAreaL (const float val) { catchableAreaL = val; }
  void setCatchableAreaW (const float val) { catchableAreaW = val; }
  void setCKickMargin (const float val) { ckickMargin = val; }
  void setControlRadius (const float val) { controlRadius = val; }
  void setDashPowerRate (const float val) { dashPowerRate = val; }
  void setEffortDec (const float val) { effortDec = val; }
  void setEffortDecThr (const float val) { effortDecThr = val; }
  void setEffortInc (const float val) { effortInc = val; }
  void setEffortIncThr (const float val) { effortIncThr = val; }
  void setEffortInit (const float val) { effortInit = val; }
  void setEffortMin (const float val) { effortMin = val; }
  void setGoalWidth (const float val) { goalWidth = val; }
  void setInertiaMoment (const float val) { inertiaMoment = val; }
  void setKickPowerRate (const float val) { kickPowerRate = val; }
  void setKickRand (const float val) { kickRand = val; }
  void setKickRandFactorL (const float val) { kickRandFactorL = val; }
  void setKickRandFactorR (const float val) { kickRandFactorR = val; }
  void setKickableMargin (const float val) { kickableMargin = val; }
  void setMaxMoment (const float val) { this->maxmoment = val; }
  void setMaxNeckAng (const float val) { this->maxneckang = val; }
  void setMaxNeckMoment (const float val) { this->maxneckmoment = val; }
  void setMaxPower (const float val) { this->maxpower = val; }
  void setMinMoment (const float val) { this->minmoment = val; }
  void setMinNeckAng (const float val) { this->minneckang = val; }
  void setMinNeckMoment (const float val) { this->minneckmoment = val; }
  void setMinPower (const float val) { this->minpower = val; }
  void setOffsideActiveAreaSize (const float val) { offsideActiveAreaSize = val; }
  void setOffsideKickMargin (const float val) { offsideKickMargin = val; }
  void setPlayerAccelMax (const float val) { playerAccelMax = val; }
  void setPlayerDecay (const float val) { playerDecay = val; }
  void setPlayerRand (const float val) { playerRand = val; }
  void setPlayerSize (const float val) { playerSize = val; }
  void setPlayerSpeedMax (const float val) { playerSpeedMax = val; }
  void setPlayerWeight (const float val) { playerWeight = val; }
  void setPRandFactorL (const float val) { prandFactorL = val; }
  void setPRandFactorR (const float val) { prandFactorR = val; }
  void setQuantizeStep (const float val) { quantizeStep = val; }
  void setQuantizeStepL (const float val) { quantizeStepL = val; }
  void setRecoverDec (const float val) { recoverDec = val; }
  void setRecoverDecThr (const float val) { recoverDecThr = val; }
  void setRecoverMin (const float val) { recoverMin = val; }
  void setSlownessOnTopForLeftTeam (const float val) { slownessOnTopForLeftTeam = val; }
  void setSlownessOnTopForRightTeam (const float val) { slownessOnTopForRightTeam = val; }
  void setStaminaIncMax (const float val) { staminaIncMax = val; }
  void setStaminaMax (const float val) { staminaMax = val; }
  void setStoppedBallVel (const float val) { stoppedBallVel = val; }
  void setTackleBackDist (const float val) { tackleBackDist = val; }
  void setTackleDist (const float val) { tackleDist = val; }
  void setTackleExponent (const float val) { tackleExponent = val; }
  void setTacklePowerRate (const float val) { tacklePowerRate = val; }
  void setTackleWidth (const float val) { tackleWidth = val; }
  void setVisibleAngle (const float val) { visibleAngle = val; }
  void setVisibleDistance (const float val) { visibleDistance = val; }
  void setWindAng (const float val) { windAng = val; }
  void setWindDir (const float val) { windDir = val; }
  void setWindForce (const float val) { windForce = val; }
  void setWindRand (const float val) { windRand = val; }

  int getGameLogVersion () const { return gameLogVersion; }
  int getGoalieMaxMoves () const { return goalieMaxMoves; }
  int getHalfTime () const { return halfTime; }
  int getHearDecay () const { return hearDecay; }
  int getHearInc () const { return hearInc; }
  int getHearMax () const { return hearMax; }
  int getMaxGoalKicks () const { return maxGoalKicks; }
  int getOlcoachPort () const { return olcoachPort; }
  int getPointToBan () const { return pointToBan; }
  int getPointToDuration () const { return pointToDuration; }
  int getPort () const { return port; }
  int getRecvStep () const { return recvStep; }
  int getSayCoachCntMax () const { return sayCoachCntMax; }
  int getSayCoachMsgSize () const { return sayCoachMsgSize; }
  int getSayMsgSize () const { return sayMsgSize; }
  int getSendStep () const { return sendStep; }
  int getSendViStep () const { return sendViStep; }
  int getSenseBodyStep () const { return senseBodyStep; }
  int getSimulatorStep () const { return simulatorStep; }
  int getSlowDownFactor () const { return slowDownFactor; }
  int getStartGoalL () const { return startGoalL; }
  int getStartGoalR () const { return startGoalR; }
  int getSynchMicroSleep () const { return synchMicroSleep; }
  int getSynchOffset () const { return synchOffset; }
  int getTackleCycles () const { return tackleCycles; }
  int getTextLogCompression () const { return textLogCompression; }
  char* getGameLogDir () const { return gameLogDir; }
  char* getGameLogFixedName () const { return gameLogFixedName; }
  char* getLandmarkFile () const { return landmarkFile; }
  char* getLogDateFormat () const { return logDateFormat; }
  char* getTextLogDir () const { return textLogDir; }
  char* getTextLogFixedName () const { return textLogFixedName; }
  bool getCoach () const { return coach; }
  bool getCoachWReferee () const { return coachWReferee; }
  bool getOldCoachHear () const { return oldCoachHear; }
  bool getWindNone () const { return windNone; }
  float getWindRandom () const { return windRandom; }
  bool getBackPasses () const { return backPasses; }
  bool getForbidKickOffOffside () const { return forbidKickOffOffside; }
  bool getFreeKickFaults () const { return freeKickFaults; }
  bool getFullstateL () const { return fullstateL; }
  bool getFullstateR () const { return fullstateR; }
  bool getGameLogDated () const { return gameLogDated; }
  bool getGameLogFixed () const { return gameLogFixed; }
  bool getGameLogging () const { return gameLogging; }
  bool getLogTimes () const { return logTimes; }
  bool getProfile () const { return profile; }
  bool getProperGoalKicks () const { return properGoalKicks; }
  bool getRecordMessages () const { return recordMessages; }
  bool getSendComms () const { return sendComms; }
  bool getSynchMode () const { return synchMode; }
  bool getTeamActuatorNoise () const { return teamActuatorNoise; }
  bool getTextLogDated () const { return textLogDated; }
  bool getTextLogFixed () const { return textLogFixed; }
  bool getTextLogging () const { return textLogging; }
  bool getUseOffside () const { return useOffside; }
  bool getVerbose () const { return verbose; }
  float getAudioCutDist () const { return audioCutDist; }
  float getBallAccelMax () const { return ballAccelMax; }
  float getBallDecay () const { return ballDecay; }
  float getBallRand () const { return ballRand; }
  float getBallSize () const { return ballSize; }
  float getBallSpeedMax () const { return ballSpeedMax; }
  float getBallWeight () const { return ballWeight; }
  float getCatchProbability () const { return catchProbability; }
  float getCatchableAreaL () const { return catchableAreaL; }
  float getCatchableAreaW () const { return catchableAreaW; }
  float getCkickMargin () const { return ckickMargin; }
  float getControlRadius () const { return controlRadius; }
  float getDashPowerRate () const { return dashPowerRate; }
  float getEffortDec () const { return effortDec; }
  float getEffortDecThr () const { return effortDecThr; }
  float getEffortInc () const { return effortInc; }
  float getEffortIncThr () const { return effortIncThr; }
  float getEffortInit () const { return effortInit; }
  float getEffortMin () const { return effortMin; }
  float getGoalWidth () const { return goalWidth; }
  float getInertiaMoment () const { return inertiaMoment; }
  float getKickPowerRate () const { return kickPowerRate; }
  float getKickRand () const { return kickRand; }
  float getKickRandFactorL () const { return kickRandFactorL; }
  float getKickRandFactorR () const { return kickRandFactorR; }
  float getKickableMargin () const { return kickableMargin; }
  float getMaxmoment () const { return maxmoment; }
  float getMaxneckang () const { return maxneckang; }
  float getMaxneckmoment () const { return maxneckmoment; }
  float getMaxpower () const { return maxpower; }
  float getMinmoment () const { return minmoment; }
  float getMinneckang () const { return minneckang; }
  float getMinneckmoment () const { return minneckmoment; }
  float getMinpower () const { return minpower; }
  float getOffsideActiveAreaSize () const { return offsideActiveAreaSize; }
  float getOffsideKickMargin () const { return offsideKickMargin; }
  float getPlayerAccelMax () const { return playerAccelMax; }
  float getPlayerDecay () const { return playerDecay; }
  float getPlayerRand () const { return playerRand; }
  float getPlayerSize () const { return playerSize; }
  float getPlayerSpeedMax () const { return playerSpeedMax; }
  float getPlayerWeight () const { return playerWeight; }
  float getPrandFactorL () const { return prandFactorL; }
  float getPrandFactorR () const { return prandFactorR; }
  float getQuantizeStep () const { return quantizeStep; }
  float getQuantizeStepL () const { return quantizeStepL; }
  float getRecoverDec () const { return recoverDec; }
  float getRecoverDecThr () const { return recoverDecThr; }
  float getRecoverMin () const { return recoverMin; }
  float getSlownessOnTopForLeftTeam () const { return slownessOnTopForLeftTeam; }
  float getSlownessOnTopForRightTeam () const { return slownessOnTopForRightTeam; }
  float getStaminaIncMax () const { return staminaIncMax; }
  float getStaminaMax () const { return staminaMax; }
  float getStoppedBallVel () const { return stoppedBallVel; }
  float getTackleBackDist () const { return tackleBackDist; }
  float getTackleDist () const { return tackleDist; }
  float getTackleExponent () const { return tackleExponent; }
  float getTacklePowerRate () const { return tacklePowerRate; }
  float getTackleWidth () const { return tackleWidth; }
  float getVisibleAngle () const { return visibleAngle; }
  float getVisibleDistance () const { return visibleDistance; }
  float getWindAng () const { return windAng; }
  float getWindDir () const { return windDir; }
  float getWindForce () const { return windForce; }
  float getWindRand () const { return windRand; }

  private:
    int catchBanCycle;
    int clangAdviceWin;
    int clangDefineWin;
    int clangDelWin;
    int clangInfoWin;
    int clangMessDelay;
    int clangMessPerCycle;
    int clangMetaWin;
    int clangRuleWin;
    int clangWinSize;
    int coachPort;
    int dropBallTime;
    int freeformSendPeriod;
    int freeformWaitPeriod;
    int gameLogCompression;
    int gameLogVersion;
    int goalieMaxMoves;
    int halfTime;
    int hearDecay;
    int hearInc;
    int hearMax;
    int maxGoalKicks;
    int olcoachPort;
    int pointToBan;
    int pointToDuration;
    int port;
    int recvStep;
    int sayCoachCntMax;
    int sayCoachMsgSize;
    int sayMsgSize;
    int sendStep;
    int sendViStep;
    int senseBodyStep;
    int simulatorStep;
    int slowDownFactor;
    int startGoalL;
    int startGoalR;
    int synchMicroSleep;
    int synchOffset;
    int tackleCycles;
    int textLogCompression;
    char *gameLogDir;
    char *gameLogFixedName;
    char *landmarkFile;
    char *logDateFormat;
    char *textLogDir;
    char *textLogFixedName;
    bool coach;
    bool coachWReferee;
    bool oldCoachHear;
    bool windNone;
    float windRandom;
    bool backPasses;
    bool forbidKickOffOffside;
    bool freeKickFaults;
    bool fullstateL;
    bool fullstateR;
    bool gameLogDated;
    bool gameLogFixed;
    bool gameLogging;
    bool logTimes;
    bool profile;
    bool properGoalKicks;
    bool recordMessages;
    bool sendComms;
    bool synchMode;
    bool teamActuatorNoise;
    bool textLogDated;
    bool textLogFixed;
    bool textLogging;
    bool useOffside;
    bool verbose;
    float audioCutDist;
    float ballAccelMax;
    float ballDecay;
    float ballRand;
    float ballSize;
    float ballSpeedMax;
    float ballWeight;
    float catchProbability;
    float catchableAreaL;
    float catchableAreaW;
    float ckickMargin;
    float controlRadius;
    float dashPowerRate;
    float effortDec;
    float effortDecThr;
    float effortInc;
    float effortIncThr;
    float effortInit;
    float effortMin;
    float goalWidth;
    float inertiaMoment;
    float kickPowerRate;
    float kickRand;
    float kickRandFactorL;
    float kickRandFactorR;
    float kickableMargin;
    float maxmoment;
    float maxneckang;
    float maxneckmoment;
    float maxpower;
    float minmoment;
    float minneckang;
    float minneckmoment;
    float minpower;
    float offsideActiveAreaSize;
    float offsideKickMargin;
    float playerAccelMax;
    float playerDecay;
    float playerRand;
    float playerSize;
    float playerSpeedMax;
    float playerWeight;
    float prandFactorL;
    float prandFactorR;
    float quantizeStep;
    float quantizeStepL;
    float recoverDec;
    float recoverDecThr;
    float recoverMin;
    float slownessOnTopForLeftTeam;
    float slownessOnTopForRightTeam;
    float staminaIncMax;
    float staminaMax;
    float stoppedBallVel;
    float tackleBackDist;
    float tackleDist;
    float tackleExponent;
    float tacklePowerRate;
    float tackleWidth;
    float visibleAngle;
    float visibleDistance;
    float windAng;
    float windDir;
    float windForce;
    float windRand;
};


class Audio {
  public:
    Audio::Audio ();
    virtual ~Audio ();

    //virtual ostream& print ( ostream& o ) const = 0;
    //friend ostream& operator<< ( ostream& o, const Audio& a ) { return a.print ( o ); }

    int getTime () const { return time; }
    void setTime ( const int t ) { time = t; }
  private:
    int time;
};


class RefAudio : public Audio {
  public:
    RefAudio ();
    virtual ~RefAudio ();

    //virtual ostream& print ( ostream& o ) const;
    //ostream& print ( ostream& o, const Side& ) const { return print ( o ); }


    Playmode getPlaymode () const { return playmode; }
    void setPlaymode ( const Playmode pm ) { playmode = pm; }
  PenaltySide getPenaltySide() const { return penaltySide; }
  void setPenaltySide( const PenaltySide ps ) { penaltySide = ps; }
    int getGoals() const { return goals; }
    void setGoals(const int g) { goals = g; }
    char const * getUnknownRefMsg () const { return unknownRefAudio; }
    void setUnknownRefMsg ( const char *m ) {
      strncpy(unknownRefAudio, m, msgLength-1);
      unknownRefAudio[msgLength-1] = '\0';
    }

  private:
    Playmode    playmode;
    PenaltySide penaltySide;
    int goals;
    static const int msgLength = 32;
    char unknownRefAudio[msgLength];
};


class SelfAudio : public Audio {
  public:
    SelfAudio ();
    virtual ~SelfAudio ();

    char const * getMsg () const { return msg; }
    void setMsg ( const char *m ) {
      strncpy(msg, m, msgLength-1);
      msg[msgLength-1] = '\0';
    }

  private:
    static const int msgLength = 128;
    char msg[msgLength];
};


enum Side {
  S_Left  = 0, 
  S_Right = 1, 
  S_Our   = 2, 
  S_Opp   = 3, 
  S_Neutral = 4, 
  S_Unknown = 5
};


class PlayerAudio : public SelfAudio {
  public:
    PlayerAudio ();
    virtual ~PlayerAudio ();

    //virtual ostream& print ( ostream& o ) const;
    //ostream& print ( ostream& o, const Side& ) const { return print ( o ); }

    float getDir () const         { return dir;  }
    void setDir ( const float d ) { dir = d;     }
    int getUnum () const          { return unum; }
    void setUnum ( const int u )  { unum = u;    }
    int getSide () const          { return side; }
    void setSide ( const Side s ) { side = s;    }

  private:
    float dir;
    int unum;
    Side side;
};


class CoachAudio : public Audio {
  public:
    CoachAudio ();
    virtual ~CoachAudio ();

    char const * getMsg () const { return msg; }
    void setMsg ( const char *m ) {
      strncpy(msg, m, msgLength-1);
      msg[msgLength-1] = '\0';
    }

    //virtual ostream& print ( ostream& o ) const;
    //ostream& print ( ostream& o, const Side& ) const;

  private:
    static const int msgLength = 128;
    char msg[msgLength];
};


class Error {
  public:
    Error ();
    virtual ~Error ();

    Error_t getType () const { return type; }
    void setType ( const Error_t e ) { type = e; }
    void setUnknownErrorMsg(const char *msg)
    {
      strncpy(unknownErrorMsg, msg, msgLength-1);
      unknownErrorMsg[msgLength-1] = '\0';
    }

    //ostream& print ( ostream& o, const Side& ) const
    //{
    //  o << Error::NAMES [ error_ ];
    //  if ( error_ == Error::say_message_too_long )
    //    o << " " << time_recv;
    //  return o;
    //}

  private:
    Error_t type;
    static const int msgLength = 32;
    char unknownErrorMsg[msgLength];
};


class Warning {
  public:
    Warning ();
    virtual ~Warning ();

    Warning_t getType () const { return type; }
    void setType ( const Warning_t w ) { type = w; }
    void setUnknownWarningMsg(const char *msg)
    {
      strncpy(unknownWarningMsg, msg, msgLength-1);
      unknownWarningMsg[msgLength-1] = '\0';
    }

    //ostream& print ( ostream& o, const Side& ) const
    //{
    //  o << Warning::NAMES [ error_ ];
    //  if ( error_ == Warning::say_message_too_long )
    //    o << " " << time_recv;
    //  return o;
    //}

  private:
    Warning_t type;
    static const int msgLength = 32;
    char unknownWarningMsg[msgLength];
};


class InitMsg {
  public:
    InitMsg ();
    virtual ~InitMsg ();

    //  ostream&
    //  print ( ostream& o ) const
    //  {
    //    if ( reconnect() )
    //      o << "(reconnect " << ( side_ == LEFT ? "l" : "r" );
    //    else
    //      o << "(init " << ( side_ == LEFT ? "l" : "r" ) << " "<< unum << " ";

    //    return play_mode.print ( o, side_ ) << ")";
    //  }

    Side getSide () const  { return side; }
    void setSide ( const Side s ) { side = s; }
    int getUnum () const  { return unum; }
    void setUnum ( const int un )  { unum = un; }
    Playmode getPlaymode () const  { return playmode; }
    void setPlayMode ( const Playmode pm )  { playmode = pm; }
  private:
    Side     side;
    int	     unum;
    Playmode playmode;
};



class PlayerParam {
public:
  PlayerParam ( );
  virtual ~PlayerParam ();

  void setAll(
              const int player_types,
              const int pt_max,
              const float random_seed,
              const int subs_max,
              const float dash_power_rate_delta_max,
              const float dash_power_rate_delta_min,
              const float effort_max_delta_factor,
              const float effort_min_delta_factor,
              const float extra_stamina_delta_max,
              const float extra_stamina_delta_min,
              const float inertia_moment_delta_factor,
              const float kick_rand_delta_factor,
              const float kickable_margin_delta_max,
              const float kickable_margin_delta_min,
              const float new_dash_power_rate_delta_max,
              const float new_dash_power_rate_delta_min,
              const float new_stamina_inc_max_delta_factor,
              const float player_decay_delta_max,
              const float player_decay_delta_min,
              const float player_size_delta_factor,
              const float player_speed_max_delta_max,
              const float player_speed_max_delta_min,
              const float stamina_inc_max_delta_factor
             );

  int getPlayerTypes () const { return playerTypes; }
  int getPtMax () const { return ptMax; }
  float getRandomSeed () const { return randomSeed; }
  int getSubsMax () const { return subsMax; }
  float getDashPowerRateDeltaMax () const { return dashPowerRateDeltaMax; }
  float getDashPowerRateDeltaMin () const { return dashPowerRateDeltaMin; }
  float getEffortMaxDeltaFactor () const { return effortMaxDeltaFactor; }
  float getEffortMinDeltaFactor () const { return effortMinDeltaFactor; }
  float getExtraStaminaDeltaMax () const { return extraStaminaDeltaMax; }
  float getExtraStaminaDeltaMin () const { return extraStaminaDeltaMin; }
  float getInertiaMomentDeltaFactor () const { return inertiaMomentDeltaFactor; }
  float getKickRandDeltaFactor () const { return kickRandDeltaFactor; }
  float getKickableMarginDeltaMax () const { return kickableMarginDeltaMax; }
  float getKickableMarginDeltaMin () const { return kickableMarginDeltaMin; }
  float getNewDashPowerRateDeltaMax () const { return newDashPowerRateDeltaMax; }
  float getNewDashPowerRateDeltaMin () const { return newDashPowerRateDeltaMin; }
  float getNewStaminaIncMaxDeltaFactor () const { return newStaminaIncMaxDeltaFactor; }
  float getPlayerDecayDeltaMax () const { return playerDecayDeltaMax; }
  float getPlayerDecayDeltaMin () const { return playerDecayDeltaMin; }
  float getPlayerSizeDeltaFactor () const { return playerSizeDeltaFactor; }
  float getPlayerSpeedMaxDeltaMax () const { return playerSpeedMaxDeltaMax; }
  float getPlayerSpeedMaxDeltaMin () const { return playerSpeedMaxDeltaMin; }
  float getStaminaIncMaxDeltaFactor () const { return staminaIncMaxDeltaFactor; }


  void setPlayerTypes (const int val) { playerTypes = val; }
  void setPTMax (const int val)  {  ptMax = val; }
  void setRandomSeed (const float val)  {  randomSeed = val; }
  void setSubsMax (const int val)  { subsMax = val; }
  void setDashPowerRateDeltaMax (const float val)  {  dashPowerRateDeltaMax = val; }
  void setDashPowerRateDeltaMin (const float val)  {  dashPowerRateDeltaMin = val; }
  void setEffortMaxDeltaFactor (const float val)  {  effortMaxDeltaFactor = val; }
  void setEffortMinDeltaFactor (const float val)  {  effortMinDeltaFactor = val; }
  void setExtraStaminaDeltaMax (const float val)  {  extraStaminaDeltaMax = val; }
  void setExtraStaminaDeltaMin (const float val)  {  extraStaminaDeltaMin = val; }
  void setInertiaMomentDeltaFactor (const float val)  {  inertiaMomentDeltaFactor = val; }
  void setKickRandDeltaFactor (const float val)  {  kickRandDeltaFactor = val; }
  void setKickableMarginDeltaMax (const float val)  {  kickableMarginDeltaMax = val; }
  void setKickableMarginDeltaMin (const float val)  {  kickableMarginDeltaMin = val; }
  void setNewDashPowerRateDeltaMax (const float val)  {  newDashPowerRateDeltaMax = val; }
  void setNewDashPowerRateDeltaMin (const float val)  {  newDashPowerRateDeltaMin = val; }
  void setNewStaminaIncMaxDeltaFactor (const float val)  {  newStaminaIncMaxDeltaFactor = val; }
  void setPlayerDecayDeltaMax (const float val)  {  playerDecayDeltaMax = val; }
  void setPlayerDecayDeltaMin (const float val)  {  playerDecayDeltaMin = val; }
  void setPlayerSizeDeltaFactor (const float val)  {  playerSizeDeltaFactor = val; }
  void setPlayerSpeedMaxDeltaMax (const float val)  {  playerSpeedMaxDeltaMax = val; }
  void setPlayerSpeedMaxDeltaMin (const float val)  {  playerSpeedMaxDeltaMin = val; }
  void setStaminaIncMaxDeltaFactor (const float val)  {  staminaIncMaxDeltaFactor = val; }


  //friend ostream& operator<< (ostream& o, const PlayerParam& sp)
  //{
  //  return sp.print ( o );
  //}


  private:
    int playerTypes;
    int ptMax;
    float randomSeed;
    int subsMax;
    float dashPowerRateDeltaMax;
    float dashPowerRateDeltaMin;
    float effortMaxDeltaFactor;
    float effortMinDeltaFactor;
    float extraStaminaDeltaMax;
    float extraStaminaDeltaMin;
    float inertiaMomentDeltaFactor;
    float kickRandDeltaFactor;
    float kickableMarginDeltaMax;
    float kickableMarginDeltaMin;
    float newDashPowerRateDeltaMax;
    float newDashPowerRateDeltaMin;
    float newStaminaIncMaxDeltaFactor;
    float playerDecayDeltaMax;
    float playerDecayDeltaMin;
    float playerSizeDeltaFactor;
    float playerSpeedMaxDeltaMax;
    float playerSpeedMaxDeltaMin;
    float staminaIncMaxDeltaFactor;
};


class CoachFreeform {
  public:
    CoachFreeform();
    virtual ~CoachFreeform();

    void setOffsideX(float x){
	offsideX = x;
    }
    void setFormation(int fm) {
      formation = fm;
    }
    void setPlayerType(int unum, int type){
	playerTypes[unum-1] = type;
    }
    const int *getPlayerTypes() const { return &playerTypes[0]; }
    float getOffsideX() const { return offsideX; }
    const int getFormation() const { return formation; }

  private:
    int playerTypes[11];
    float offsideX;	
    int formation;
};

class PlayerType {
  public:
    PlayerType ();
    virtual ~PlayerType ();

    void setAll(
                const int id,
                const float player_speed_max,
                const float stamina_inc_max,
                const float player_decay,
                const float inertia_moment,
                const float dash_power_rate,
                const float player_size,
                const float kickable_margin,
                const float kick_rand,
                const float extra_stamina,
                const float effort_max,
                const float effort_min
               );

    int   getTypeId () const { return id; }
    float getPlayerSpeedMax () const { return playerSpeedMax; }
    float getStaminaIncMax () const { return staminaIncMax; }
    float getPlayerDecay () const { return playerDecay; }
    float getInertiaMoment () const { return inertiaMoment; }
    float getDashPowerRate () const { return dashPowerRate; }
    float getPlayerSize () const { return playerSize; }
    float getKickableMargin () const { return kickableMargin; }
    float getKickRand () const { return kickRand; }
    float getExtraStamina() const { return extraStamina; }
    float getEffortMax () const { return effortMax; }
    float getEffortMin () const { return effortMin; }
    float getActualPlayerSpeedMax () const { return playerSpeedMaxAct; }


  private:
    int id;
    float playerSpeedMax;
    float staminaIncMax;
    float playerDecay;
    float inertiaMoment;
    float dashPowerRate;
    float playerSize;
    float kickableMargin;
    float kickRand;
    float extraStamina;
    float effortMax;
    float effortMin;
  float playerSpeedMaxAct;
};


class Substitution {
  public:
    Substitution ();
    virtual ~Substitution ();

    //virtual ostream& print ( ostream& o ) const
    //{
    //  o << "(change_player_type " << unum;
    //  if ( type_ != PlayerType::UNKNOWN_ID )
    //    o << " " << type_;
    //  return o << ")";
    //}

    void setUnum( const int u ) { unum = u; }
    void setType( const int t ) { type = t; }

    int getUnum () const { return unum; }
    int getType () const { return type; }

  private:
    int unum;
    int	type;
};


enum Location {
  //in order to be compatible to rccparser
  Loc_Top = 2,
  Loc_Bottom = 3,
  Loc_Left = 4,
  Loc_Right = 5,
  Loc_Center = 6,
  Loc_Penalty = 7,
  Loc_Goal = 8
};

}; //Sense


#endif //SENSEOBJECTS_HH

