/**********************************************************************
 * main.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   function main()
 *
 * requires:
 *
 * Created on: 21.04.2002 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 * $Id: main.cc,v 1.3 2004/03/26 11:52:05 harlekin Exp $
 *
 **********************************************************************/

#include <iostream>
#include <sys/time.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>

#include "Agent.hh"
#include "SenseThread.hh"
#include "ConfigFile.hh"
#include "debug.hh"

//forward
void  showHelp();
void  sigintHandler(int s);
void  doTerminate();
void  printStatistics();

const int KEYWORD_NOT_FOUND = -1;

/// Returns KEYWORD_NOT_FOUND if keyword was not found or the index position
int findKeyword(char* key, int argc, char** argv)
{
  int i;
  for (i = 0; i < argc; i++)
  {
    if(strcmp(argv[i], key) == 0)
    {
      return i;
    }
  }
  return KEYWORD_NOT_FOUND;
}

void removeKeyword(char* key, int& argc, char** argv)
{
  int index = findKeyword(key, argc, argv);
  if (index != KEYWORD_NOT_FOUND)
  {
    while (index < argc-1)
    {
      argv[index] = argv[index+1];
      index++;
    }
    argc--;
  }
}

int main(int argc, char** argv){
  DEBUG1_SYNCHRO("[INIT] \tmain(): called");

#ifdef  USE_PLANNER
  if (findKeyword("-net", argc, argv) != KEYWORD_NOT_FOUND)
  {
    cerr << "Keyword -net found." << endl;
    Agent::useNetExecutor = true;
    removeKeyword("-net", argc, argv);
  }
  else
  {
    Agent::useNetExecutor = false;
    cerr << "Keyword -net NOT found." << endl;
  }
#endif

  if((argc == 4 && strcmp(argv[3], "-goalie")) || argc > 4 ){
    showHelp();
    exit(-1);
  }

  if(!ConfigFile::parseConfigFile("player.conf")){
    cerr << "Unable to parse <player.conf>!" << endl;
    exit(-1);
  }
  Formations::readFormations("formations.conf");
  Formations::setCurrentFormation(ConfigFile::standardFormation);


  //by omission of commandline parameters take the config file values
  const char *serverHost = ConfigFile::serverHost.c_str();
  const char *teamname   = ConfigFile::teamname.c_str();
  int         serverPort = ConfigFile::serverPort;
  bool        goalie     = false;


  switch(argc){
  case 4:
    goalie=true;
  case 3:
    ConfigFile::teamname   = argv[2];
  case 2:
    ConfigFile::serverHost = argv[1];
    serverHost = argv[1];
  default:
    break;
  }

  

  Agent agent    = Agent(serverHost, serverPort, teamname, goalie);
  WorldModel *wm = Agent::getWorldModel();
  Sensors    *s  = Agent::getSensors();
  Param param(wm, s);


  //start the senseThread with pthread_create()
  int rc;
  pthread_t sense_thr;
  rc = pthread_create(&sense_thr, NULL,
                      &SenseThread::senseCallback, &param);
  if(rc)
    cerr << "pthread_create() failed, rc=" << rc << ", errno=" << errno << endl;


  // init CTRL-C handler
  signal(SIGINT, &sigintHandler);

  //start the reasoning thread
  agent.init();
  agent.run();

  printStatistics();

  DEBUG1("main(): exiting...");
  exit(rc);
}

void printStatistics(){
  ServerCommandCounters &qcc =
    Agent::getWorldModel()->getQueuedCommandCounters();
  BodyInfo &bi =
    Agent::getWorldModel()->myself.bodyInfo;
  cout << "**************************************************************\n";
  cout << "kicks        queued: " << qcc.kicks       << "  \t server received: "
    << bi.getKickCount() << endl;
  cout << "dashes       queued: " << qcc.dashes      << "  \t server received: "
    << bi.getDashCount() << endl;
  cout << "turns        queued: " << qcc.turns       << "  \t server received: "
    << bi.getTurnCount() << endl;
  cout << "says         queued: " << qcc.says        << "  \t server received: "
    << bi.getSayCount() << endl;
  cout << "turn_necks   queued: " << qcc.turnNecks   << "  \t server received: "
    << bi.getTurnNeckCount() << endl;
  cout << "catches      queued: " << qcc.catches     << "  \t server received: "
    << bi.getCatchCount() << endl;
  cout << "moves        queued: " << qcc.moves       << "  \t server received: "
    << bi.getMoveCount() << endl;
  cout << "change_views queued: " << qcc.changeViews << "  \t server received: "
    << bi.getChangeViewCount() << endl;
  cout << "tackles      queued: " << qcc.tackles << "  \t server received: "
    << bi.getTackleCount() << endl;
  cout << "pointTos     queued: " << qcc.pointTos << "  \t server received: "
    << bi.getArmCount() << endl;
  cout << "attentionTos queued: " << qcc.attentionTos << "  \t server received: "
    << bi.getFocusCount() << endl;


  cout << "**************************************************************\n\n";
}


void sigintHandler(int s){
  cout << "\n\n SIGINT caught " << s << endl;
  printStatistics();
  exit(0);
}


void doTerminate(){
  cout << "\n\n";
  cout << "The server probably dead: exiting" << endl;
  printStatistics();
  exit(0);
}


void showHelp(){
  cout << "usage: agent <host <teamname <-goalie>>>" << endl;
}

