/******************************************************************************
 * Behaviour.cc
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Behaviour
 *
 * requires:
 *
 *
 * Created on: ??.??.2003 by majunor        
 * Maintainer: harlekin
 * Group:
 * Co-authors: harlekin, cniehaus, tbogon
 *
 * Changelog:
 *
 * $Id: Behaviour.cc,v 1.91 2004/06/29 10:57:44 mosh Exp $
 *
 ******************************************************************************/
#include "Agent.hh"
#include "debug.hh"


/** The Constructor initializes the (mostly) private Variables*/
Behaviour::Behaviour(){
  worldModel = Agent::getWorldModel();
  myself = &(worldModel->myself);
  ball = &(worldModel->ball);
  skillPool = Agent::getSkillPool();
  usePriorityPassing = ConfigFile::usePriorityPassing;
  runsFree = false;
  currentlyDribbling = false;
  kickInMinDelay = 0; 
  kickInScanDelay = 0;
  kickInState  = 0;
}

/** Destructor destroying previously instantiated Variables
    Currently it does nothing, but this may change over Time
*/
Behaviour::~Behaviour(){
}

// true:  right
// false: left
bool Behaviour::getDribbleSide(){
  const Vector2D *ballPos   = ball->getPositionPt();
  float myBodyDir = myself->getBodyDir();

  //consider field borders
  if( fabs(ballPos->getY()) >= 32 ){
    if( ballPos->getY() > 0 ){
      return ( myBodyDir > 90 );
    }else{
      return ( myBodyDir < 90 );
    }
  }
  if( fabs(ballPos->getX()) > 50 ){
    if( ballPos->getX() > 0 ){
      return ( myBodyDir > 0 );
    }else{
      return ( myBodyDir < 0 );
    } 
  }
     
  int opponents =  worldModel->numInSetInCone(PS_Opponents,myself->getPosition(), myself->getPosition() + Vector2D(8, myself->getBodyDir(), POLAR), 1);
  
  if (opponents == 0){
    //return ballPos->getY() < 0;

    // keep ball on the actualcurrentnutten side
    if (worldModel->angleToBallRel2Body() < 0) return false;
    else return true;
    
  }
  
  const Player* nme = worldModel->closestInSetToBall(PS_Opponents);
  
  if(nme == NULL){
    if (worldModel->angleToBallRel2Body() < 0) return false;
    else return true;
  }else{
    if(worldModel->angleToPointRel2Body(nme->getPosition()) < 0) return true;
    else return false;
  }
} 

/** Returns the potential Pass Candidate
    @return the Pass Candidate
    @param offensive if this is an offensive pass, default value is true
*/
const Player* Behaviour::getPassCandidate(bool offensive){
  return worldModel->getPassCand(offensive);
}


/** Returns if stopping the Ball is necessary
    @return true, if the Ball needs to be stopped, false othewise
*/
bool Behaviour::isStopBallNeeded(){
  if(currentlyDribbling){
    currentlyDribbling = false;
    return false;
  }
  static float powerThr = ConfigFile::stopBallPowerThr;
  float angle = worldModel->angleToBallRel2Body();
  float dist  = worldModel->distanceToBall();
  static Vector2D ballVel;
  ballVel.setVector2D(ball->getVelocity());
  float power = (ballVel.getLength() / ball->calculateKickRate(dist, angle)) / Configuration::kickPowerRate;
  
  if(power <= powerThr){
    return false;
  }else{
    return true;
  }
}

/** Returns if running free is a good choice, for a Player at the overgiven Position
    A Good Choice is the following:
    1. we have the Ball
    2. I'm currently intersected from the Player with the Ball
    3. If I run 10m in front, i will be no longer intersected
    @param player the Vector of the Player that should run free
    @return true, if it is a good choice, false otherwise
    @todo this is currently in a testing Stage DO NOT USE YET
*/
bool Behaviour::shouldRunFree(const Vector2D &player){
  if(!worldModel->hasBallTeam()){
    return false;
  }
  if(!worldModel->isIntersected(player)){
    return false;
  }
  float playerX = player.getX();
  float playerY = player.getY();
  Vector2D freeRunVector(playerX+10, playerY);
  if(worldModel->isIntersected(freeRunVector)){
    return false;
  }
  return true;
}

bool Behaviour::shouldDribble(){
  // Should not (normal/complex) dribble if I am attacked
  if (worldModel->isAttacked(worldModel->myself))
    return false;
  // Should not dribble if out of stamina
  float st = 1500;
  if (myself->getStamina() <= st)
    return false;
  return true;
}

// Determine if the point is a good point to pass to
// The pass to this point IS secure. Just determine if
// it is really needed or a good idea.
bool Behaviour::shouldPassTo(const Vector2D &pos){
  // Should pass to players who are at a good point to score
  if (worldModel->isAGoodPointToScore(pos)) 
    return true;  
  //Pass dont seem to make sense
  if (pos.distanceTo( myself->getPositionPt()) < 2)
    return false; 
 // Should pass to players who are in the enemy penalty area
  if (Field::OPP_PENALTY_AREA.isInside(pos))
    return true;
 
  Vector2D goal = Vector2D(52,0);
  // Should pass to players who are in front of me
  if (pos.distanceTo(goal) < myself->getPositionPt()->distanceTo(goal) - 10)
    return true;
  // Should pass to players who are in front of me
  if (pos.distanceTo(goal) < myself->getPositionPt()->distanceTo(goal) - 7
      && pos.getX() > 10)
    return true;
   if (pos.distanceTo(goal) < myself->getPositionPt()->distanceTo(goal) - 4
      && pos.getX() > 25)
    return true;
    
  
  return false;
}

/** Returns if the Player needs to turn to a Pass Reciever
    @param passCand the PassCandidate
    @return true, if a turn is needed, false otherwise
*/
bool Behaviour::isTurnToPassCandNeeded(const Player* passCand){
  Vector2D targetPos = passCand->getPosition();
  if ((fabs(worldModel->angleToPointRel2Head(targetPos))>15)||
      (fabs(worldModel->angleToPointRel2Head(targetPos))>15)){
    return true;
  }
  return false;
}

/** Returns if the currentSeeTime is equal to the currentTime
    @return true if the currentTime is equal to the last See
*/
// TODO: Why is this in the behavour? Should go into the worldModel
bool Behaviour::isLastSeeAtCurrentTime(){
  return worldModel->getLastSeeSSTime() == *(worldModel->getTime());
}

/** Handles the Behaviour if the Agent has the Ball (the Ball is kickable)*/
void Behaviour::playOnWithBall(){
  // SHOULD NEVER _"NEVER"_ NEVER BE USED!
  DEBUG1("Behaviour::playOnWithBall() called: Should NEVER happen!");
  exit(-1);
}

bool Behaviour::acceptPass(){
  // HACK: Dont use this function
  return false;

  if (worldModel->passPointToGoCycles == 0 ||
      worldModel->passPointToGo.isInvalid()){
    // Reseting information concerning passPoint
    worldModel->passPointToGo.setVector2D(NAN, NAN);
    worldModel->passPointToGoCycles = 0;
    return false;
  }
  worldModel->passPointToGoCycles--;

  if (skillPool->goToPoint(worldModel->passPointToGo, 1.0, 100)){
    // Reached passpoint
    worldModel->passPointToGo.setVector2D(NAN, NAN);
    worldModel->passPointToGoCycles = 0;
    return false;
  }
  // Running towards the communicated point
  lookAround(true);
  return true;
}

/** Handles the behaviour if the Agent does not have the Ball (the Ball is not kickable)*/
void Behaviour::playOnWithoutBall(){
  // if the player looses the ball while dribbling he would
  // otherwise still think to be dribbling if he gets into the
  // reach of another ball... later... uhm... and would not stop it...
  currentlyDribbling = false;
  //APL_POINT3_MOSH(myself->getPosition(),APLLevel(1),255,0,0); 
  const Player *fastest = worldModel->fastestInSetToBall(PS_Teammates, true);
  if(fastest != NULL) {
    if(worldModel->mayKickBall() && *fastest == *myself){
      if(skillPool->getBall()){	
	// should not happen
	skillPool->stopBall();
      }
      lookAround(true);
      return;
    }
  }else{  //no one is fastest to ball
    if(worldModel->mayKickBall() && worldModel->isClosestTeammateToBall(*myself)){
      if(skillPool->getBall()){       
	skillPool->stopBall();
      }
      lookAround(true);      
      return;
    }
  }
 

  // offensive look around Behaviour ==(Hack)=====>
  
  if(ball->getPosition().getX()>34){
    int ti = worldModel->getTime()->getCycles();    
    
    //Behaviour for players in good position for a pass a bevor the goal 
    if(fabs(ball->getPosition().getY())>15){
      if( ti%2 ==0)
      skillPool->faceOnlyNeckToPoint(Vector2D(40,0));
    else
      skillPool->faceOnlyNeckToPoint(ball->getPosition());
     return;
    }//Behaviour for players awaiting this pass to shoot a goal
    else{
      if( ti%2 ==0)
	skillPool->faceOnlyNeckToPoint(Vector2D(52,0));
      else
	skillPool->faceOnlyNeckToPoint(ball->getPosition());
      return;
    }
  }
  //
  

  //set enemy under pressure
  bool hasBallPossessionOpp = worldModel->hasBallPossessionOpp();
  if( hasBallPossessionOpp){
    /****  we need interceptionPoint for later Calculations  *****/
    static Vector2D intPoint; 
    int cycles;
    if(!worldModel->interceptionPoint(*myself, intPoint, cycles, false))
      intPoint=ball->getPosition(); 
    if(intPoint.getX() < myself->getPosition().getX()+5
       && myself->getPosition().distanceTo(intPoint) < 15
       && getStrategicPosition().distanceTo(intPoint) < 25 ){
      
      if(skillPool->getBall()){       
	skillPool->stopBall();
      }
      lookAround(true);      
      return;
    }
  }

  
  doPositioning();
}

void Behaviour::assistTeammate(){
  // Should I assist?
  if ( worldModel->isSecondClosestTeammateToPoint((const Player&)*myself, ball->getPosition(), ConfigFile::highConfidencePlayer, NULL ) ){

    // No rules for assisting on own half yet
    if (ball->getPositionPt()->getX() < -9){
      return playOnWithoutBall();
    }

    if (Formations::currentFormation->getBehaviour(myself->getUniformNumber()) < B_WingMidfielderLeft)
      return playOnWithoutBall();

    // Assist teammate on opponent half
    if (ball->getPositionPt()->getX() > -10){

      static Vector2D assistPoint;
      float ballY = ball->getPositionPt()->getY();
      /*
      if (fabs(ballY) < 8.0){
	assistPoint.setVector2D(Field::LENGTH/2 -  8,  8 * sign(myself->getPosition().getY()));
      }else{
	assistPoint.setVector2D(Field::LENGTH/2 - 11,  0);
           }
      */
      Vector2D stratPos= getStrategicPosition();

      Vector2D toBall=ball->getPosition()-stratPos;
      toBall.setLength(toBall.getLength()/3);

      Vector2D toGoal=Vector2D(52,0)-myself->getPosition();
      toGoal.setLength(ball->getPosition().getX()/5);
      Vector2D toX = Vector2D(12,0);
      assistPoint.setVector2D(stratPos+toBall+toGoal+toX);

      if(assistPoint.distanceTo(ball->getPosition())>15){
	assistPoint.setVector2D(stratPos+toBall+toGoal+toX);
      }else     
	assistPoint.setVector2D(stratPos+toGoal+toX);
      
      static Line myWay = Line(0,0,0);
      myWay.updateLineFromTwoPoints(myself->getPosition(), assistPoint);
      //Vector2D toGo = assistPoint;
      Vector2D cPoL = myWay.getPointOnLineClosestTo(ball->getPosition());
      float cPoL_length = (cPoL - ball->getPosition()).getLengthSq();
      
      if(assistPoint.distanceTo(ball->getPosition())<8){
	//assistPoint= (ball->getPosition() +
	//      (( assistPoint - ball->getPosition() ).setLength(8)) );
	assistPoint= (ball->getPosition() +      
		      (( (stratPos +toGoal +toX) - ball->getPosition() ).setLength(8)) );
	      }
     
      //freien punkt suchen
      // const Player *closestEnemy = worldModel->closestInSetToMe( PS_Opponents,0.95);
      const Player *closestEnemy = worldModel->closestInSetToMe( PS_Players,0.95);
      Vector2D current=assistPoint;
      float closestDist=5;
      if(closestEnemy!=NULL){
	Vector2D abweichung=Vector2D(2,0);
	for(int t =1 ;t <=8;t++)
	  if(closestEnemy->getPosition().distanceTo(current) <5 ){
	    abweichung.rotate(45);
	    Vector2D temp=assistPoint+abweichung;
	    if(closestEnemy->getPosition().distanceTo(temp)<closestDist 
	       < closestDist){
	      closestDist =closestEnemy->getPosition().distanceTo(temp);
	      current=assistPoint+temp;
	    }
	  }
	abweichung=Vector2D(4,0);
	for(int t =1 ;t <=8;t++)
	  if(closestEnemy->getPosition().distanceTo(current) <5 ){
	    abweichung.rotate(45);
	    Vector2D temp=assistPoint+abweichung;
	    if(closestEnemy->getPosition().distanceTo(temp)<closestDist 
	     < closestDist){
	      closestDist =closestEnemy->getPosition().distanceTo(temp);
	      current=temp;
	    }
	    
	  }
	abweichung=Vector2D(6,0);
	for(int t =1 ;t <=8;t++)
	  if(closestEnemy->getPosition().distanceTo(current) <5 ){
	    abweichung.rotate(45);
	    Vector2D temp=assistPoint+abweichung;
	    if(closestEnemy->getPosition().distanceTo(temp)<closestDist 
	     < closestDist){
	      closestDist =closestEnemy->getPosition().distanceTo(temp);
	      current=temp;
	    }
	    
	  }
      }
      assistPoint=current;

      if (false){//cPoL_length < 12.25 /* 3.5 */ && 
	
      }
      else{      
	float offsideX = worldModel->getOffsideX();
	if (offsideX < assistPoint.getX()){
	  offsideX--; // avoid offside
	  assistPoint.setVector2D( offsideX, myWay.getYGivenX(offsideX) );
	}		
	
	//keep point in field
	if (!Field::FIELD.isInside(assistPoint)){
	  if(assistPoint.getY()>31)
	    assistPoint.setY(31);
	  if(assistPoint.getY()<-31)
	  assistPoint.setY(-31);
	  if(assistPoint.getX()>52)
	  assistPoint.setX(51);
	}
	
	
	//laufe nicht in den ballfuehrenden Spieler
	if(ball->getPosition().distanceTo(myself->getPosition())<3
	   && ball->getPosition().getX() > myself->getPosition().getX()){
	  skillPool->faceOnlyBodyToPoint(worldModel->field.getMarkerPosition(GOAL_R));
	}
	else
	  if (skillPool->goToPoint(assistPoint, 2.0, 100)){
	    skillPool->faceOnlyBodyToPoint(worldModel->field.getMarkerPosition(GOAL_R));
	  }     
      }
    }
    lookAround(true);
    return;
  }
  return playOnWithoutBall();
}


/** The standard play On Behaviour calls specific Behaviours*/
void Behaviour::playOnBehaviour(){
  // If I am told to accept a pass: Do it!
  if (acceptPass()) return;

  bool hasBallPossessionOpp = worldModel->hasBallPossessionOpp();
  if (worldModel->mayKickBall() && worldModel->isBallKickable()){
    return playOnWithBall();
  }else{
    if (!hasBallPossessionOpp)
      return assistTeammate();
    else
      return playOnWithoutBall();
  }
}

/**
 * This method determines wether I should perform a freeKick/kickIn,
 * assist or just ignore it.
 *
 * @param  pos    Where the freeKick is performed (ballPos)
 * @param  assist In this bool will be stored if I should assist
 * @return        if I should perform the freeKick
 *
 **/
bool Behaviour::shouldPerformKickIn(const Vector2D &pos, bool &assist){
  if (pos.isInvalid()){
    assist = false;
    return false;
  }

  const Formation* f = Formations::getCurrentFormation();
  int cUnum = 0;
  int aUnum = 0;
  float aDist = BIG_FLOAT; // assist
  float cDist = BIG_FLOAT; // perform
  static Vector2D cPos, aPos;
  cPos.setVector2D(UNDEFINED_FLOAT, NAN);
  aPos.setVector2D(UNDEFINED_FLOAT, NAN);

  bool ignoreDefenders = true; //ball->getPositionPt()->getX() > 0;
  // Determine the closest and second closest distances
  for (int i = 2; i <= 11; ++i){
    // Ignore players we have not seen for some time (crashed agents)
    if (worldModel->getPlayerByUnum(i)->getConfidence(worldModel->getTime()) == 0) continue;
    if (ignoreDefenders)
      if (f->getBehaviour(i) < B_WingMidfielderLeft)
	continue;
    const Vector2D &hP = f->getHomePosition(i);
    float dist = hP.distanceToSq(pos);
    if (dist < cDist){
      aUnum = cUnum;
      aDist = cDist;
      cDist = dist;
      cUnum = i;
    }else if(dist < aDist){
      aDist = dist;
      aUnum = i;
    }
  }

  int lUnum     = 0; // leftest (perform)
  int laUnum    = 0; // second leftest (assist)
  float maxDist = 289; // 17 //cDist + 3.0;
  
  // Determine the two players who are farthest left but close
  for (int i = 2; i <= 11; ++i){
    // Ignore players we haven't seen for a LONG time
    if (worldModel->getPlayerByUnum(i)->getConfidence(worldModel->getTime()) == 0) continue;
    if (i == cUnum) continue;
    if (ignoreDefenders)
      if (f->getBehaviour(i) < B_WingMidfielderLeft)
	continue;
    const Vector2D &hP = f->getHomePosition(i);
    float dist = hP.distanceToSq(pos);
    if (dist < maxDist){
      if (hP.getX() < cPos.getX()){
	aPos   = cPos;
	laUnum = lUnum;
	cPos   = hP;
	lUnum  = i;
      }else if (hP.getX() < aPos.getX()){
	aPos   = hP;
	laUnum = i;
      }
    }
  }

  int unum, unumA;

  // Do we have a "leftest" player?
  if (lUnum != 0){
    unum = lUnum;
  }else{ // take the closest
    unum = cUnum;
  }

  // Do we have a second "leftest" player?
  if (laUnum != 0){
    unumA = laUnum;
  }else{
    // None found, take closest
    if (unum != cUnum){
      unumA = cUnum; // take closest
    }else{ // take the secondclosest
      unumA = aUnum;
    }
  }

  // store and return values
  assist = (unumA == myself->getUniformNumber());
  return (unum == myself->getUniformNumber());
}

// Also used for kickIns. Should be considered
void Behaviour::freeKickBehaviour(){
  const Vector2D &ballPos = ball->getPosition();
  const Player* goalie = worldModel->myTeam.getGoalie();
  if(goalie != NULL){ // if I know the goalie
    // and he is near the ball   
    if(worldModel->distanceFromPlayerToBall(*goalie) < 2.0){ 
      // and the ball is inside the penalty area
      if(Field::OWN_PENALTY_AREA.isInside(ballPos)){ 
	doPositioning();
	return;	    
      }
    }
  }

  // Use the same method as for kickIn
  bool perform, assist;
  perform = shouldPerformKickIn(ballPos, assist);

  if (perform && assist) assist = false;

  if ( assist ){
    static Vector2D assistPoint;
    assistPoint.setVector2D(Field::OPP_PENALTY_POINT - ball->getPosition());
    assistPoint.setLength(ConfigFile::assistanceDistance);
    assistPoint += ball->getPosition();

    // TODO: Should consider offside during indirect_freekick
    if (!worldModel->isOffside(assistPoint)){
      if (skillPool->goToPoint(assistPoint, 1.0, 80)){
	skillPool->faceOnlyBodyToBall();
      }
      lookAround(true);
      return;
    }
  }

  if (!perform){
    doPositioning();
    return;
  }

  // perform

  static Vector2D opponent_goal_mid = worldModel->field.getMarkerPosition(GOAL_R);
  double prefferedDistance = Configuration::getPlayerSize(myself->getType()) +
    Configuration::getKickableMargin(myself->getType())/2 +
    Configuration::ballSize;
  int passMinDelay  = ConfigFile::passMinDelay;
  int passScanDelay = ConfigFile::passScanDelay;
  float power = ConfigFile::walkPower;
  Vector2D freeKickPosition = ballPos + Vector2D(prefferedDistance, worldModel->angleFromPointToPoint(opponent_goal_mid, ballPos),POLAR); 

  // walk slow
  if (myself->getPosition().distanceToSq(freeKickPosition) < 12.25 /* 3.5 */){
    power = 42;
  }

  Playmode pm = worldModel->getPlaymode();
  bool ready = pm == PM_OurFreeKick || 
    pm == PM_OurIndirectFreeKick || 
    pm == PM_OurKickIn;
    
  // Perform the kickin
  if(skillPool->goToPoint(freeKickPosition, prefferedDistance/2, power)){
    if(skillPool->faceOnlyBodyToPoint(opponent_goal_mid)){
      if (ready){
	if(!performKickIn(true, passMinDelay, passScanDelay, 1.2)){
	  skillPool->kickAway(false, true);
	} // ]pass partner found? 
      } // ]waiting for ready
    } // ]turning to point
  } // ]going to point
  lookAround(true);
}

/** The general Behaviour that tells the Agent what to do in the current Situation, e.g Free Kick, play on*/
void Behaviour::generalBehaviour(){
  //the following number are not cycles, but the number of calls of this
  //function, ratio to the cycles: 5:3 im default view mode
  static int  passMinDelay     = ConfigFile::passMinDelay;
  static int  passScanDelay    = ConfigFile::passScanDelay;
  static int  walkPower        = ConfigFile::walkPower;
  static bool fsUsed           = ConfigFile::useFullstate;
  static bool penaltyPerformed = false;
  static int  nextToShoot      = 2;

  static Vector2D penShootPoint     = Vector2D(40, 0);
  static Vector2D opponent_goal_mid = worldModel->field.getMarkerPosition(GOAL_R);
  static Vector2D opponent_rc       = worldModel->field.getMarkerPosition(FLAG_RB);
  static Vector2D opponent_lc       = worldModel->field.getMarkerPosition(FLAG_RT);
  
  /*   Test fuer Anstoss */
  static Vector2D own_goal = worldModel->field.getMarkerPosition(FLAG_PLC);
  Playmode pm = worldModel->getPlaymode();
  const Vector2D &ballPos = ball->getPosition();

  int stuckedForCycles = myself->bodyInfo.getTackleExpires();
  if (stuckedForCycles > 0) {
    DEBUG1_HARLEKIN("WorldModel::generalBehaviour(): Standing up...");
    lookAround(true);
    return; 
  }

  worldModel->debugSeeTimes();

  /*
  if (pm != PM_BeforeKickOff && myself->bodyInfo.getViewWidth() == VW_NARROW){
    if (!worldModel->haveOptimalSeeTimes()){
      skillPool->doSSProfiling();
      DEBUG1("[CRITICAL] Game started before SSProfiling finished!");
    }else{
      DEBUG1("[CRITICAL] Agent thinks Profiling is done but has wrong VIEW!");
    }
  }
  */
  
  if(pm != PM_BeforeKickOff){
    if(!ball->isPositionKnown()){
      //if(!ball->isGoodConfident(worldModel->getTime())){
      skillPool->scanFieldWithBody();
      return;
    }
  }
  
  if (ConfigFile::useDifferentViewAngles &&
      (pm != PM_BeforeKickOff || worldModel->getTime()->getTime() > 0)){
    if (worldModel->switchViewNeeded())
      skillPool->doSwitchView();
  }

  switch(pm){
  case PM_BeforeKickOff:
    if( !fsUsed ){
      if(!worldModel->haveOptimalSeeTimes()){
	skillPool->doSSProfiling();
	lookAround(true);
	return;
      }
    } 
    
    if(!skillPool->moveToInitialPosition()){
      lookAround(false);
    }else if(!ball->isPositionKnown()){
      skillPool->scanFieldWithBody();
    }else{
      skillPool->faceNeckAndBodyToBall();
    } 
    
    break;
    case PM_PlayOn:
      //reset all flags
      resetKickIn();
      playOnBehaviour();
      break;
    case PM_OurKickOff:   //kick off in the middle
      if(worldModel->isClosestTeammateToBall(*myself)){
	if(skillPool->getBall(walkPower, false, true)){
	  const Player* cand = worldModel->getPassCand(true, false, false);
	  if (cand != NULL){
	    skillPool->passBallToPlayer(cand);
	    lookAround(true);
	  }else{
	    skillPool->kickToPoint(Vector2D(-20,0), 90 * ball->calculateKickRate(worldModel->distanceToBall(), worldModel->angleToBallRel2Body()), true);
	    lookAround(true);
	  }
	}
      }else{
	skillPool->faceNeckAndBodyToBall();
      }      
      break;
  case PM_TheirKickOff:
    if(!skillPool->moveToInitialPosition()){
      lookAround(true);
    }else{
      skillPool->faceNeckAndBodyToBall();
    }
    break;
  case PM_TheirOffside:  //lasts a bit, after it comes freekick
  case PM_OurIndirectFreeKick:
  case PM_TheirFreeKickFault: // PM_OurFreeKick will follow
  case PM_TheirBackPass: // dito
  case PM_OurFreeKick:  //after the goalie has catched the ball,
    freeKickBehaviour();
    break;
  case PM_OurKickIn:    //unser Einwurf
    kickInBehaviour();
    break;
  case PM_TheirIndirectFreeKick:
  case PM_TheirFreeKick:
  case PM_OurBackPass: // PM_TheirFreeKick will follow
  case PM_OurFreeKickFault: // dito
  case PM_TheirKickIn:   //deren Einwurf
  case PM_OurGoalKick: // nothing to do
  case PM_TheirGoalKick: // nothing to do
  case PM_OurOffside:
    doPositioning();
    break;
  case PM_OurCornerKick:
    if( worldModel->isMyHomePosClosestToPos(ballPos) ){
      if(skillPool->getBall(walkPower, false)){
	if(skillPool->faceOnlyBodyToPoint(penShootPoint)){
	  if(!performKickIn(true, passMinDelay, passScanDelay, 1.5)){
	    skillPool->kickToPoint(penShootPoint, 100, true);
	    lookAround(true);
	  }
	}
      }
    }else if(skillPool->goToPositionOurCKick()){
      skillPool->faceNeckAndBodyToBall();
    }    
    break;
  case PM_TheirCornerKick:
    if(skillPool->goToPositionTheirCKick()){
      skillPool->faceOnlyBodyToBall();
    }
    lookAround(true);
    break;
  case PM_TheirGoal:
  case PM_OurGoal:
    if (worldModel->getTime()->getStoppedCount() + 8 >= 50){
      if(!skillPool->moveToInitialPosition()){
	lookAround(false);
      }else{
	skillPool->faceOnlyBodyToPoint(opponent_goal_mid);
	skillPool->faceOnlyNeckToBall();
      }
    }else{
      // TODO: If moveLikeAnIdiot supports TheirGoal this should
      //       be changed.
      if(pm == PM_TheirGoal){
	if (skillPool->moveToInitialPosition()){
	  skillPool->faceOnlyBodyToPoint(Vector2D(0,0));
	}
      }else
	skillPool->moveLikeAnIdiot();	
      lookAround(true);
    }
    break;
  case PM_OurPenaltyKick:        //penalty kick
      if( worldModel->isMyHomePosClosestToPos(ballPos) ){
	if(skillPool->getBall(walkPower, false)){
	  if(skillPool->faceOnlyBodyToPoint(opponent_goal_mid)){
	    if(!performKickIn(true, passMinDelay, passScanDelay, 1.2)){
	      skillPool->scoreGoal();
	    }
	  }
	}
      }else{
	if(myself->getPosition().getX() > (worldModel->getOffsideX()-3) ){
	  doPositioning();
	  break;
	}
	doPositioning();
      }
      break;

  case PM_TheirPenaltyKick:
  case PM_OurGoalieCatchBall:
  case PM_TheirGoalieCatchBall:
    doPositioning ();
    break;    
  case PM_OurPenaltySetup:
    penaltyPerformed = false;
    if(myself->getUniformNumber() == nextToShoot){
      skillPool->getBall();
    }else{
      skillPool->goToPoint(Vector2D(0, 0), 3.0,80.0);
    }
    break;
  case PM_OurPenaltyTaken:
  case PM_OurPenaltyReady:
    if(myself->getUniformNumber() == nextToShoot){
      doPenalty();
    }
    break;
  case PM_OurPenaltyScore:
  case PM_OurPenaltyMiss:
    if (penaltyPerformed == false){
      nextToShoot = (nextToShoot % 11) + 1;
      penaltyPerformed = true;
    }
    break;
  case PM_TheirPenaltyTaken:
  case PM_TheirPenaltyScore:
  case PM_TheirPenaltyMiss:
  case PM_TheirPenaltyReady: // goalie should perform
    break;
  case PM_TheirPenaltySetup:
    skillPool->goToPoint(Vector2D(0, 0), 8.0, 80);
    break;
  case PM_TimeOver:
    break;
  default:
    DEBUG2("Behaviour::fieldPlayerGeneralBehaviour():\
             not supported playmode appeared: ", pm);
  }
  
  // Communicate Ballposition, GoodPositioning etc.
  reportStatus();

}

void Behaviour::reportStatus(){
  // report the ball data in every cycle!
  /*
  if (worldModel->passPointToSayCycles != 0){
  DEBUG1_HARLEKIN("Behaviour::reportStatus(): communicating good pass position");
  worldModel->passPointToSayCycles--;
  skillPool->reportPassPoint();
  return;
  }
  */
  if (worldModel->isClosestToBall(*myself)  && 
      ball->isHighConfident(worldModel->getTime()) &&
      ball->isVelHighConfident(worldModel->getTime()) &&
      worldModel->getPlaymode() != PM_BeforeKickOff){
    skillPool->reportBallStatus(currentlyDribbling);
    return;
  }
  if (Field::OPP_PENALTY_AREA.isInside(myself->getPosition()) &&
      Field::OPP_PENALTY_AREA.isInside(ball->getPosition())){
    skillPool->reportPosition();
    return;
  }
  if (worldModel->getPlaymode() == PM_PlayOn &&
      worldModel->isAGoodPointToScore(myself->getPosition())){
    skillPool->reportPosition();
    return;
  }  
}

/** Handles Penalty Shootout*/
void Behaviour::doPenalty(){
  PenaltySide ps = worldModel->getPenaltySide();
  if(!skillPool->getBall()){
    return;
  }
  float myX = myself->getPosition().getX();
  Vector2D whereToDribble(UNDEFINED_FLOAT, UNDEFINED_FLOAT);
  if (isAGoodPenaltyShotPoint()){
    if(ps != PS_OurGoal){
      skillPool->scoreGoal();
    }else{
      skillPool->scoreGoal(true);
    }
    return;
  }
  if(myX < 25){
    whereToDribble.setX(myself->getPositionPt()->getX() + 20);
    whereToDribble.setY(myself->getPositionPt()->getY());
    if (ps == PS_OurGoal){
      whereToDribble.setX(whereToDribble.getX() - 40);
    }
    skillPool->simpleDribbleToPoint(whereToDribble, 4); 
    return;
  }else{
    whereToDribble.setX(myself->getPositionPt()->getX() + 10);
    whereToDribble.setY(myself->getPositionPt()->getY());
    if (ps == PS_OurGoal){
      whereToDribble.setX(whereToDribble.getX() - 20);
    }
    // do other dribbling here
    skillPool->simpleDribbleToPoint(whereToDribble, 30); 
    return;
  }
}

bool Behaviour::tryToForceScoreGoal(){
  //DEBUG1_MOSH("TRY_FORCE");
  //return false;
  bool shouldTry = false;
  const Vector2D &ballPos = ball->getPosition();
  float ballX = ballPos.getX();
  const Vector2D &myPos = myself->getPosition();

  // Just in case. 
  // If the current half is nearly over try to
  // score a goal of kick the ball away if we
  // are defending.
  int cycle = worldModel->getTime()->getTime();
  if (cycle < 12000 && cycle % 3000 > 2950){
    shouldTry = true;
    // timeplay
    if (ballX < 0.0){
      skillPool->kickAway(true, true);
      return true;
    }
  }
 
  if (ballX < 30.0){
    return false;
  }

  const Player* goalie = worldModel->theirTeam.getGoalie();
  if (goalie != NULL && goalie->isHighConfident(worldModel->getTime())){
    const Vector2D &goaliePos = goalie->getPosition();
    // dont consider him if he is behind us
    if (goaliePos.getX() >= myPos.getX()){
      float distToGoalie = goaliePos.distanceToSq(myPos);
      // dont shoot if he is too far away
      if (distToGoalie < 4.0f /*3.5f*/){
	shouldTry = true;
      }
    }
  }

  bool shouldTackle = false;
  Vector2D shortestPoint = Vector2D(NAN, NAN); 

  Line l1 = Line::makeLineFromTwoPoints(myPos, myPos + Vector2D(10, myself->getBodyDir(), POLAR));
  float cross = l1.getYGivenX(Field::OPP_GOAL_OUTLINE);
  if (ConfigFile::useTackleForScoring && shouldTry &&
      worldModel->isBallTackable() &&
      fabs(cross) < Field::GOAL_WIDTH/2 && fabs(myself->getBodyDir()) < 90.0 ){
    shouldTackle = true; 
    //shouldTry = true;
  }else{
    // get closest point with high prob
    int width = (int)((Field::GOAL_WIDTH - 1.5f)/2.0f);
    float    shortestDist  = BIG_FLOAT;
    float    shortestProb  = 0.0f;
    Vector2D goalMid = worldModel->field.getMarkerPosition(GOAL_R);
    for (int i = -width; i < width; ++i){
      Vector2D point = goalMid + Vector2D(0, i);
      float dist = point.distanceToSq(ballPos);
      float prob = worldModel->probToVector(ballPos, point);
      if (prob > shortestProb){
	shortestProb  = prob;
	shortestDist  = dist;
	shortestPoint = point;
      }
    }
    if (shortestProb < 0.50f){
      // wouldn't hit the goal
      return false;
    }
    if (shortestDist < 16){//156.25f /*12.5f*/){
      shouldTry = true;
    }
  }


  if (shouldTry){
    if (shouldTackle){
      // use brute force
      skillPool->tackle(100);
      APL_ELLIPSE3_MOSH(Vector2D(-10,10),0.5,0.5, (APLLevel) 4  , 255, 255, 0);    DEBUG1_MOSH("FORCE1");
      return true;
    }
    skillPool->kickToPoint(shortestPoint, 100, true);
    APL_ELLIPSE3_MOSH(Vector2D(-10,-10),0.5,0.5, (APLLevel) 4  , 255, 255, 0);    
    DEBUG1_MOSH("FORCE2");
    return true;    
  }
  // no chance
  return false;
}

void Behaviour::kickInBehaviour( int wantedRecoveryDelay ) {
  // counter for stamina-recovery delay
  kickInMinDelay++;
  static bool watchingLeft = false;
  static int kickInTimer = 0;

  // do I have to perform the kick in?

  bool assist, perform;
  perform = shouldPerformKickIn(ball->getPosition(), assist);

  // Just to be sure (can't do both)
  if (assist && perform) assist = false;
  
  static Vector2D assistPoint;
  //assistPoint.setVector2D(Field::OPP_PENALTY_POINT - ball->getPosition());
  if(ball->getPosition().getY()>0){
    assistPoint =(Vector2D(52,10) - ball->getPosition());
  }
  else{
    assistPoint =(Vector2D(52,-10) - ball->getPosition());
  }

  assistPoint.setLength(ConfigFile::assistanceDistance);
  assistPoint += ball->getPosition();  
  APL_ELLIPSE3_MOSH(assistPoint,0.5,0.5, (APLLevel) 4  , 255, 255, 0);    

  // I shall not perform so assist...
  if ( assist ){
    
    if (skillPool->goToPoint(assistPoint, 1.0, 80)){
      skillPool->faceOnlyBodyToBall();
    }
    lookAround(true);
    return;
  }
  
  // I should not assist and not perform 
  if ( !perform ) {
    doPositioning();
    return;
  }
  
  static Vector2D shootingPos;
  // adjust position
  shootingPos.setVector2D(ball->getPosition());
  if ( shootingPos.getY() > 0 ) {
    shootingPos.moveBy( 0.0f,  0.5f);
  } else {
    shootingPos.moveBy( 0.0f, -0.5f);
  }

  // go to the ball - returns true, if we are close to the ball
  // tolerance is 0.5m
  if ( !skillPool->goToPoint( shootingPos, 0.5, 42 ) ) {
    // player is going to the ball
    kickInTimer = 0;
    lookAround(true);
    return;
  }      
  
  // LOOK AROUND (while standing next to the ball)
  // Calculate angle
  float angleToFaceTo = 0.0;
  if (shootingPos.getX() <= 0.0){
    angleToFaceTo = -45.0 * sign(shootingPos.getY());
  }else{
    // if we just got a see, calculate the new angle
    if (isLastSeeAtCurrentTime())
      watchingLeft = !watchingLeft;	  
    if (watchingLeft){
      float angleToBorder = 0.0;
      if (shootingPos.getY() < 0.0){
	angleToBorder = worldModel->angleFromPointToPoint(shootingPos, worldModel->field.getMarkerPosition(FLAG_CB));
	angleToFaceTo = angleToBorder - 45.0;
      }else{
	angleToBorder = worldModel->angleFromPointToPoint(shootingPos, worldModel->field.getMarkerPosition(FLAG_CT));	    	    
	angleToFaceTo = angleToBorder + 45.0;
      }
    }else{
      float angleToBorder = 0.0;
      if (shootingPos.getY() < 0.0){
	angleToBorder = worldModel->angleFromPointToPoint(shootingPos, Vector2D(Field::LENGTH/2, -Field::PENALTY_AREA_WIDTH/2));
	angleToFaceTo = angleToBorder + 45.0;
      }else{
	angleToBorder = worldModel->angleFromPointToPoint(shootingPos, Vector2D(Field::LENGTH/2,  Field::PENALTY_AREA_WIDTH/2));	    	    
	angleToFaceTo = angleToBorder - 45.0;
      }	    
    }	  
  }
  
  // calc the point to look to (after getting the angle)
  static Vector2D pointToFace;
  pointToFace.setVector2D(myself->getPosition() + Vector2D(20, angleToFaceTo, POLAR));
  // turn body to enemy penalty point (the assisting player is
  // standing in that direction)
  //if (!skillPool->faceOnlyBodyToPoint(Vector2D(42, 0))){
  if (!skillPool->faceOnlyBodyToPoint(assistPoint)){
    lookAround(true);
    return;
  } 

  // wait at least 30 "cycles" (after positioning)
  if (kickInTimer++ < 30){
    skillPool->faceOnlyNeckToPoint(pointToFace);	  
    return;
  }
  // wait at least wantedRecoveryDelay-cycles (at all)
  if (kickInMinDelay < wantedRecoveryDelay){
    lookAround(true);
    return;
  }
  
  // PERFORM THE KICKIN
  // is our worldmodel up to date (see-message arrived this cycle)?
  if ( isLastSeeAtCurrentTime() ) {
    kickInTimer = 0;
    
    // get the best passcand (backpasses are allowed, if necessary)
  
    /*
    skillPool->kickToPoint( (assistPoint+Vector2D(5,0)), 75, false );
    return;
    
   Vector2D passPoint,passVel;
   bool thereIsAPassPoint = worldModel->getPassPoint(passPoint,passVel);
   if (thereIsAPassPoint){
     float passSpeed = passVelocity.getLength();
     float passPower = worldModel->calcKickPowerForSpeed( passSpeed );
     if(passPower < 50 )
       passPower=50;
     skillPool->kickToPoint( passPoint, passPower, false );
     lookAround(true);
     return;
   }
    */

  const Player *passCand = worldModel->getPassCand(true /*not used*/);
    
  if ( passCand != NULL){
    if (skillPool->passBallToPlayer( passCand, ConfigFile::passEndSpeed, true )){
      lookAround(true);
      return;
    }
  }
    

   skillPool->kickAway(false, true);
  } // else: no see arrived, wait a bit/cycle.
  lookAround(true);
  return;
}


/** Determines whether the current Player Position is god to score a goal on penalties
    @return true if it is a good point, false otherwise
*/
bool Behaviour::isAGoodPenaltyShotPoint(){
  PenaltySide ps = worldModel->getPenaltySide();
  if(ps != PS_OurGoal){
    return worldModel->isAtGoodPointToScore(*myself, false);
  }else{
    return worldModel->isAtGoodPointToScore(*myself, true);
  }
  return false;
}
/*******************************************************************************/
Vector2D Behaviour::getStrategicPosition(){
  static float posTolerance = ConfigFile::positioningTolerance;

  Playmode p = worldModel->getPlaymode();
  Formation *f = Formations::getCurrentFormation();
  Vector2D pos = myself->getHomePosition();
  Vector2D myPos = myself->getPosition();
  Vector2D ballPos = ball->getPosition();
  Vector2D relBallPos = ballPos - myself->getPosition();

  int unum = myself->getUniformNumber();
  bool behindBall = f->getBehindBall(unum);
  int behID = f->getBehaviour(unum);
  int speed = 80;

  // Do not run arround while stamina is low
  if (worldModel->distanceToBall() > 40.0){
    float s = myself->getStamina();
    float t = 3000;
    float s_inc = Configuration::getStaminaIncMax(myself->getType());
    if (s < t){
      speed = (int)(s_inc / 2.0);
    }
  }

  //ball attraction factors
  float factorX = f->getAttractionX(unum);
  float factorY = f->getAttractionY(unum);

  /*
  if (ballPos.getX() < -35.0f && behindBall){
    ballPos    = Vector2D(-52,0);
    relBallPos = ballPos - myself->getPosition();
  }
  */

  //dependence on the ball position
  pos.setX(pos.getX() + relBallPos.getX()*factorX);
  pos.setY(pos.getY() + relBallPos.getY()*factorY);

  bool nearOwn = ballPos.getX() < -30.0f && (behindBall || behID == B_CenterMidfielder) &&
    (p == PM_PlayOn || p == PM_TheirFreeKick || 
     p == PM_TheirIndirectFreeKick || p == PM_TheirKickIn || 
     p == PM_OurOffside || p == PM_OurBackPass || 
     p == PM_OurFreeKickFault);
  bool nearNME = ballPos.getX() > 30.0f && !behindBall &&
    (p == PM_PlayOn || p == PM_OurFreeKick || p == PM_OurIndirectFreeKick ||
     p == PM_OurKickIn || p == PM_TheirOffside || p == PM_TheirBackPass || 
     p == PM_TheirFreeKickFault);

  // Defenders should stay near own goal
  if (nearOwn){
    pos.setX(pos.getX() + (Vector2D(-52.0, 0.0) - myPos).getX() * factorX);
    pos.setY(pos.getY() + (Vector2D(-52.0, 0.0) - myPos).getY() * factorY);
  }
  
  // Attackers should stay near enemy goal
  if (nearNME){
    factorX = 0.8;
    pos.setX(pos.getX() + (Vector2D( 52.0, 0.0) - myPos).getX() * factorX);
    pos.setY(pos.getY() + (Vector2D( 52.0, 0.0) - myPos).getY() * factorY);
  }
  

  //avoid the boundaries of the field (width of field = 68.0)
  if (fabs(pos.getY()) > 30.0){ 
    pos.setY(pos.getY() - 0.5*factorY*relBallPos.getY());
  }

  // make sure the defence goes behind the ball
  // dont run behind the goalie
  if ((p != PM_OurFreeKick) && (p != PM_OurGoalKick)){
    /*
    if (f->getBehindBall(unum) && ballPos.getX() < pos.getX()){ 
      pos.setX(ballPos.getX());
      }*/
    if (behindBall){
      if (ballPos.getX() - 5.0 < pos.getX()){
	pos.setX(ballPos.getX() - 5.0);
      }
    }
  }

  //make sure no one goes closer to the enemies goal as the ball if the enemy
  //is kicking an offside free kick
  if (p == PM_OurOffside) pos.setX(min(pos.getX(), ballPos.getX()));

  //monitor an maximum and a minimum of the x-coord of the strategic position
  //if (!nearOwn && !nearNME){
  pos.setX(min((double)pos.getX(), f->getMaxX(unum)));
  pos.setX(max((double)pos.getX(), f->getMinX(unum)));
  //}

  // Consider offside...
  pos.setX(min(pos.getX(), worldModel->getOffsideX()-3));


  // try to cover a enemy play
  // stay away from the ball to save stamina
  goToStrategicPosition_StayAway(pos);
  
  return pos; 

  //go to this determined position
  // speed depends on stamina-recovery 
  //float dashPower = worldModel->getDashPowerForPositioning();
                                             // or faster if you are close to the ball
}


/*******************************************************************************/
bool Behaviour::goToStrategicPosition(){
  static float posTolerance = ConfigFile::positioningTolerance;

  Playmode p = worldModel->getPlaymode();
  const Formation *f = Formations::getCurrentFormation();
  static Vector2D pos;
  pos.setVector2D(myself->getHomePosition());
  const Vector2D &myPos   = myself->getPosition();
  const Vector2D &ballPos = ball->getPosition();
  static Vector2D relBallPos;
  relBallPos.setVector2D(ballPos - myself->getPosition());

  int unum = myself->getUniformNumber();
  bool behindBall = f->getBehindBall(unum);
  int behID = f->getBehaviour(unum);
  int speed = 80;

  // Do not 'run' around while stamina is low
  if (worldModel->distanceToBall() > 40.0){
    float s = myself->getStamina();
    float t = 3000;
    float s_inc = Configuration::getStaminaIncMax(myself->getType());
    if (s < t){
      speed = (int)(s_inc / 2.0);
    }
  }

  // ball attraction factors
  float factorX = f->getAttractionX(unum);
  float factorY = f->getAttractionY(unum);

  // dependence on the ball position
  pos.setVector2D((pos.getX() + relBallPos.getX()*factorX),
		  (pos.getY() + relBallPos.getY()*factorY));

  // Why and when should we use the following hacks:
  bool nearOwn = ballPos.getX() < -30.0f && (behindBall || behID == B_CenterMidfielder) &&
    (p == PM_PlayOn || p == PM_TheirFreeKick || 
     p == PM_TheirIndirectFreeKick || p == PM_TheirKickIn || 
     p == PM_OurOffside || p == PM_OurBackPass || 
     p == PM_OurFreeKickFault);
  bool nearNME = !behindBall && ballPos.getX() > 30.0f &&
    (p == PM_PlayOn || p == PM_OurFreeKick || p == PM_OurIndirectFreeKick ||
     p == PM_OurKickIn || p == PM_TheirOffside || p == PM_TheirBackPass || 
     p == PM_TheirFreeKickFault);

  // Evil HACK
  // Defenders should stay near own goal
  if (nearOwn){
    pos.setX(pos.getX() + (Vector2D(-52.0, 0.0) - myPos).getX() * factorX);
    pos.setY(pos.getY() + (Vector2D(-52.0, 0.0) - myPos).getY() * factorY);
  }
  
  // Evil HACK
  // Attackers should stay near enemy goal
  if (nearNME){
    factorX = 0.8;
    pos.setX(pos.getX() + (Vector2D( 52.0, 0.0) - myPos).getX() * factorX);
    pos.setY(pos.getY() + (Vector2D( 52.0, 0.0) - myPos).getY() * factorY);
  }
  
  // Avoid the boundaries of the field (width of field = 68.0)
  if (fabs(pos.getY()) > 30.0){ 
    pos.setY(pos.getY() - 0.5*factorY*relBallPos.getY());
  }

  // make sure the defense goes behind the ball
  if ((p != PM_OurFreeKick) && (p != PM_OurGoalKick)){
    if (behindBall){
      if (ballPos.getX() - 5.0 < pos.getX()){
	pos.setX(ballPos.getX() - 5.0);
      }
    }
  }

  // make sure no one goes closer to the 
  // enemies goal as the ball if the enemy
  // is kicking an offside free kick
  if (p == PM_OurOffside) 
    pos.setX(min(pos.getX(), ballPos.getX()));

  // monitor a maximum and a minimum of the 
  // x-coord of the strategic position
  pos.setX(min((double)pos.getX(), f->getMaxX(unum)));
  pos.setX(max((double)pos.getX(), f->getMinX(unum)));

  // Consider offside... 3 meter cause of inaccuracy and goTo-tolerance
  pos.setX(min(pos.getX(), worldModel->getOffsideX() - 3.0f));
  
  // try to cover an enemy play
  // NOT IMPLEMENTED

  // stay away from the ball to save stamina (during enemy freekick etc.)
  goToStrategicPosition_StayAway(pos);
  
  // go to the determined position
  // speed depends on stamina-recovery 
  //float dashPower = worldModel->getDashPowerForPositioning();
  return skillPool->goToPoint(pos, posTolerance,  speed); 
  // or faster if you are close to the ball
}

void Behaviour::goToStrategicPosition_StayAway(Vector2D &pos){
  Playmode p = worldModel->getPlaymode();
  
  if ((p == PM_TheirFreeKick)         || 
      (p == PM_TheirIndirectFreeKick) || 
      (p == PM_TheirKickOff)          || 
      (p == PM_TheirKickIn)){

    const Vector2D &myPos   = myself->getPosition();
    const Vector2D &ballPos = ball->getPosition();

    Vector2D posI1 = Vector2D(UNDEFINED_FLOAT, UNDEFINED_FLOAT);
    Vector2D posI2 = Vector2D(UNDEFINED_FLOAT, UNDEFINED_FLOAT);
    static Line myWay = Line(0,0,0);
    myWay.updateLineFromTwoPoints(myPos, pos);
    Circle boundary  = Circle(ballPos, 10.00f /*radius*/);
    Circle boundary2 = Circle(ballPos,  9.15f /*radius*/); 
    
    // Stay away if:
    //  - I am in the circle or I want to go into the circle
    if (boundary2.isInside(pos)){
      pos = myPos;
    }else{
    //  - my way at least crosses the circle once and
    //  - the ball/circle lies on my way (is not behind me)
      if (myWay.getCircleIntersectionPoints(boundary, &posI1, &posI2) != 0){  
	if (myWay.isInBetween(ballPos, myPos, pos)){
	  Vector2D closest = myWay.getPointOnLineClosestTo(ballPos);
	  closest = closest - ballPos;
	  closest.setLength(10.0f);
	  pos = ballPos + closest;
	}// Circle lies behind me
      }// My way does not cross the circle
    }
  }
}


/*******************************************************************************/
void Behaviour::doPositioning(){
  if(ball->isPositionKnown()){
    if(goToStrategicPosition()){
      skillPool->faceOnlyBodyToBall();
      lookAround(false);
    }else{
      if (worldModel->ball.isHighConfident(worldModel->getTime())){
	lookAround(false);
      }else{
	lookAround(true);
      }
    }
  }else{
    skillPool->scanFieldWithBody();
  }
}

/*******************************************************************************/
static bool noPassCand = false;
bool Behaviour::passIfHaveACandidate(bool offensive, float endSpeed){
  const Player *pc = worldModel->getPassCand(offensive, true, false);
  if ( pc != NULL ){
    skillPool->passBallToPlayer(pc, endSpeed, true);
    noPassCand = false;
    return true;
  }else{
    noPassCand = true;
    return false;
  }
}

void Behaviour::resetKickIn(){
  kickInState     = 0;
  kickInMinDelay  = 0;
  kickInScanDelay = 0;
}


/*******************************************************************************/
/**
 * WRITE DOCUMENTATION
 */
/*******************************************************************************/
bool Behaviour::performKickIn(bool offensive, int minDelay,
                              int scanDelay, float endSpeed){

  skillPool->scanFieldWithNeck();
  if( kickInMinDelay++ < minDelay )
    return true;

  if( kickInScanDelay++ >= scanDelay )
    return false;

  return passIfHaveACandidate(offensive, endSpeed);
}


/*******************************************************************************/
/**
 * the method tryToScoreGoal uses worldModel function getBestPointToShootTo
 * @return true if a good point to shoot to has been found, which has a good chance
 *              to let the ball end up in the goal by passing the goalie and
 *              find a secure way into the net. It differs between direct shooting
 *              and indirect shooting by laying the ball in front of the body 
 *              first.
 *	   false otherwise ...
 */
/*******************************************************************************/

bool Behaviour::shouldTryToScoreGoal(){
  //dont shoot at the goal if you are in your half of the field
  double x = worldModel->myself.getPositionPt()->getX();
  if (x <= 0.0) return false;

  // dont shoot if you have not seen the goal for some time
  int markerID;
  if (myself->getPositionPt()->getY() < 0 ){
    markerID = FLAG_GRB;
  }else{ 
    markerID = FLAG_GRT;
  }
  float goalConf = worldModel->field.getMarkerConfidence(markerID, worldModel->getTime());
  if (goalConf < 0.97) {
    return false;
  }
  return true;
}

bool Behaviour::tryToScoreGoal(){

  APL_ELLIPSE3_MOSH(Vector2D(48,25),0.5,0.5, (APLLevel) 4  , 255, 255, 0);    
  if (!shouldTryToScoreGoal()) return false;

  if (worldModel->theirTeam.getGoalie()->isGoodConfident(worldModel->getTime())){
    //DEBUG1_HARLEKIN("Versuche Torschuss mit bekanntem Torwart.");
    return tryToScoreGoalWithGoalie();
  }else{
    //return false;
    //DEBUG1_HARLEKIN("Versuche Torschuss mit unbekanntem Torwart.");
    return tryToScoreGoalWithoutGoalie();
  }
  // Should not be reached
  return false;
}

bool Behaviour::tryToScoreGoalWithGoalie(){

 

  static Vector2D point;
  DEBUG1_MOSH("with");
  point = worldModel->getBestPointToShootTo(ball->getPosition(), ConfigFile::powerForGoalScoring);
  if (!point.isInvalid()){
    // if there is a good point then shoot direct
    skillPool->kickToPoint(point, 100, true);
    APL_POINT3_MOSH(point,(APLLevel)4,255,0,0); 
    return true;
  }else{
    float kickMargin = Configuration::getKickableMargin(myself->getType());
    Vector2D ballNCFront = myself->predictedPosition() +
      Vector2D(kickMargin/4.0f, myself->getBodyDir(), POLAR);
    Vector2D pointFront = worldModel->getBestPointToShootTo(ballNCFront, ConfigFile::powerForGoalScoring);
    if (!pointFront.isInvalid()){
      // if there is a good point then shoot
      //skillPool->kickToPoint(pointFront, 100, false);
      //skillPool->kickCloseToBody(0);
      int myType = myself->getType();      
      skillPool->kickBallCloseToBody( 0, Configuration::getKickableMargin( myType )/4 );
      APL_POINT3_MOSH(pointFront,(APLLevel) 4,255,255,255); 
      return true;
    }
  }
  // none found
  return false;
}

bool Behaviour::tryToScoreGoalWithoutGoalie(){
 
 //sonst schiessen zu haeufig aus dem Mittelfeld
  if(ball->getPosition().distanceTo(Vector2D(52,0)) > 20 )
    return false;
  
 DEBUG1_MOSH("withou");
  static Vector2D point;
  point = worldModel->getClosestPointToShootTo(ball->getPosition(), ConfigFile::powerForGoalScoring);
  if (!point.isInvalid()){
    skillPool->kickToPoint(point, 100, true);
    APL_ELLIPSE3_MOSH(point,0.5,0.5, (APLLevel) 4  , 255, 255, 0);    
    return true;
  }
  // no good point was found
  return false;
}


/**********************************************************************************/
/**
 * WRITE DOCUMENTATION
 */
/**********************************************************************************/
bool Behaviour::doPenaltyScore(){
  PenaltySide ps = worldModel->getPenaltySide();
  const float KICK_POWER = 100;

  static Vector2D leftGoalCorner;
  static Vector2D rightGoalCorner;
  static Vector2D goalMid;
  if (ps == PS_TheirGoal){
    leftGoalCorner  = worldModel->field.getMarkerPosition(FLAG_GRT);
    rightGoalCorner = worldModel->field.getMarkerPosition(FLAG_GRB);
    goalMid         = worldModel->field.getMarkerPosition(GOAL_R);
  }else{
    leftGoalCorner  = worldModel->field.getMarkerPosition(FLAG_GLT);
    rightGoalCorner = worldModel->field.getMarkerPosition(FLAG_GLB);
    goalMid         = worldModel->field.getMarkerPosition(GOAL_L);
  }

  static Vector2D kickPoint, goaliePos;

  const Player *enemyGoalie = worldModel->theirTeam.getGoalie();
  goaliePos = enemyGoalie->getPosition();

  float yTop    = fabs((goaliePos - leftGoalCorner ).getY());
  float yBottom = fabs((goaliePos - rightGoalCorner).getY());
  
  if (yTop > yBottom){
    kickPoint.setVector2D(leftGoalCorner.getX(), leftGoalCorner.getY() + 0.5);
    skillPool->kickToPoint(kickPoint, KICK_POWER, true);
  }else{
    kickPoint.setVector2D(rightGoalCorner.getX(), rightGoalCorner.getY() - 0.5);
    skillPool->kickToPoint(kickPoint, KICK_POWER, true);
  }
  return true;
}


/**********************************************************************************/
/**
 * WRITE DOCUMENTATION
 */
/**********************************************************************************/
void Behaviour::lookAround(bool forceBall){
  // As there is no lookAround-Behaviour for different angles
  // just watch for the ball. Should be replaced...
  if ((ConfigFile::useDifferentViewAngles && !myself->isGoalKeeper()) ||
      (ConfigFile::useDifferentViewAnglesGoalie && myself->isGoalKeeper())){
    skillPool->faceOnlyNeckToBall();
    return;
  }

  static bool watchingLeft     = false; // which side of field to look to
  static int  watchingBall     = 1;     // should watch ball or field
  static bool watchingBallLeft = false; // which side of ball  to look to

  // if kicking the ball we should use the next position of it
  // otherwise we would foul ourself.
  bool calcPredicted = ((worldModel->isBallKickable()) && 
			(worldModel->getSentCommands().getFst() == CMD_KICK));

  // Switch between the views if we got a see this turn
  if( isLastSeeAtCurrentTime() ){
    if (!watchingBall)
      watchingLeft = !watchingLeft;
    else
      watchingBallLeft = !watchingBallLeft;
    watchingBall = (watchingBall + 1)%3;
  }

  // There are some situations in which it is useful to watch for
  // the ball at every cycle.
  if (forceBall) watchingBall = 1;
  
  // HACK! This method is not tested very good, yes...
  // So this deactivated the switching between 'Watch Ball'
  // and 'Watch Field'. 
  watchingBall = 1;


  float ballDir1;
  float ballDir2;
  float etcDir1;
  float etcDir2;

  // Get the position of the ball we perhaps will watch for
  static Vector2D ballPos;
  if (!calcPredicted){
    ballPos = ball->predictedPosition();
  }else{
    float kickDir = worldModel->getSentCommands().getFstParam2();
    float kickPwr = worldModel->getSentCommands().getFstParam1();
    kickDir += myself->getBodyDir();
    kickDir = normalizeTo180Deg(kickDir);
    ballPos = ball->predictedPosition(1, kickDir, kickPwr, 
				      worldModel->distanceToBall(),
				      worldModel->angleToBallRel2Body());
  }

  float angle    = worldModel->angleToPoint( ballPos );
  float distance = worldModel->distanceFromPointToPoint(myself->predictedPosition(), 
							ballPos);
  
  // This determines where we will watch exactly. Turn your neck so the ball will be near the
  // edge of the visible field. This way we keep an eye on the ball and another on the field.
  // Doesn't really make sense but sounds nice. ;)
  float changeAngle = (float)min( min( max( (double)distance * 2, 
					    ConfigFile::minViewChange ), 
				       ConfigFile::maxViewChange ), 
				  (double)myself->getViewAngle()/2 ); 
  ballDir1 = normalizeTo180Deg(angle + changeAngle);
  ballDir2 = normalizeTo180Deg(angle - changeAngle);

  // Determine wether we should watch for the goal
  static const Vector2D &goalR = worldModel->field.getMarkerPosition(GOAL_R);
  float angleToGoalFlag = worldModel->angleToPoint(goalR);  
  float ballAngle = normalizeTo180Deg(angle - myself->getBodyDir());
  
  // Get the right flag
  // Since there is no flag in the middle of the enemy goal we
  // have to use the goal posts. The farthest away post will be taken.
  // So we will see the larger part of the enemy penalty area.
  int flag_id = 0;
  if (myself->getPosition().getY() < 0) flag_id = FLAG_GRB;
  else                                  flag_id = FLAG_GRT;
  float flagConf = worldModel->field.getMarkerConfidence(flag_id, worldModel->getTime());

  // Every player should keep track of the enemy goal so he can
  // really decide if he is at a good point to score or not
  // which is important for communication. The goal will not move,
  // even the stupid agent is not THAT stupid. But looking out for
  // the goal keeps also track of enemy defenders and the annoying goalie.
  bool hasBall = worldModel->isBallKickable();
  float confidence = 0.90; // (Every tenth cycle)

  // If we are in the possession of the ball we should look out quite often.
  // (Every fourth cycle)
  if (hasBall) confidence = 0.96;
  
  // If the "informations" about the goal flag are obsolete and we are near
  // the goal (max. 32m) look in its direction.
  if (flagConf < confidence &&
      worldModel->distanceToPointSq(goalR) < 1024.0 /* 32.0 */){
    angle = angleToGoalFlag;
  }else{    
    // We decided not to watch for the goal. So what to do now?
    // determine which sides of the field should be watched if
    // we won't watch for the ball.
    // Those directions are for a view angle of 45. Other angles
    // will not work as intended.
    if (!watchingBall){
      if (ballAngle < -45.0){
	etcDir1 = normalizeTo180Deg(   0 + myself->getBodyDir());
	etcDir2 = normalizeTo180Deg(  90 + myself->getBodyDir());
      }else if (ballAngle > 45.0){
	etcDir1 = normalizeTo180Deg(   0 + myself->getBodyDir());
	etcDir2 = normalizeTo180Deg( -90 + myself->getBodyDir());
      }else{
	etcDir1 = normalizeTo180Deg(  90 + myself->getBodyDir());
	etcDir2 = normalizeTo180Deg( -90 + myself->getBodyDir());
      }
    }
    
    // Chose in which direction (ball or field) to look.
    if (watchingBall){
      if (watchingBallLeft) angle = ballDir1;
      else                  angle = ballDir2;
    }else{
      if (watchingLeft)     angle = etcDir1;
      else                  angle = etcDir2;
    }
  }

  // If we cant turn our neck that far just turn it towards
  // the ball. Even if thats not possible the maximum angle
  // will be used. So there is no need to recalculate the new
  // angle here.
  if (fabs(normalizeTo180Deg(angle - myself->getBodyDir())) > 
      Configuration::maxneckang){
    skillPool->faceOnlyNeckToBall();
    return;
  }

  // This only applies for WingDefenders in OurFreeKick-PlayMode.
  // if the goalie is performing a free kick in our penalty area
  // try to watch for both possible goalie jump points
  lookAround_GoalieKickOffPoint(angle);

  static Vector2D watchPoint;
  watchPoint = worldModel->myself.getPosition() + Vector2D(10, angle, POLAR); 
  skillPool->faceOnlyNeckToPoint(watchPoint);
  
  return;
}

void Behaviour::lookAround_GoalieKickOffPoint(float &angle){
  // This only applies for WingDefenders in OurFreeKick-PlayMode.
  // if the goalie is performing a free kick in our penalty area
  // try to watch for both possible goalie jump points
  if (worldModel->getPlaymode() == PM_OurFreeKick &&
      Field::OWN_PENALTY_AREA.isInside(ball->getPosition())){
    const Formation *f = Formations::getCurrentFormation();
    int behID = f->getBehaviour(myself->getUnum());

    if (behID == B_WingDefenderLeft ||
	behID == B_WingDefenderRight   ){
      if (myself->getPosition().distanceTo(Field::GOALIE_KICK_OFF_POINT_LEFT) <
	  myself->getPosition().distanceTo(Field::GOALIE_KICK_OFF_POINT_RIGHT)){			       
	angle = worldModel->angleFromPointToPoint(myself->getPosition(), 
						  Field::GOALIE_KICK_OFF_POINT_LEFT);
      }else{
	angle = worldModel->angleFromPointToPoint(myself->getPosition(), 
						  Field::GOALIE_KICK_OFF_POINT_RIGHT);
      }
    }
  }
}

// This method returns the best distance for dribbling.
// Because it uses all enemies it is quite expensive
int Behaviour::getDribbleDistance(float angle){
  Vector2D invVel = -myself->getVelocity();
  for (int i = 10; i > 2; --i){
    const Player* fastest;
    Vector2D point  = myself->getPosition() + Vector2D(  i, angle, POLAR);
    Vector2D point2 = myself->getPosition() + Vector2D(2*i, angle, POLAR);
    // Avoid the borders
    if (!Field::FIELD.isInside(point))  continue;
    if (!Field::FIELD.isInside(point2)) continue;
    int cycles = myself->predictedCyclesToPoint(point+invVel) + 1;
    fastest = worldModel->fastestInSetToPoint(PS_Opponents, point, cycles,
					      ConfigFile::highConfidencePlayer, false);
    if (fastest == NULL ){
      return i;
    }
  }
  return -1;
}

bool Behaviour::considerCouldPass(){
  // If we are attacked we should pass to save the ballposession
  // This method is only called if the agent believes dribbling is no option
  float passSpeed, passPower;
  if (canPassDirect && !canPassIndirect){
    skillPool->passBallToPlayer(passCand);
    return true;    
  }

  if (!canPassDirect && canPassIndirect){
    passSpeed = passVelocity.getLength();
    passPower = worldModel->calcKickPowerForSpeed( passSpeed );
    skillPool->kickToPoint( passPoint, passPower, false );
    return true;
  }

  if (canPassDirect && canPassIndirect){
    // Prefer points which are closer to the enemy goal
    if (passPoint.getX() > passCand->getPositionPt()->getX()){
      passSpeed = passVelocity.getLength();
      passPower = worldModel->calcKickPowerForSpeed( passSpeed );

      worldModel->passPointToSayCycles = 3;
      worldModel->passPointToSay = passPoint;
    
      skillPool->kickToPoint( passPoint, passPower, false );
      return true;
    }
     
    skillPool->passBallToPlayer(passCand);    
    return true;
  }

  // Can't pass
  return false;
}

bool Behaviour::considerShouldPass(){
  /* initialize values */
  passPoint    = Vector2D(NAN,NAN);
  passVelocity = Vector2D(NAN,NAN);
  passCand     = worldModel->getPassCand(true, true);
  canPassDirect   = passCand != NULL;
  canPassIndirect = worldModel->getPassPoint(passPoint, passVelocity);
  shouldPassDirect   = false;
  shouldPassIndirect = false;

  // Determine if we shall pass or only can pass
  if (canPassDirect && shouldPassTo(passCand->getPosition())){
      shouldPassDirect = true;    
  }
  if (canPassIndirect && shouldPassTo(passPoint)){
      shouldPassIndirect = true;   
  }
  
  static float passSpeed, passPower;
  // Determine which type of pass to use
  if (shouldPassDirect && !shouldPassIndirect){
    skillPool->passBallToPlayer(passCand);
    return true;    
  }

  if (!shouldPassDirect && shouldPassIndirect){
    passSpeed = passVelocity.getLength();
    passPower = worldModel->calcKickPowerForSpeed( passSpeed );
    skillPool->kickToPoint( passPoint, passPower, false );
    return true;
  }

  if (shouldPassDirect && shouldPassIndirect){
    if (passPoint.getX() > passCand->getPosition().getX()){
      passSpeed = passVelocity.getLength();
      passPower = worldModel->calcKickPowerForSpeed( passSpeed );
    
      worldModel->passPointToSayCycles = 3;
      worldModel->passPointToSay = passPoint;
    
      skillPool->kickToPoint( passPoint, passPower, false );
      return true;
    }     
    skillPool->passBallToPlayer(passCand);
    return true;        
  }
  // Passing is no good option
  return false;
}

// Calculates maximum inertia of agent and how far he can use ball to
// and other is blah and so on...
float Behaviour::getDribbleAngle(){
  float speed = myself->getVelocity().getLength();
  speed *= Configuration::getPlayerDecay(myself->getType());
  float inertiaMoment = Configuration::getInertiaMoment(myself->getType());
  float maxAngle = 180.0 / ( 1.0 + inertiaMoment * speed );
  float angle = min(maxAngle, 35.0f);
  return angle;
}

bool Behaviour::considerDribbling(){
  
  float ballX    = worldModel->ball.getPositionPt()->getX();
  float ballY    = worldModel->ball.getPositionPt()->getY();
  float dribbleY = worldModel->myself.getPositionPt()->getY();

  static Vector2D dribblePoint;
  if (ballX < 38.0){
    // avoid sides
    if (fabs(dribbleY) > 31.0) dribbleY = 31.0 * sign(dribbleY);
    // dribble close to goal line
    dribblePoint = Vector2D(50.0, dribbleY);
  }else{
    // if close to goal line to get near the goal
    if (ballY < 0.0){
      dribblePoint = Vector2D(50.5, -10.0);
    }else{
      dribblePoint = Vector2D(50.5,  10.0);
    }
  }
  float dribbleAngle = worldModel->angleFromPointToPoint(myself->getPosition(), dribblePoint);

  // EMERGENCY DRIBBLING
  const Player* enemy = worldModel->closestInSetToPlayer(PS_Opponents, *myself, ConfigFile::highConfidencePlayer);
  if (enemy != NULL){
    bool isClose = enemy->isClose();
    // Enemy is close so we are attacked from behind... pervert...
    if (isClose){
      // dribble forward
      //static Vector2D dribblePoint;
      //dribblePoint = myself->getPosition() + Vector2D(10, 0, POLAR);
      int i = getDribbleDistance(dribbleAngle);
      if (i > 0){
	skillPool->simpleDribbleToPoint(dribblePoint, i);
	// we have done something (dribbling)
	return true;
      }
      // we dont find a good distance for simple dribbling. 
      // so we have to do some smart dribbling
    }else{     
      // We are attacked from the front so dribble sideways
      Vector2D relPos = enemy->getPosition() - myself->getPosition();
      float dist = relPos.getLengthSq();
      if (dist < 16.0 /* 4.0 */){
	float enemyAngle = worldModel->angleFromPointToPoint(myself->getPosition(), enemy->getPosition());
	enemyAngle = normalizeTo180Deg(dribbleAngle - enemyAngle);
	float angle = -getDribbleAngle() * sign(enemyAngle);
	
	angle += dribbleAngle;
	angle  = normalizeTo180Deg(angle);
	
	static Vector2D dribblePoint;
	dribblePoint = myself->getPosition() + Vector2D(10, angle, POLAR);
	int i = getDribbleDistance(angle);
	if (i > 0){
	  skillPool->simpleDribbleToPoint(dribblePoint, i);	
	  // we have done something
	  return true;
	}
      }
    }
  }

  // NORMAL DRIBBLING
  // No passing to be done so try to dribble
  if (shouldDribble()){
     if (worldModel->myself.getPositionPt()->distanceToSq(dribblePoint) > 1.0f){
      skillPool->smartDribbleToPoint(dribblePoint, 100, getDribbleSide());
      return true;
    }
  }

  // Not dribbling
  return false;
}




















