/******************************************************************************
 * SkillPool.cc
 * Copyright (c) 2001 Virtual Werder
 *
 *
 * implements:
 *   class SkillPool
 *
 * requires:
 *
 *
 * Created on: 26.10.2001 by wenns        
 * Maintainer: cniehaus
 * Group:
 * Co-authors: dbernau, harlekin, tbogon
 *
 * Changelog:
 *
 * $Id: SkillPool.cc,v 1.67 2004/06/11 12:24:09 mosh Exp $
 *
 ******************************************************************************/

#include "SkillPool.hh"
#include "Agent.hh"
#include "WorldModel.hh"
#include "Effectors.hh"
#include "debug.hh"
#include "Geometry.hh"
#include "ConfigFile.hh"
#include "Behaviour.hh"
#include "GoalieBehaviour.hh"
#include "WingAttackerBehaviour.hh"

#include <math.h>
#include <assert.h>
#include <string>


/*******************************************************************************/
SkillPool::SkillPool(){
  worldModel = Agent::getWorldModel();
  effectors  = Agent::getEffectors();
  myself     = &(worldModel->myself);
  ball       = &(worldModel->ball);
}


//------------------ face-Skills Start --------------------------------------------//


/*******************************************************************************/
/**
 * Turns the body of the agent in the direction of the given point
 * If the body can't be turned in one cycle because of the limitation from 
 * the agents speed, the body is turned as far as possible 
 * @param point to turn the body to
 * @return true, if no turning is needed (no command was send)... so if the agent
 *               body facing already the right direction         
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceOnlyBodyToPoint(const Vector2D &point){
  if(point.isInvalid()){
    DEBUG1("SkillPool::faceOnlyBodyToPoint(): the point is invalid !!!");
    return false;
  }

  static Vector2D relPos;
  relPos.setVector2D(point - myself->predictedPosition());

  float angle       = relPos.getDirection();
  float angleToTurn = normalizeTo180Deg(angle - myself->getBodyDir());
  float speed       = myself->getVelocity().getLength();

  if(fabs(angleToTurn) >= ConfigFile::angleWhenToTurn){
    effectors->queueTurn(angleToTurn, speed);
    return false;
  }
  return true;
}


/*******************************************************************************/
/**
 * Turns the back of the agent in the direction of the given point
 * If the body can't be turned in one cycle because of the limitation from 
 * the agents speed, the body is turned as far as possible 
 * @param point to turn the back of the agent to
 * @return true, if no turning is needed (no command was send)... so if the agent
 *               body facing already the right direction         
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceOnlyBackToPoint(const Vector2D &point){
  if(point.isInvalid()){
    DEBUG1("SkillPool::faceOnlyBodyToPoint(): the point is invalid !!!");
    return false;
  }

  static Vector2D relPos;
  relPos.setVector2D(point - myself->predictedPosition());
  float angle       = normalizeTo180Deg(relPos.getDirection() + 180);
  float angleToTurn = normalizeTo180Deg(angle - myself->getBodyDir());
  float speed       = myself->getVelocity().getLength();

  if(fabs(angleToTurn) >= ConfigFile::angleWhenToTurn){
    effectors->queueTurn(angleToTurn, speed);
    return false;
  }
  return true;
}


/*******************************************************************************/
/**
 * Turns only the neck of the agent in the direction of the given point.
 * If the full turn is not possible because of the turn-neck-limit,
 * the turn will be executed with maximal angle. The body direction
 * is not influenced.
 * @param point the point to turn the neck to
 * @return true, if no turning neck is needed (no command was send),
 *               so if the agent head facing already the right direction         
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceOnlyNeckToPoint(const Vector2D &point){
  if(point.isInvalid()){
    DEBUG1("SkillPool::faceOnlyNeckToPoint(): the point is invalid!");
    return false;
  }

  //current angles
  float bodyDir = myself->getBodyDir();
  float headDir = myself->getHeadDir();

  //check for possible turn in this cycle
  //if TURN, then add it on the current body- and headangle
  //else take the current values
  ServerCommandHistory &sch = worldModel->getSentCommands();
  if(sch.getFst() == CMD_TURN){
    bodyDir = normalizeTo180Deg(bodyDir + sch.getFstParam1());
    headDir = normalizeTo180Deg(headDir + sch.getFstParam1());
  }

  //calculate the desired angle of the neck relative to body
  float angleToPos = (point - myself->predictedPosition()).getDirection();
  float desTurnNeckAngleToBody = normalizeTo180Deg(angleToPos - bodyDir);
  float headDirRelToBody       = normalizeTo180Deg(headDir - bodyDir);
  float turnNeckAngle          = 0;

  //decision making : which turnAngle to take
  if(desTurnNeckAngleToBody < Configuration::minneckang){
    turnNeckAngle = Configuration::minneckang - headDirRelToBody;
  }
  else if(desTurnNeckAngleToBody > Configuration::maxneckang){
    turnNeckAngle = Configuration::maxneckang - headDirRelToBody;
  }
  else{
    turnNeckAngle = desTurnNeckAngleToBody - headDirRelToBody;
  }

  //sending the command and return false when turn neck needed
  //else no command sent and return true because already 
  //facing the right direction 
  if(fabs(turnNeckAngle) >= ConfigFile::angleWhenToTurnNeck){
    effectors->queueTurnNeck(turnNeckAngle);
    return false;
  }

  return true;
}


/*******************************************************************************/
/**
 * Turns the neck to a given point.
 * If the neck can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the body is turned as 
 * far as needed. ATTENTION: only use this skill, if no other skill sends
 * an TURN command in the current cycle. 
 * @return true, if the agent already facing the given point 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceNeckToPoint(const Vector2D &point){
  float angleToTurn = (point - myself->predictedPosition()).getDirection();
  float bodyDir = myself->getBodyDir(); 
  
  //angle between my bodyDir and the point
  float desTurnNeckAngleToBody = normalizeTo180Deg(angleToTurn - bodyDir);
  float turnAngle = 0;
  
  //if angle smaller then maxNeckTurnAngle simply turn only neck 
  if(desTurnNeckAngleToBody < Configuration::minneckang){
    turnAngle = desTurnNeckAngleToBody + myself->getViewAngle()/2;
    effectors->queueTurn(turnAngle);
    faceOnlyNeckToPoint(point);
    return false;
  }
  else if(desTurnNeckAngleToBody > Configuration::maxneckang){
    turnAngle = desTurnNeckAngleToBody - myself->getViewAngle()/2;
    effectors->queueTurn(turnAngle);
    faceOnlyNeckToPoint(point);
    return false;
  }
  else{
    return faceOnlyNeckToPoint(point);
  }
}


/*******************************************************************************/
/**
 * Turns the body and the neck to a given point.
 * If the body and the neck can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the body and the neck  
 * is turned as far as needed. 
 * @return true, if the agent already facing the given point 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceNeckAndBodyToPoint(const Vector2D &point){
  bool fobtp = faceOnlyBodyToPoint(point);
  bool fontp = faceOnlyNeckToPoint(point);
  return (fobtp && fontp);
}


/*******************************************************************************/
/**
 * Turns only the body to the position of the ball.
 * If the body can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the body is turned as 
 * far as possible. 
 * @return true, if the agent already facing the given point 
 *         false, otherwise... 
 */
/*******************************************************************************/
bool SkillPool::faceOnlyBodyToBall(){
  if(ball->predictedPosition().distanceTo(ball->getPosition()) <= 
     Configuration::ballSpeedMax)
    return faceOnlyBodyToPoint(ball->predictedPosition());
  else
    return faceOnlyBodyToPoint(ball->getPosition());
}


/*******************************************************************************/
/**
 * Turns only the neck to the position of the ball.
 * If the neck can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the neck is turned as 
 * far as possible. 
 * @return true, if the agent already facing the given point 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceOnlyNeckToBall(){
  if(ball->predictedPosition().distanceTo(ball->getPosition()) <= 
     Configuration::ballSpeedMax)
    return faceOnlyNeckToPoint(ball->predictedPosition());
  else
    return faceOnlyNeckToPoint(ball->getPosition());
}


/*******************************************************************************/
/**
 * Turns the neck to the position of the ball.
 * If the neck can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the body is turned as 
 * far as needed.  
 * @return true, if the agent already facing the given point 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceNeckToBall(){
  static Vector2D predictedBallPosition;
  predictedBallPosition.setVector2D(ball->predictedPosition());
  if (predictedBallPosition.distanceTo(ball->getPosition()) <= Configuration::ballSpeedMax)
    return faceNeckToPoint(predictedBallPosition);
  else
    return faceNeckToPoint(ball->getPosition());
}


/*******************************************************************************/
/**
 * Turns the body and the neck to the ball.
 * If the body and the neck can't be turned in one cycle as much as needed, 
 * because of the limitation from the agents speed, the body and the neck 
 * is turned as far as needed. 
 * @return true, if the agent already facing the given point 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::faceNeckAndBodyToBall(){
  static Vector2D predictedBallPosition;
  predictedBallPosition.setVector2D(ball->predictedPosition());
  if (predictedBallPosition.distanceTo(ball->getPosition()) <= Configuration::ballSpeedMax)
    return faceNeckAndBodyToPoint(predictedBallPosition);
  else
    return faceNeckAndBodyToPoint(ball->getPosition());
}


//------------------ face-Skills Ende --------------------------------------------//


//------------------ goTo-Skills Start -------------------------------------------//


/*********************************************************************************/
/**
 * The skill getBall tries to intercept the ball. Therefore it uses the lower level 
 * skill goToPoint with a tolerance useful for intercepting. In order to intercept
 * the ball at a right distance (kickableMargin/4) from the agent, it uses the skills 
 * interceptClose  and interceptCloseGoalie for the last two steps to the ball. 
 * @param power dashPower for Interception
 * @param forCatch true, if goalie intercepts the ball
 * @param closeInterception true, for using the close interception skill
 * @return true, if the agent got the ball. 
 *         false, otherwise...
 */
/*********************************************************************************/
bool SkillPool::getBall(float power, bool forCatch, bool closeInterception){
  //set the tolerance for using goToPoint()
  float tolerance;
  if( forCatch ) 
    tolerance = Configuration::catchableAreaL - ConfigFile::toleranceBuffer;
  else           
    tolerance = Configuration::getPlayerSize(myself->getType()) +
      Configuration::ballSize + 
      Configuration::getKickableMargin(myself->getType()) - 
      ConfigFile::toleranceBuffer;
  
  const Vector2D *ballPos = ball->getPositionPt();
  if ( worldModel->distanceToPoint(ballPos) < tolerance)
    return true;
  
  // Those methods return true if they queue a command so this
  // method has to return false which means it has not reached
  // the wished point
  if(closeInterception){
    //interceptClose for the correct last two steps to the ball, for fieldplayer only
    if(!forCatch){
      if(interceptClose()){
	return false; 
      }
    }else{
      if(interceptCloseGoalie()){
        return false; 
      }
    }
  }

  //look for a possible interception point or else for the ball position.
  static Vector2D intPoint;
  int cycles;
  if(worldModel->interceptionPoint(*myself, intPoint, cycles, forCatch)){
    goToPoint(intPoint, tolerance, power);
    return false;
  }else{
    return goToPoint(ballPos, tolerance, power);
  }
}


/*******************************************************************************/
/**
 * The skill interceptClose creates the last two steps to the ball. Therefore it 
 * checks all possible combinations of dash and turn commands. The goal is to bring
 * the agent within a good distance to the ball (kickableMargin/4). It checks 
 * the single dash, the combination of first turn when dash and the combination
 * of two dashes.
 * @return true, if the agent can intercept within two cycles. 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::interceptClose(){
  //necessary information
  int   playerType     = myself->getType();
  float playerSize     = Configuration::getPlayerSize(playerType);
  float kickableMargin = Configuration::getKickableMargin(playerType);
  float dashPowerRate  = Configuration::getDashPowerRate(playerType);
  float playerDecay    = Configuration::getPlayerDecay(playerType);
  float maxPlayerSpeed = Configuration::getMaximumPlayerSpeed(playerType);

  //max distances
  float maxKickDist       = playerSize + Configuration::ballSize + kickableMargin;
  float maxKickDistSecure = playerSize + Configuration::ballSize + kickableMargin * 0.8;
  float maxReachDist      = 2 * maxPlayerSpeed + (1.0 + Configuration::ballDecay) * 
                            Configuration::ballSpeedMax + maxKickDist; 
  float perfectBallDist   = playerSize + Configuration::ballSize + kickableMargin/4;
  
  //my data
  const Vector2D &myPos  = myself->getPosition();
  const Vector2D &myVel  = myself->getVelocity();
  float mySpeed   = myVel.getLength();
  float myBodyDir = myself->getBodyDir();

  //ball data
  Vector2D ballPosIn1Cycle  = ball->predictedPosition(1);
  Vector2D ballPosIn2Cycles = ball->predictedPosition(2);

  float dashPower, neededAngle;

  //STEP1 : check for max distance, can ball be reached in 2 cycles
  if(worldModel->distanceToBall() > maxReachDist){  
    return false;
  }

  //STEP2 : intercept in one cycle with a single dash
  Line myPath = Line::makeLineFromPositionAndAngle(myPos, myBodyDir);
  Circle ballCircle = Circle(ballPosIn1Cycle, perfectBallDist);
  static Vector2D InterSecPoint_1, InterSecPoint_2, goToPoint;
  int intersectionPoints = myPath.getCircleIntersectionPoints(ballCircle, &InterSecPoint_1, &InterSecPoint_2);
  if(intersectionPoints > 0){
    //selection of the best intersection point
    if(intersectionPoints == 1){
      goToPoint = InterSecPoint_1;
    }
    if(intersectionPoints == 2){
      if(worldModel->distanceToPoint(InterSecPoint_1) <= 
	 worldModel->distanceToPoint(InterSecPoint_2))
	goToPoint = InterSecPoint_1;
      else
	goToPoint = InterSecPoint_2;
    }
    dashPower = myself->getPowerForDash(goToPoint - myPos, myBodyDir, myVel, myself->getEffort());
    dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    Vector2D myPosIn1CycleDash = myself->predictedPosition(1, dashPower, 0);
    //checking for distance
    if(myPosIn1CycleDash.distanceTo(ballPosIn1Cycle) < maxKickDistSecure){
	effectors->queueDash(dashPower);
	return true;
    }
  }
  
  //STEP3 : intercept in two cycles with a turn followed by a dash
  float angleToBallIn2Cycles = normalizeTo180Deg((ballPosIn2Cycles - 
						  myself->predictedPosition()).getDirection());
  neededAngle = normalizeTo180Deg(angleToBallIn2Cycles - myBodyDir); 
  neededAngle = worldModel->calcAngleForTurn(angleToBallIn2Cycles, mySpeed);  
  //first check angle to turn 
  if((fabs(neededAngle) < Configuration::maxmoment) && (fabs(neededAngle) > ConfigFile::angleWhenToTurn)){
    //better prediction
    Vector2D myPosIn1CycleTurn = myPos + myVel;
    //Vector2D myPosIn1CycleTurn = myself->predictedPosition(1, 0, angleToBallIn2Cycles);
    Vector2D velIn1CycleTurn = (myVel * playerDecay).rotate(angleToBallIn2Cycles);
    float myBodyDirIn1Cycle = velIn1CycleTurn.getDirection();
    Vector2D relWay2PerfectPos = ballPosIn2Cycles - myPosIn1CycleTurn - 
	                         Vector2D(perfectBallDist, myBodyDirIn1Cycle, POLAR);
    dashPower = myself->getPowerForDash(relWay2PerfectPos, myBodyDirIn1Cycle, velIn1CycleTurn, myself->getEffort());
    dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    //calculation of the predicted position after turn and dash
    Vector2D myPosIn2CyclesTurnDash;
    if(dashPower > 0){
      myPosIn2CyclesTurnDash = myPosIn1CycleTurn + velIn1CycleTurn +
	  Vector2D(dashPower * dashPowerRate, myBodyDirIn1Cycle, POLAR);
    }else{
      myPosIn2CyclesTurnDash = myPosIn1CycleTurn + velIn1CycleTurn +
	  Vector2D(fabs(dashPower/2 * dashPowerRate), normalizeTo180Deg(myBodyDirIn1Cycle+180), POLAR);
    }
    //checking for distance
    if(myPosIn2CyclesTurnDash.distanceTo(ballPosIn2Cycles) < maxKickDistSecure){
      effectors->queueTurn(angleToBallIn2Cycles, mySpeed);
      return true;
    }
  }

  //STEP4 : intercept in two cycles with two dashes
  Vector2D myPosIn1CycleFullDash = myself->predictedPosition(1, 100, 0);
  myPath = Line::makeLineFromPositionAndAngle(myPos, myBodyDir);
  ballCircle = Circle(ballPosIn2Cycles, perfectBallDist);
  intersectionPoints = myPath.getCircleIntersectionPoints(ballCircle, &InterSecPoint_1, &InterSecPoint_2);
  //checking for intersections between ballCircle and myPath
  if(intersectionPoints > 0){
    if(intersectionPoints == 1){
      goToPoint = InterSecPoint_1;
    }
    if(intersectionPoints == 2){
      if(worldModel->distanceToPoint(InterSecPoint_1) <= worldModel->distanceToPoint(InterSecPoint_2))
	goToPoint = InterSecPoint_1;
      else
	goToPoint = InterSecPoint_2;
    }
    dashPower = myself->getPowerForDash(goToPoint - myPos, myBodyDir, myVel, myself->getEffort());
    dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    float firstDashPower = dashPower;
    if(dashPower < 0)
	myPosIn1CycleFullDash = myself->predictedPosition(1, -100, 0);
    Vector2D velIn1CycleDash = (myPosIn1CycleFullDash - myPos) * playerDecay;
    dashPower = myself->getPowerForDash(goToPoint - myPosIn1CycleFullDash, myBodyDir, 
					velIn1CycleDash, myself->getEffort());
    dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    //calculation of the predicted position after dash and dash
    Vector2D myPosIn2CyclesDashDash;
    if(dashPower > 0){
	myPosIn2CyclesDashDash = myPosIn1CycleFullDash + velIn1CycleDash +
	    Vector2D(dashPower * dashPowerRate, myBodyDir, POLAR);
    }else{
	myPosIn2CyclesDashDash = myPosIn1CycleFullDash + velIn1CycleDash +
	    Vector2D(fabs(dashPower/2 * dashPowerRate), normalizeTo180Deg(myBodyDir+180), POLAR);
    }
    //checking for distance 
    if(myPosIn2CyclesDashDash.distanceTo(ballPosIn2Cycles) < maxKickDistSecure){
      effectors->queueDash(sign(firstDashPower) * 100);
      return true;
    }
  }
  return false;
}


/*******************************************************************************/
/**
 * The skill interceptCloseGoalie creates the last two steps to the ball, if a 
 * goalie wants to intercept the ball. Therefore it checks all possible combinations 
 * of dash and turn commands. The goal is to bring the goalie within the 
 * catchable distance to the ball (Configuration::catchableAreaL). It checks  
 * the single dash, the combination of first turn when dash and the combination
 * of two dashes.
 * @return true, if the goalie can catch the ball within two cycles. 
 *         false, otherwise...
 */
/*******************************************************************************/
bool SkillPool::interceptCloseGoalie(){
  //necessary information about the player values
  int   playerType     = myself->getType();
  float dashPowerRate  = Configuration::getDashPowerRate(playerType);
  float playerDecay    = Configuration::getPlayerDecay(playerType);
  float maxPlayerSpeed = Configuration::getMaximumPlayerSpeed(playerType);
  float catchableAreaSq= Configuration::catchableAreaL * Configuration::catchableAreaL;

  float maxDist, dashPower;
  Vector2D closestPosToBall, predPos;

  //myself data
  const Vector2D &myPos  = myself->getPosition();
  const Vector2D &myVel  = myself->getVelocity();
  float mySpeed   = myVel.getLength();
  float myBodyDir = myself->getBodyDir();
  Line myPath     = Line::makeLineFromPositionAndAngle(myPos, myBodyDir);
    
  //ball data
  float ballDist = worldModel->distanceToBall();
  Vector2D ballPosIn1 = ball->predictedPosition(1);
  Vector2D ballPosIn2 = ball->predictedPosition(2);
    
  //step 1 : try a single dash
  maxDist = maxPlayerSpeed + Configuration::ballSpeedMax + Configuration::catchableAreaL;
  if(ballDist < maxDist){
    closestPosToBall = myPath.getPointOnLineClosestTo(ballPosIn1);
    dashPower = myself->getPowerForDash(closestPosToBall - myPos, myBodyDir, myVel, myself->getEffort());
    // This will be done in predictedPosition
    //dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    predPos   = myself->predictedPosition(1, dashPower, 0);
    if(predPos.distanceToSq(ballPosIn1) < catchableAreaSq){
      effectors->queueDash(dashPower);
      return true;
    }
  }

  // Distance we (and the ball) can travel at best case
  maxDist = 2 * maxPlayerSpeed + (1.0 + Configuration::ballDecay) * 
    Configuration::ballSpeedMax + Configuration::catchableAreaL;
  if(ballDist < maxDist){
    //step 2 : try two dashes
    dashPower = myself->getPowerForDash(ballPosIn2 - myPos, myBodyDir, myVel, myself->getEffort());
    // This will be done in predictedPosition
    //dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    predPos   = myself->predictedPosition(1, dashPower, 0);
    float firstDashPower = dashPower;
    Vector2D velIn1 = (predPos - myPos) * playerDecay;
    dashPower = myself->getPowerForDash(ballPosIn2 - predPos, myBodyDir, velIn1, myself->getEffort());
    dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
    //calculation of the predicted position after dash two
    if(dashPower > 0){
      predPos = predPos + velIn1 + 
	        Vector2D(dashPower * dashPowerRate, myBodyDir, POLAR);
    }else{
      predPos = predPos + velIn1 + 
	        Vector2D(fabs(dashPower/2 * dashPowerRate), normalizeTo180Deg(myBodyDir+180), POLAR);
    }
    if(predPos.distanceToSq(ballPosIn2) < catchableAreaSq){
      effectors->queueDash(firstDashPower);
      return true;
    }
	
    //step 3 : try one turn followed by one dash
    float angleToBallIn2 = normalizeTo180Deg((ballPosIn2 - myself->predictedPosition()).getDirection());
    float neededAngle = normalizeTo180Deg(angleToBallIn2 - myBodyDir);
    neededAngle = worldModel->calcAngleForTurn(angleToBallIn2, mySpeed);  
    //first check angle to turn too high too low
    if((fabs(neededAngle) < Configuration::maxmoment) && 
       (fabs(neededAngle) > ConfigFile::angleWhenToTurn)){
      //better prediction
      Vector2D myPosIn1 = myPos + myVel;
      //Vector2D myPosIn1 = myself->predictedPosition(1, 0, angleToBallIn2);
      Vector2D velIn1Cycle = (myVel * playerDecay).rotate(angleToBallIn2);
      float myBodyDirIn1 = velIn1Cycle.getDirection();
      Vector2D relWay2Pos = ballPosIn2 - myPosIn1;
      dashPower = myself->getPowerForDash(relWay2Pos, myBodyDirIn1, velIn1Cycle, myself->getEffort());
      dashPower = myself->correctDashPowerForStamina(dashPower, myself->bodyInfo.getStamina());
      if(dashPower > 0){
      	predPos = predPos + velIn1Cycle + 
	          Vector2D(dashPower * dashPowerRate, myBodyDirIn1, POLAR);
      }else{
	predPos = predPos + velIn1Cycle + 
	          Vector2D(fabs(dashPower/2 * dashPowerRate), normalizeTo180Deg(myBodyDirIn1+180), POLAR);
      }
      if(predPos.distanceToSq(ballPosIn2) < catchableAreaSq){
        effectors->queueTurn(angleToBallIn2, mySpeed);
        return true;
      }
    }
  }
  return false;
}


/*******************************************************************************/
/**
 * Lets the agent go to the specified point. The point is reached, if the
 * distance to them is smaller than the specified tolerance. 
 * @param target the point to go to
 * @param tolerance the aforesaid tolerance
 * @param dashPower the dash power to use
 * @return true, if no soccer server command is queued (job done) and the agent 
 *         is in the right place
 *         false, otherwise
 */
/*******************************************************************************/
bool SkillPool::goToPoint(const Vector2D &target, float tolerance, float dashPow){
  static float angleWhenToTurn = ConfigFile::angleWhenToTurn;

  //target condition  
  float distance = worldModel->distanceToPoint(target);
  if(distance < tolerance)
    return true;
  
  static Vector2D predPos;
  predPos.setVector2D(myself->predictedPosition());
  float myBodyDir  = myself->getBodyDir();
  
  //running backwards
  /*  if(dashPow < 0){
    myBodyDir = normalizeTo180Deg(myBodyDir + 180);
  }
  */

  bool backwards  = dashPow < 0;
  bool shouldTurn = false;

  if (backwards){
    myBodyDir = normalizeTo180Deg(myBodyDir + 180.0f);
  }
  
  // Angle to target
  float turnAng  = normalizeTo180Deg((target-predPos).getDirection() - myBodyDir);  
  const Vector2D &myVel = myself->getVelocity();
  float mySpeed  = myVel.getLength();
  Line myPath    = Line::makeLineFromPositionAndAngle(myself->getPosition(), myBodyDir);
  static Vector2D closestPos;
  closestPos.setVector2D(myPath.getPointOnLineClosestTo(target));
  // Distance to target when I am next to it without turning.
  float distanceToPath = closestPos.distanceTo(target);
  

  // TURN NEEDED?
  // A small tolerance does not work with the new method
  if(tolerance < 0.5){
    if(fabs(turnAng) > angleWhenToTurn){
      effectors->queueTurn(turnAng, mySpeed);
      return false;
    }
  }else{
    // (turnAng!=0) is a protecting workaround because distanceToPath becomes way to high (>0)
    // if the line crosses the point (turnAng==0).
    if ( turnAng && distanceToPath > tolerance ){
      shouldTurn = true;
    }else if ((distance < 5 && /*mySpeed > 0.1 &&*/ fabs(turnAng) >= 170/*170*/)){
      shouldTurn = false;
      backwards  = true;
    }else if ((fabs(turnAng) >  80)){
      shouldTurn = true;
    }

    if( shouldTurn ){
      effectors->queueTurn(turnAng, mySpeed);
      return false;
    }
  }

  //Don't dash harder than needed
  // TODO: Does this work? Should be checked
  float maxSpeed = myself->getPowerForDash(closestPos - myself->getPosition(), myBodyDir, 
					   myVel, myself->getEffort());
  //Don't dash too hard (stamina buffer)
  if (backwards) dashPow *= -1;
  float minSpeed = myself->correctDashPowerForStamina(dashPow, myself->bodyInfo.getStamina());
  if (!backwards)
    dashPow = min(maxSpeed, minSpeed);
  else
    dashPow = minSpeed;

  if (backwards){
    // Make sure we are really running backwards
    dashPow = -1 * fabs(dashPow);
    DEBUG2_HARLEKIN("Backwards with: ", dashPow);
  }

  //if dash needed !?  
  if(fabs(dashPow) > 0){
    effectors->queueDash(dashPow);
    return false;
  }

  // TODO: What now?
  // We have not arrived and still nothing is done. 
  // This could happen due to stamina problems. 
  // Should we return 'true' so the player can do something else?
  // Would it be better just to run slower before this happens?
  //DEBUG1_HARLEKIN("SkillPool::goToPoint(...), well, I am doing nothing. Lalala...");
  return false;
}


/*******************************************************************************/
/**
 * Lets the agent go to a position assigned to him in the case of our
 * corner kick
 * KICK IT OUT OF THE SKILLPOOL AND PUT IT INTO A CONFIGURATIONFILE
 */
/*******************************************************************************/
bool SkillPool::goToPositionOurCKick(){
  // use static vectors to safe time. dont care
  // about memory usage
  static const Vector2D NO_2_POS_OC  = Vector2D( 30, -15);
  static const Vector2D NO_3_POS_OC  = Vector2D( 30, -6 );
  static const Vector2D NO_4_POS_OC  = Vector2D( 30,  6 );
  static const Vector2D NO_5_POS_OC  = Vector2D( 30,  15);
  static const Vector2D NO_6_POS_OC  = Vector2D( 42, -20);
  static const Vector2D NO_7_POS_OC  = Vector2D( 42,  0 );
  static const Vector2D NO_8_POS_OC  = Vector2D( 42,  20);
  static const Vector2D NO_9_POS_OC  = Vector2D( 45,  -9);
  static const Vector2D NO_10_POS_OC = Vector2D( 45,   0);
  static const Vector2D NO_11_POS_OC = Vector2D( 45,   9);

#define DASH_POW_OC     60
#define TOLERANCE_OC    1

  int uNum = myself->getUniformNumber();
  switch(uNum){
    case 2:  { return goToPoint(NO_2_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 3:  { return goToPoint(NO_3_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 4:  { return goToPoint(NO_4_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 5:  { return goToPoint(NO_5_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 6:  { return goToPoint(NO_6_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 7:  { return goToPoint(NO_7_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 8:  { return goToPoint(NO_8_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 9:  { return goToPoint(NO_9_POS_OC,  TOLERANCE_OC, DASH_POW_OC); }
    case 10: { return goToPoint(NO_10_POS_OC, TOLERANCE_OC, DASH_POW_OC); }
    case 11: { return goToPoint(NO_11_POS_OC, TOLERANCE_OC, DASH_POW_OC); }
    default: {
      DEBUG1("SkillPool::moveToOurCKickPosition(): NO UNIFORMNUMBER!");
      return false;
    }
  }
}


/*******************************************************************************/
/**
 * Lets the agent go to a position assigned to him in the case of a opponent
 * corner kick
 * KICK IT OUT THE SKILLPOOL AND PUT IT INTO A CONFIGURATIONFILE
 */
/*******************************************************************************/
bool SkillPool::goToPositionTheirCKick(){
  // Use static vectors instead of initialising 
  // one each cycle. Safes some time.
  static const Vector2D NO_2_POS_TC  = Vector2D( -45, -15 );
  static const Vector2D NO_3_POS_TC  = Vector2D( -51, - 6 );
  static const Vector2D NO_4_POS_TC  = Vector2D( -51,   6 );
  static const Vector2D NO_5_POS_TC  = Vector2D( -45,  15 );
  static const Vector2D NO_6_POS_TC  = Vector2D( -45, -25 );
  static const Vector2D NO_7_POS_TC  = Vector2D( -40,   0 );
  static const Vector2D NO_8_POS_TC  = Vector2D( -45,  25 );
  static const Vector2D NO_9_POS_TC  = Vector2D( -35, -20 );
  static const Vector2D NO_10_POS_TC = Vector2D( -35,   0 );
  static const Vector2D NO_11_POS_TC = Vector2D( -35,  20 );

#define DASH_POW_TC     60
#define TOLERANCE_TC    1

  int uNum = myself->getUniformNumber();
  switch(uNum){
    case 2:  { return goToPoint(NO_2_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 3:  { return goToPoint(NO_3_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 4:  { return goToPoint(NO_4_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 5:  { return goToPoint(NO_5_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 6:  { return goToPoint(NO_6_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 7:  { return goToPoint(NO_7_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 8:  { return goToPoint(NO_8_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 9:  { return goToPoint(NO_9_POS_TC,  TOLERANCE_TC, DASH_POW_TC); }
    case 10: { return goToPoint(NO_10_POS_TC, TOLERANCE_TC, DASH_POW_TC); }
    case 11: { return goToPoint(NO_11_POS_TC, TOLERANCE_TC, DASH_POW_TC); }
    default: {
      DEBUG1("SkillPool::moveToTheirCKickPosition(): NO UNIFORMNUMBER!");
      return false;
    }
  }
}


//------------------ goTo-Skills Ende --------------------------------------------//


//------------------ kickTo-Skills Start -----------------------------------------//


/**********************************************************************************/
/**
 * This skill accelerate the ball to a wished velocity, but it only does so,if its
 * possible, else the ball is shot smaller, if variable kickSmallerToo is set. 
 * The skill also prevent the ball from colliding with the agent by shooting
 * the ball to a save distance from the body.
 * @param wishedVelocity the desired velocity
 * @param kickSmallerToo, if one shot cant get the ball to the wished 
 *        velocity, shot smaller but in the correct direction
 * @return true, if velocity can be reached
 *         false, otherwise ...
 */
/**********************************************************************************/
bool SkillPool::accelerateBallToVelocity(const Vector2D &wishedVel, bool kickSmallerToo ){
  static float power, angle;
  static Vector2D lastVel;
  // Abort this function as this would result in a deadlock
  // remember to reset this values before "returning"
  if (!lastVel.isInvalid() && wishedVel == lastVel){
    DEBUG1_HARLEKIN("...leaving cause of infinite recursion");
    lastVel.setVector2D(999.333, 999.333);
    return false;
  }

  float myBodyDir      = myself->getBodyDir();
  const Vector2D &ballVel = ball->getVelocity();
  static Vector2D wishedAccel;
  wishedAccel = wishedVel - ballVel;

  if (!ball->isVelHighConfident(worldModel->getTime())){
    // DEBUG2_HARLEKIN("  Ball-Velocity is not high confident: ", ballVel);
    if (ballVel.isInvalid()){
      wishedAccel = wishedVel;
      // DEBUG1_HARLEKIN("  Use Velocity of (0,0), just to be sure.");
    }else{
      // TODO: What should be done in this case?
    }
  }
 
  //if power under max, correct acceleration is possible
  static float ballSpeedMaxSq = Configuration::ballSpeedMax * Configuration::ballSpeedMax;
  if( wishedAccel.getLengthSq() < ballSpeedMaxSq ){
    power = worldModel->calcKickPowerForSpeed( wishedAccel.getLength() );
    
    angle = normalizeTo180Deg( wishedAccel.getDirection() - myBodyDir );
    if( power <= Configuration::maxpower ){
      effectors->queueKick( power, angle );
      lastVel.setVector2D(999.333,999.333);
      return true;
    }
  }

  if( kickSmallerToo ){
    //for power higher than max, the shooting vector is shorter but
    //in the same direction (with max power) 
    power           = Configuration::maxpower;
    float ballDist  = worldModel->distanceToBall();
    float ballAngle = worldModel->angleToBallRel2Body();
    float speed     = power * Configuration::kickPowerRate *
	              ball->calculateKickRate( ballDist, ballAngle );
    float tmp       = ballVel.rotated( -(wishedVel.getDirection()) ).getY();
    angle           = wishedVel.getDirection() - asinDeg( tmp / speed );
    angle           = normalizeTo180Deg( angle - myBodyDir );
    
    //check for collision after kick
    int playerType = myself->getType();
    const Vector2D &ballPos   = ball->getPosition();
    Vector2D ballPosAfterKick = ballPos + ballVel +
	                        Vector2D( speed, normalizeTo180Deg( angle + myBodyDir ), POLAR );
    float collDist = Configuration::getPlayerSize( playerType ) + Configuration::ballSize;
    float bestDist = collDist + Configuration::getKickableMargin( playerType )/4;
    if( myself->predictedPosition().distanceTo( ballPosAfterKick ) <= collDist ){
      //create new secure kickPoint 
      static Line ballWay = Line(0,0,0);
      ballWay.updateLineFromTwoPoints( ballPos, ballPosAfterKick);
      Circle goodDist = Circle( myself->predictedPosition(), bestDist );
      Vector2D InterSecPoint_1, InterSecPoint_2, kickPoint;
      int intersectionPoints = ballWay.getCircleIntersectionPoints(goodDist, &InterSecPoint_1, &InterSecPoint_2);
      if( intersectionPoints > 0 ){
        //selection of the best intersection point
	if( intersectionPoints == 1 ){
	  kickPoint = InterSecPoint_1;
	}else if( intersectionPoints == 2 ){
	  if( worldModel->distanceFromPointToBall(InterSecPoint_1) <= 
	      worldModel->distanceFromPointToBall(InterSecPoint_2) )
	    kickPoint = InterSecPoint_1;
	  else
	    kickPoint = InterSecPoint_2;
	}
	Vector2D antiCollVel = kickPoint - ( ballPos );
	//kick to new kickPoint - recursive!
	// save actual parameter to avoid infinite loops
	lastVel = wishedVel;
	return accelerateBallToVelocity( antiCollVel, true );
      }
      // would collide with ball and cant see another solution
      lastVel.setVector2D(999.333,999.333);
      return false;
    }
    // will not collide: send command
    effectors->queueKick( power, angle );
    lastVel.setVector2D(999.333,999.333);
    return true;
  }
  lastVel.setVector2D(999.333,999.333);
  return false;
}


/**********************************************************************************/
/**
 * This skill mainly uses the low level skill accelerateBallToVelocity and adds
 * the functionality of putting the ball in right place within the agent
 * kickArea in order to get the right velocity in the following cycle, if the
 * needed velocity cant be reached. The variable oneKick is needed to execute kickIns 
 * and freeKicks.
 * @param kickPoint target point
 * @param pow the kick power
 * @param oneKick one kick only, needed for kickIns
 * @return true if the shot can be execute
 *         false otherwise ...
 */
/**********************************************************************************/
bool SkillPool::kickToPoint(const Vector2D &kickPos, float power, bool oneKick)  {

  //wished velocity calculation
  float targetAngle = worldModel->angleFromPointToPoint( ball->getPosition(), kickPos );
  Vector2D wishedVel = Vector2D( power * Configuration::kickPowerRate, targetAngle, POLAR );

  //subbehaviour kickToPoint
  if( oneKick ){
    return accelerateBallToVelocity( wishedVel, true );
  }else{
    if( accelerateBallToVelocity( wishedVel, false ) ){
      return true;
    }else{
	int myType = myself->getType();
	if( fabs( worldModel->angleToBallRel2Body() ) > 10){
	  return kickBallCloseToBody( 0, Configuration::getKickableMargin( myType )/4 );
	}else{
	  //simple - just kick the ball in front
	  return accelerateBallToVelocity( wishedVel, true );
       }
    }
    return false;
  }
}


/*********************************************************************************/
/**
 * kicks the ball to a positon in the agents kickableArea on the 
 * desired distance to the body and the desired angle according 
 * to the bodydirection.
 * the default value for distanceToBody should be
 * Configuration::kickableMargin/4, now it is set to 0.175.
 * @param angleToBody desired angle according the bodydirection
 * @param distanceToBody distance to the body, a value between 0 and kickableMargin
 * @return true if the shot can be execute (command queued)
 *         false otherwise ...
 */
/*********************************************************************************/
bool SkillPool::kickBallCloseToBody(float angleToBody, float distanceToBody){
  //DEBUG5_HARLEKIN("kickBallCloseToBody(", angleToBody, ",", distanceToBody, ")");
  //predicted position myself
  Vector2D predPosMyself = myself->predictedPosition();
  //current ball position
  const Vector2D &ballPos = ball->getPosition();
  
  //desired distance from ball to body 
  float dDist = Configuration::getPlayerSize( myself->getType() ) +
    Configuration::ballSize + distanceToBody;
  //angle (angleToBody) relative to body for closekick
  float globalAngle = normalizeTo180Deg( angleToBody + myself->getBodyDir() );
  
  //desired ballVector and kickRate
  Vector2D vecDesired = predPosMyself + Vector2D( dDist, globalAngle, POLAR ) - ballPos;
  return accelerateBallToVelocity( vecDesired, true );
}


/*******************************************************************************/
/**
 * allows the agent to turn with the ball to a desired global angle. 
 * this happens in three step, if necessary:
 * 1. kick the ball close to body in the right direction (use of min-threshold
 *    for ball angle to direction when to turn)
 * 2. stop the ball (use of min-threshold for ball velocity when to stopBall)
 * 3. face only body to the desired position, which is simple  
 *    the predicted ball position.
 * @param globalAngle angle to turn with ball to
 * @return true if done, no server command is queued
 */
/*******************************************************************************/
bool SkillPool::turnWithBallTo(float globalAngle){
  //ballAngle min-threshold when to turn
  static float turnWithBallAngle = ConfigFile::turnWithBallAngle;
  //ballVelocity min-treshold when to stopBall
  static float turnWithBallSpeedSq = ConfigFile::turnWithBallStopSpeed * 
    ConfigFile::turnWithBallStopSpeed;
  
  //globalAngle from player to the ball
  float myBodyDir = myself->getBodyDir();
  float ballAngle = normalizeTo180Deg( myBodyDir + worldModel->angleToBallRel2Body() );

  //decision-tree for turnWithBallTo
  if( fabs(normalizeTo180Deg( globalAngle - ballAngle ) ) > turnWithBallAngle ){
    float directionToKick = normalizeTo180Deg( globalAngle - myBodyDir );
    return !kickBallCloseToBody( directionToKick, ConfigFile::turnWithBallKickDist );
  }else if( ball->getVelocity().getLengthSq() > turnWithBallSpeedSq ||
	    !worldModel->willBallBeKickable() ){
    return !stopBall();
  }else{
    return faceOnlyBodyToBall(); 
  }
}


/**********************************************************************************/
/**
 * this skill allows the agent to hold the ball away from the nearest 
 * opponent to the ball, but if there is no nearest opponent or it is too far away, 
 * the agent simply holds the ball in front of the body - the distance to hold 
 * the ball away is written down in the ConfigFile
 */
/**********************************************************************************/
void SkillPool::holdBall(){
  const Player *closestOpp = worldModel->closestInSetToBall( PS_Opponents );
  if( closestOpp != NULL && 
      worldModel->distanceToPoint( closestOpp->getPosition() ) <= 10.0 ){
    float angleToOpp = worldModel->angleToPointRel2Body( closestOpp->getPosition() );
    float angleAwayFromOpp = normalizeTo180Deg( angleToOpp + 180 );
    kickBallCloseToBody( angleAwayFromOpp, ConfigFile::holdDist );
  }else{
    //simply holds the ball in front of the body
    kickBallCloseToBody( 0.0, ConfigFile::holdDist );
  }
}

// Not really testet. DONT USE IT!!!
void SkillPool::holdBall_H(){
  holdBall();
  return;
  
  ////////////////////////////////
  const Player *closestOpp = worldModel->closestInSetToBall( PS_Opponents );
  if ( closestOpp != NULL ){
    if ( closestOpp->getConfidence(worldModel->getTime()) >= 0.98 ){
      float range = Configuration::getKickableMargin(closestOpp->getType()) +
	Configuration::getPlayerSize(closestOpp->getType()) + 
	Configuration::ballSize;
      float myRange = Configuration::getKickableMargin(myself->getType()) +
	Configuration::getPlayerSize(myself->getType()) +
	Configuration::ballSize;
      float angle1 = normalizeTo180Deg(closestOpp->getBodyDir() + 90);
      float angle2 = normalizeTo180Deg(closestOpp->getBodyDir() - 90);
      Vector2D point1 = closestOpp->getPosition() + Vector2D(range, angle1);
      Vector2D point2 = closestOpp->getPosition() + Vector2D(range, angle2);
      Line nme1 = Line::makeLineFromPositionAndAngle(point1, 
						     closestOpp->getBodyDir());
      Line nme2 = Line::makeLineFromPositionAndAngle(point2, 
						     closestOpp->getBodyDir());
      Circle circle = Circle(myself->getPosition(), myRange);
      Vector2D ip11, ip12, ip21, ip22;
      int num1 = nme1.getCircleIntersectionPoints(circle, &ip11, &ip12);
      int num2 = nme2.getCircleIntersectionPoints(circle, &ip21, &ip22);
      if (num1 > 1){
	Vector2D c = ip11 + ((ip12 - ip11)/2);
	c = c - myself->getPosition();
	float angle = c.getDirection();
	float dist = c.distanceTo(myself->getPosition());
	dist = (myRange - dist)/2 + dist;
	kickBallCloseToBody( angle, dist );	
	return;
      }
      if (num2 > 1){
	Vector2D c = ip21 + ((ip22 - ip21)/2);
	c = c - myself->getPosition();
	float angle = c.getDirection();
	float dist = c.distanceTo(myself->getPosition());
	dist = (myRange - dist)/2 + dist;
	kickBallCloseToBody( angle, dist );
	return;
      }
    }
    holdBall();
    return;
  }
  stopBall();
}


/**********************************************************************************/
/**
 * Lets the agent stop the ball, i. e. kick the ball in such a manner, that
 * the ball looses all his velocity relativ to the agent velocity.
 * @return true if server command is queued (job done)
 */
/**********************************************************************************/
bool SkillPool::stopBall(){
  //DEBUG1_HARLEKIN("stopBall()");
  //calc stop vector
  static Vector2D neededVel, myPredPos, ballPos;
  neededVel  = -ball->getVelocity(); 
  myPredPos.setVector2D(myself->predictedPosition());
  ballPos.setVector2D(ball->getPosition());

  float colDist  = Configuration::ballSize + Configuration::getPlayerSize(myself->getType());
  float kickDist = colDist + Configuration::getKickableMargin(myself->getType());
  colDist  *= colDist;
  kickDist *= kickDist;
  float predDist = ballPos.distanceToSq(myPredPos);
  if (predDist < colDist || predDist > kickDist){
    neededVel += myself->getVelocity();     
  }
  
  //calc kick power and angle
  float power = worldModel->calcKickPowerForSpeed( neededVel.getLength() ); 
  float invRelVelAng = normalizeTo180Deg( neededVel.getDirection() - myself->getBodyDir() );
  
  effectors->queueKick(power, invRelVelAng);

  if (!ball->isVelHighConfident(worldModel->getTime())){
    DEBUG2_HARLEKIN("Working with invalid low confidence: ", ball->getVelConfidence(worldModel->getTime()));
  }
  if (ball->getVelocity().isInvalid()){
    DEBUG5_HARLEKIN("Working with invalid ballVelocity, kicking with (", power, ",", invRelVelAng, ") anyway...");
  }

  return true;
}


/**********************************************************************************/
/**
 * Let the agent pass the ball to a position using accelerateBallToVelocity
 * @param pos the position on the field to pass to ATTENTION: please set the targetVel!
 * @param targetVel the speed, that the ball should have at the target
 *        position
 */
/**********************************************************************************/
bool SkillPool::passBallToPos(const Vector2D &pos, float targetVel, bool oneKick){

  if (targetVel == -1.0) 
    targetVel = ConfigFile::passEndSpeed;
  float dist          = pos.distanceTo( ball->getPosition() );
  float initVelAmount = worldModel->calcKickSpeedToTravel( dist, targetVel );
  float power         = initVelAmount / Configuration::kickPowerRate;
 
  return kickToPoint( pos, power, oneKick );
}


/**********************************************************************************/
/**
 * Let the agent pass the ball to a player with a given targetVelocity
 * using the skill passBallToPos. ATTENTION: please set the targetVel!
 * @param passCand the position of the player on the field to pass to
 * @param targetVel the speed, that the ball should have when
 * the player intercepts the ball
 */
/**********************************************************************************/
bool SkillPool::passBallToPlayer(const Player* passCand, float targetVel, bool oneKick){
  if (targetVel == -1.0) 
    targetVel = ConfigFile::passEndSpeed;
  return passBallToPos( passCand->getPosition(), targetVel, oneKick );
  
  // Communication
  //worldModel->passPointUnumToSay   = passCand->getUniformNumber();
  //worldModel->passPointToSay       = passCand->getPosition();
  //worldModel->passPointToSayCycles = 3;
  /*
  DEBUG4_HARLEKIN("SkillPool::passBallToPlayer, setting passPoint: ", passCand->getPosition(),
		  " Unum: ", passCand->getUniformNumber());
  */
}


/**********************************************************************************/
/**
 * The skill enables the agent to dribble with the ball to a given point. The agent
 * therefore kicks the ball in such a way that he can intercept it a given 
 * distance ahead (kickAhead). This dribbling is a subbehaviour in which the agent
 * changes from kickToPoint() to getBall() and back. 
 * @param targetPos point to dribble with ball to
 * @param kickAhead distance which the agent kicks the ball ahead
 * @return at the moment simply true 
 */
/**********************************************************************************/
bool SkillPool::simpleDribbleToPoint(const Vector2D &targetPos, float kickAhead){

  //first: change kickAhead, if distance to target is too small
  const Vector2D &ballPos = ball->getPosition();
  float dist       = targetPos.distanceTo( ballPos );
  if( dist < kickAhead )
    kickAhead = dist;

  //calc kickAheadPoint : point to intercept the ball after getBall
  float angle = worldModel->angleFromPointToPoint( ballPos, targetPos );
  Vector2D kickAheadPoint = ballPos + Vector2D( kickAhead, angle, POLAR );

  int predCycles  = myself->predictedCyclesToPoint( kickAheadPoint, false );
  //predCycles + 1 because the agent needs one cycle to kick the ball
  //           + 1 because the agent will loose velocity while kicking
  float kickPower = worldModel->calcInitSpeedForDistInCycles( kickAhead, predCycles + 1 ) /
                    Configuration::kickPowerRate;

  //subbehaviour dribbling
  if( worldModel->isBallKickable() ){
    kickToPoint( targetPos, kickPower, false );
  }else{
    //no slow interception, because it is too slow
    getBall( 100, false, false );
  }
  return true;
}


/**********************************************************************************/
/**
  * Lets the agent perform a smart dribbling in the direction of the
  * targetPos. The dribbling goal is to get to the targetPos without 
  * losing the control over the ball, therefore the agent tries to hold
  * the ball kickable in every cycle by performing kicks or dashs.
  * The angle of the ball relativ to the body is controled over the parameter
  * dribbleAngle, so we can keep the ball away from opponents and still
  * dribble along.
  * @param targetPos the position in which direction should be dribbled
  * @param maxDashPower the dash power for dribbling
  * @param dribbleSide true for right side and false for left side
  * @param fuzzy use random function or not
  * @return true, if a turn, kick or dash is queued(so always true!)
  */
/**********************************************************************************/
bool SkillPool::smartDribbleToPoint(const Vector2D &targetPos, int maxDashPower, bool dribbleSide, bool fuzzy){

  //--INSERT DATA--
  //initializing posIn2Cycles
  static Vector2D posIn2Cycles;
  posIn2Cycles.setVector2D(0,0);
  //kick-variable  
  bool mustKick = false;
  //turn-variable with possible need in the next cycle  
  static bool mustTurn = false;
  
  //Configuration-Data
  int   playerType     = myself->getType();
  float playerSize     = Configuration::getPlayerSize(playerType);
  float kickableMargin = Configuration::getKickableMargin(playerType);
  float dashPowerRate  = Configuration::getDashPowerRate(playerType);
  float playerDecay    = Configuration::getPlayerDecay(playerType);

  //ConfigFile-Data 
  static float collisionBuffer                = ConfigFile::collisionBuffer;
  static float kickableBuffer                 = ConfigFile::kickableBuffer;
  static float wishedDistance                 = ConfigFile::stdDribbleDist * kickableMargin;
  static float wishedDribbleAngle             = ConfigFile::stdDribbleAng;
  static float angleWhenToTurnDribble         = ConfigFile::angleWhenToTurnDribble;
  static float angleWhenToTurnWithBallDribble = ConfigFile::angleWhenToTurnDribble * 2;
  static float angleToBodyForceKick           = ConfigFile::angleToBodyForceKick;
  
  //Myself and Ball-Data
  const Vector2D &ballPos = ball->getPosition();
  Vector2D ballPosPred    = ball->predictedPosition();
  const Vector2D &myVel   = myself->getVelocity();
  const Vector2D &myPos   = myself->getPosition();
  float    myBodyDir      = myself->getBodyDir();
  float    myEffort       = myself->getEffort();
  static Line     bodyDirLine    = Line(0,0,0);
  bodyDirLine.updateLineFromPositionAndAngle(myPos, myBodyDir);

  //calculation of correct maxDashPower
  maxDashPower = (int)myself->correctDashPowerForStamina(maxDashPower, myself->bodyInfo.getStamina());
  maxDashPower = (int)myself->getPowerForDash( targetPos - myPos, myBodyDir, myVel, myEffort, maxDashPower );
  
  //calculation of myself predicted position with maxDashPower
  Vector2D myPosPred    = myself->predictedPosition(1, maxDashPower, 0);
  
  //changing dribbling side
  if( !dribbleSide )
    wishedDribbleAngle = -wishedDribbleAngle;

  //buffering data
  float kickableMarginBallPlayer  = playerSize + Configuration::ballSize + kickableMargin;
  float bufferedKickableDistance  = kickableMarginBallPlayer - kickableBuffer;
  float bufferedCollisionDistance = playerSize + Configuration::ballSize + collisionBuffer;

  //--TURN-DECISION--
  //turn condition: if bodyDir angle to target larger than angleWhenToTurnDribble
  float angleToTarget = fabs( worldModel->angleToPointRel2Body( targetPos ));
  if( angleToTarget >= angleWhenToTurnDribble || mustTurn ){
    if( angleToTarget >= angleWhenToTurnWithBallDribble ){
      if (!turnWithBallTo(( targetPos - myPosPred ).getDirection())){
	mustTurn = false;
	return true;
      }
    }else{
      if (!faceOnlyBodyToPoint( targetPos )){
	mustTurn = false;
	return true;
      }
    }    
  }

  //--KICK-DECISION-- FUZZY OR NOT
  if( fuzzy ){
    //distance ball to agent next cycle 
    float distBallAgentNC    = min((myPosPred - ballPosPred).getLength() - playerSize - 
			       Configuration::ballSize, kickableMargin);
    //distance ball to kickableMarginBorder next cycle 
    float distBallKickAreaNC = max(kickableMarginBallPlayer -
			       (myPosPred - ballPosPred).getLength(), (float)0.0);
    //angle to ball rel to bodyDir next cycle
    float angleBallBodyNC    = normalizeTo180Deg((ballPosPred - myPosPred).getDirection() - 
			       myBodyDir);
    //distance to ball from bodyLine next cycle
    float distBallBodyLineNC = min(kickableMarginBallPlayer, 
                               bodyDirLine.distanceToPoint(ballPosPred));
    
    float normDistBallAgentNC, normDistBallKickAreaNC, normAngleBallBodyNC, normDistBallBodyLineNC, maxValue;
    
    //normalize value between 0 and 1 - find maxValue - set mustKick
    if( fabs( angleBallBodyNC ) <= 90 ){
	normDistBallAgentNC    = distBallAgentNC / kickableMargin;
	normDistBallAgentNC    = min(1.0, max( (-1.0 / 0.3) * normDistBallAgentNC + 1.0, 0.0 ) );
	normDistBallBodyLineNC = 1 - ( distBallBodyLineNC / kickableMarginBallPlayer ); 
	normDistBallBodyLineNC = min(1.0, max( (-1.0 / 0.3) * normDistBallBodyLineNC + 1.0, 0.0 ) );
	maxValue = max(normDistBallAgentNC, normDistBallBodyLineNC); 
	mustKick = maxValue > drand48();
    }else{
	normDistBallAgentNC    = distBallAgentNC / kickableMargin;
	normDistBallAgentNC    = min(1.0, max( (-1.0 / 0.3) * normDistBallAgentNC + 1.0, 0.0 ) );
	normDistBallKickAreaNC = distBallKickAreaNC / kickableMargin;
	normDistBallKickAreaNC = min(1.0, max( (-1.0 / 0.3) * normDistBallKickAreaNC + 1.0, 0.0 ) );
	normAngleBallBodyNC    = 1 - ( fabs( angleBallBodyNC ) / 180 );
	normAngleBallBodyNC    = min(1.0, max( (-1.0 / 0.3) * normAngleBallBodyNC + 1.0, 0.0 ) );
	maxValue = max( max( normDistBallAgentNC, normDistBallKickAreaNC ) , normAngleBallBodyNC);
	mustKick = maxValue > drand48();
    }
  }
  // not fuzzy
  else{
    float distance      = (myPosPred - ballPosPred).getLength();
    float predBallAngle = normalizeTo180Deg((ballPosPred - myPosPred).getDirection() - myBodyDir);
    float distanceLine  = bodyDirLine.distanceToPoint(ballPosPred);
    //check for upcoming collision in the next cycle, assuming we dash with maxDashPower
    bool  collisionNextCycle  = distance <= bufferedCollisionDistance;
    //check for kickable ball in the next cycle
    bool notKickableNextCycle = distance >= bufferedKickableDistance;
    //check for the right ball position, so ball is not too far behind(NOT USED)
    bool toFarBehindNextCycle = fabs(predBallAngle) > angleToBodyForceKick;
    //check for ball is too far or too near from agent bodyDirLine
    bool tooFarFromLine = distanceLine >= bufferedKickableDistance || distanceLine <= bufferedCollisionDistance;
    
    //decision over kick the ball - set mustKick
    mustKick = collisionNextCycle || 
	       ((toFarBehindNextCycle || notKickableNextCycle) && 
		(fabs(predBallAngle) > 90)) || 
	       tooFarFromLine;
  }
  
  //check on the right ball position for dribble side
  bool onWrongSideNextCycle = sign(normalizeTo180Deg((ballPosPred - myPosPred).getDirection() - myBodyDir)) != 
                              sign(wishedDribbleAngle);

  //calculation for future position of player and ball
  float acceleration         = maxDashPower * dashPowerRate * myEffort;
  float wishedDistanceToBall = playerSize + Configuration::ballSize + wishedDistance;
  float wishedAngleToBodyDir = normalizeTo180Deg(myBodyDir + wishedDribbleAngle);
  
  //--DECISION OVER KICK-- for above reasons or for turn
  //calculation of best ball position in 2 cycles - for 2 dashes
  if( mustKick || onWrongSideNextCycle ){
    
    posIn2Cycles = myPos + myVel + myVel * playerDecay + 
                   Vector2D(acceleration, myBodyDir, POLAR) +
                   Vector2D(wishedDistanceToBall, wishedAngleToBodyDir, POLAR);
    static Line in2CyclesLine = Line(0,0,0);
    in2CyclesLine.updateLineFromPositionAndAngle(posIn2Cycles, myBodyDir);
    float distanceBallToLine = in2CyclesLine.distanceToPoint(ballPos);
    float wishedLineDistance = min(fabs(bufferedKickableDistance - bufferedCollisionDistance)/4, 
			       distanceBallToLine/2);
    if(bodyDirLine.distanceToPoint(ballPos) <  bodyDirLine.distanceToPoint(posIn2Cycles))
      wishedLineDistance *= -1;
    
    posIn2Cycles += Vector2D(wishedLineDistance, normalizeTo180Deg(myBodyDir + 
		    (sign(worldModel->angleToBallRel2Body())*90)),POLAR);
  }
  //calculation of best ball position in 2 cycles - for turn and dash
  float turningAng = normalizeTo180Deg( worldModel->angleFromPointToPoint( myPosPred, targetPos ) - myBodyDir );
  if( fabs(turningAng) > angleWhenToTurnDribble ){
    mustTurn = true;      
    posIn2Cycles = /*myself->predictedPosition(1, 0, turningAng)*/ myPos + myVel + 
                   myVel.rotated(turningAng) * playerDecay + 
	           Vector2D( acceleration, worldModel->angleFromPointToPoint(myPosPred, targetPos), POLAR ) + 
	           Vector2D( playerSize + Configuration::ballSize + 
			     fabs(bufferedKickableDistance - bufferedCollisionDistance), 
			     normalizeTo180Deg(sign(wishedAngleToBodyDir) * 90 + myBodyDir + turningAng), POLAR);
  }
  
  //--CALCULATION FOR BEST BALL POSITION IN 2 CYCLES--
  if(posIn2Cycles != Vector2D(0,0)){
    //calculation for needed ball vector to get the ball to the wished point in two cycles
    float distanceBall    = (posIn2Cycles - ballPos).getLength();
    float angleBall       = (posIn2Cycles - ballPos).getDirection();
    float wishedBallSpeed = distanceBall / (1 + Configuration::ballDecay);
    
    Vector2D ballVector = Vector2D(wishedBallSpeed, angleBall, POLAR);
    return accelerateBallToVelocity( ballVector, false );
  }else{
    //the decision criteria for kicking the ball failed, so just go forward
    effectors->queueDash(maxDashPower);
    return true;
  }
  return true;
}


/**********************************************************************************/
/**
 * ROAD WORKS! please come around later for results :)
 * should be called scorePenalty no scoreGoal!
 * @param oppGoal selection of the goal - true for former oppGoal - false for ownGoal
 * @return always true
 */
/**********************************************************************************/
bool SkillPool::scoreGoal( bool oppGoal){
  // old method

  const float KICK_POWER = 100;
  const float Y_OFFSET   = 1.5;  //meter

  if(!worldModel->isBallKickable())
    return false;

  static Vector2D leftGoalCorner  = worldModel->field.getMarkerPosition(FLAG_GRT);
  static Vector2D rightGoalCorner = worldModel->field.getMarkerPosition(FLAG_GRB);
  static Vector2D goalMid         = worldModel->field.getMarkerPosition(GOAL_R);
  if(oppGoal){
    leftGoalCorner  = worldModel->field.getMarkerPosition(FLAG_GLT);
    rightGoalCorner = worldModel->field.getMarkerPosition(FLAG_GLB);
    goalMid         = worldModel->field.getMarkerPosition(GOAL_L);
  }

  const Vector2D  &myPos        = myself->getPosition();
  Vector2D  kickPoint;
  Vector2D *corner;
  const Player *enemyGoalie     = worldModel->theirTeam.getGoalie();

  float goalAngle   = UNDEFINED_FLOAT;
  float angleOffset = UNDEFINED_FLOAT;
  if(oppGoal){
    goalAngle = (goalMid - myPos).getDirection();
    angleOffset = 0;//8 + calcAngleOffset(goalAngle);
    //cerr << angleOffset << endl;
  }else{
    goalAngle   = (goalMid - myPos).getDirection();
    angleOffset = calcAngleOffset(goalAngle);
  }
  //  float angleOffset = calcAngleOffset(goalAngle);
  float yOffset;

  if (enemyGoalie == NULL){ //the opponent goalie is not seen
    //kick to the nearest corner of the goal
    if( myPos.distanceToSq(leftGoalCorner) < myPos.distanceToSq(rightGoalCorner) )
      corner = &leftGoalCorner, yOffset = Y_OFFSET;
    else
      corner = &rightGoalCorner, yOffset = -Y_OFFSET;
  }else{
    //draw the line from the own position to each goal corner
    Line toLeftCorner  = Line::makeLineFromTwoPoints(leftGoalCorner, myPos);
    Line toRightCorner = Line::makeLineFromTwoPoints(rightGoalCorner, myPos);
    const Vector2D &goaliePos = enemyGoalie->getPosition();

    //determine the distance from these lines to the position of the goalie
    float distToLeftLine  = toLeftCorner.distanceToPoint(goaliePos);
    float distToRightLine = toRightCorner.distanceToPoint(goaliePos);

    //kick there, where the distance is bigger
    if(distToLeftLine > distToRightLine)
      corner = &leftGoalCorner, yOffset = Y_OFFSET;
    else
      corner = &rightGoalCorner, yOffset = -Y_OFFSET;
  }

  kickPoint.setVector2D(corner->getX(), corner->getY() + yOffset + angleOffset);

  kickToPoint(kickPoint, KICK_POWER, true);
  return true;
 
  // new method
  /*
  Vector2D oppGoalTop    = worldModel->field.getMarkerPosition(FLAG_GRT);
  Vector2D oppGoalBottom = worldModel->field.getMarkerPosition(FLAG_GRB);
  Vector2D oppGoalMid    = worldModel->field.getMarkerPosition(GOAL_R);
  oppGoalTop.setY(oppGoalTop.getY() + 1.5);
  oppGoalBottom.setY(oppGoalBottom.getY() - 1.5);
  
  Vector2D relBallPos = ball->getPosition() - myself->getPosition();
  float ballDist  = relBallPos.getLength();
  float ballAngle = worldModel->angleToBallRel2Body();
  float ballSpeed = Configuration::maxpower * 
    ball->calculateKickRate(ballDist, ballAngle) * 
    Configuration::kickPowerRate;

  if (!worldModel->isIntersected_NH(oppGoalMid, ball->getPosition(), ballSpeed))
    kickToPoint(oppGoalMid, ballSpeed, true);
  else if (!worldModel->isIntersected_NH(oppGoalTop, ball->getPosition(), ballSpeed))
    kickToPoint(oppGoalTop, ballSpeed, true);
  else if (!worldModel->isIntersected_NH(oppGoalBottom, ball->getPosition(), ballSpeed))
    kickToPoint(oppGoalBottom, ballSpeed, true);
  else
    return false;

  return true;
  */
}


/**********************************************************************************/
/**
 * Lets the agent kick the ball into the open space between the opponents, if
 * "defensive" is set, to the nearest side out
 * @param defensive if set the agent shoots the ball to the nearest side out
 * @param oneKick for kickIn and freeKick usage
 */
/**********************************************************************************/
void SkillPool::kickAway(bool defensive, bool oneKick){
  
  const Vector2D &ballPos = ball->getPosition();
  static Vector2D kickAwayPoint;
  static Vector2D leftBorderPoint, rigthBorderPoint;
  //if defensive is set, then kick the ball to the nearest side out, 
  //but only if the nearest side out isn't intersectable by an opponent.
  //else normal kickAway for the position on the field
  if( defensive ){
    //calc kickAwayPoint defensive
    kickAwayPoint = Vector2D( ballPos.getX(), sign( ballPos.getY() ) * Field::WIDTH/2 );
    //check for intersection
    float ballDist  = worldModel->distanceToBall();
    float ballAngle = worldModel->angleToBallRel2Body();
    float ballSpeed = Configuration::ballAccelMax * ball->calculateKickRate( ballDist, ballAngle );
    if( !worldModel->isIntersected_NH( kickAwayPoint, ball->getPosition(), ballSpeed ) ){
      kickToPoint( kickAwayPoint, 100, oneKick );
      return;
    }
  }

  if( ballPos.getX() < -10 ){
    leftBorderPoint  = worldModel->field.getMarkerPosition( FLAG_T0   ); 
    rigthBorderPoint = worldModel->field.getMarkerPosition( FLAG_B0   );
  }else if( ballPos.getX() < 20 ){
    leftBorderPoint  = worldModel->field.getMarkerPosition( FLAG_RT30 ); 
    rigthBorderPoint = worldModel->field.getMarkerPosition( FLAG_RB30 );
  }else if( ballPos.getY() <  0 ){
    leftBorderPoint  = worldModel->field.getMarkerPosition( GOAL_R    ); 
    rigthBorderPoint = worldModel->field.getMarkerPosition( FLAG_PRC  );
  }else{
    leftBorderPoint  = worldModel->field.getMarkerPosition( FLAG_PRC  ); 
    rigthBorderPoint = worldModel->field.getMarkerPosition( GOAL_R    );
  }	
  kickAwayPoint = ballPos + Vector2D( 10, worldModel->getFreeAngle( ballPos, leftBorderPoint, rigthBorderPoint ), POLAR );
  kickToPoint( kickAwayPoint, 100, oneKick );
}


//------------------ kickTo-Skills Ende ------------------------------------------//


//------------------ other-Skills Start ------------------------------------------//


/**********************************************************************************/
/**
 * Moves the agent to his initial position before kickoff and after goals
 */
/**********************************************************************************/
bool SkillPool::moveToInitialPosition(){
  // don't use reference, will change it later
  Vector2D pos = myself->getInitPosition();
  Playmode  pm = worldModel->getPlaymode();
  FieldSide fs = worldModel->getFieldSide();
  
  bool canGetToMiddle = true;
  int cycle = worldModel->getTime()->getTime() % 6000;
  if (pm == PM_TheirKickOff) canGetToMiddle = false;
  if (pm == PM_OurGoal) canGetToMiddle = false;
  if (pm == PM_BeforeKickOff && cycle < 100 &&
      fs == FS_Right) canGetToMiddle = false;
  if (pm == PM_BeforeKickOff && cycle > 2000 &&
      fs == FS_Left) canGetToMiddle = false;

  // Keep away from the ball as we are not allowed near it
  if (!canGetToMiddle){
    if (pos.getLengthSq() < 81 /* 9 */){
      pos.setLength(10.5);
    }
  }

  if (myself->getPosition().distanceTo(pos) > 1){
    effectors->queueMove(pos.getX(), pos.getY());
    return false;
  }
  return true;
}


/**********************************************************************************/
/** 
 * Moves the goalie to a specific position in non-playOn-modi 
 * @param whereToMove the position to whichthe agent should be moved
 */
/**********************************************************************************/
void SkillPool::moveToPosition(const Vector2D &whereToMove){
  effectors->queueMove(whereToMove.getX(), whereToMove.getY());
}


/**********************************************************************************/
/**
 * Turns the neck in steps to the right. If no turn to the right is
 * possible, the neck will be turned to the maximum left position.
 * Calling this function a couple of cycles in a row, the field in front
 * of the agent can be scanned
 */
/**********************************************************************************/
void SkillPool::scanFieldWithNeck(){
  float step  = (myself->getViewAngle()/2);
  int   steps = (int)(180.0/step) + 1;
  int t = worldModel->getTime()->getCycles() % steps;
  float angle = normalizeTo180Deg(myself->getBodyDir() + step*t - 
				  Configuration::maxneckang);

  Vector2D watchPoint = myself->getPosition();
  watchPoint += Vector2D(10, angle, POLAR);
  faceOnlyNeckToPoint(watchPoint);
}


/**********************************************************************************/
/**
 * Turns the body in steps to the right. Calling this function a couple of
 * cycles in a row, the field around of the agent can be scanned
 */
/**********************************************************************************/
void SkillPool::scanFieldWithBody(){
 
  if(*(worldModel->getTime()) != worldModel->getLastSeeSSTime()) 
    return;
  Vector2D firstLook = Vector2D(10, myself->getBodyDir(), POLAR);
  float headRel2Body = normalizeTo180Deg(myself->getHeadDir() - myself->getBodyDir());
  //first align neck with body
  if(fabs(headRel2Body) > ConfigFile::angleWhenToTurn){
    faceNeckAndBodyToPoint(firstLook);
    return;
  }
  
  //with little threshold -> not yet in the configFile
  const float step = myself->getViewAngle() - 10;
  
  float speed = myself->getVelocity().getLength();
  effectors->queueTurn(step, speed);
  return;
}


/**********************************************************************************/
/**   
 * allows the agent to cover a line between two given points. 
 * the agent covers the line at a given coverDistance from the from-point,   
 * but if the coverDistance is larger than the distance between the two point
 * than the half of the distance between the point is taken.
 * @param Vector2D pos1 - from-point 
 * @param Vector2D pos2 - to-point
 */
/**********************************************************************************/
void SkillPool::coverLineFromTo(const Vector2D &pos1, const Vector2D &pos2){

  float coverAngle = worldModel->angleFromPointToPoint(pos1, pos2);
  Vector2D coverPoint = pos1 + Vector2D(ConfigFile::coverDistance, coverAngle, POLAR);
  
  //if distance between points smaller than coverDistance: 
  //take the half of the distance between points as coverDistance
  Line line = Line::makeLineFromTwoPoints(pos1,pos2);
  if(!line.isInBetween(coverPoint, pos1, pos2)){
    coverPoint = pos1 + Vector2D((worldModel->distanceFromPointToPoint(pos1, pos2)/2), coverAngle, POLAR);
  }  
  goToPoint(coverPoint, 1, 100);
  return;
}


/*******************************************************************************/
/**
 * the skill allows the agent to cover the line from the ball to the goal.
 * it decides which point of the goal has to be covered according
 * to the agent position
 */
/*******************************************************************************/
void SkillPool::coverGoal(){
  const Vector2D &myselfPos = myself->getPosition();
  const Vector2D &ballPos   = ball->getPosition();
  
  static const Vector2D &topGoalCorner    = worldModel->field.getMarkerPosition(FLAG_GLT);
  static const Vector2D &bottomGoalCorner = worldModel->field.getMarkerPosition(FLAG_GLB);
  static const Vector2D &goalMid          = worldModel->field.getMarkerPosition(GOAL_L);
  float distToTPath, distToMPath, distToBPath;
  Line lTop    = Line::makeLineFromTwoPoints(ballPos, topGoalCorner);
  Line lMid    = Line::makeLineFromTwoPoints(ballPos, goalMid);
  Line lBottom = Line::makeLineFromTwoPoints(ballPos, bottomGoalCorner);
  distToTPath = lTop.distanceToPoint(myselfPos);
  distToMPath = lMid.distanceToPoint(myselfPos);
  distToBPath = lBottom.distanceToPoint(myselfPos);

  if(distToTPath < distToMPath && distToTPath < distToBPath){
    coverLineFromTo(ballPos, topGoalCorner);
    return;
  }else if(distToBPath < distToMPath && distToBPath < distToTPath){
    coverLineFromTo(ballPos, bottomGoalCorner);
    return;
  }else{
    coverLineFromTo(ballPos, goalMid);
    return;
  }
}


/*******************************************************************************/
/**
 * this skill covers offensive: the line between the ball and a given enemy
 * or covers defensive: the line between the goal and a given enemy
 * @param const Player* enemy  
 * @param bool offensiv  
 */
/*******************************************************************************/
void SkillPool::coverEnemy(const Player* enemy, bool offensive){
  const Vector2D &enemyPos = enemy->getPosition();
  if(offensive){
    //cover offensive: cover line between ball and enemy
    const Vector2D &ballPos = ball->getPosition();
    coverLineFromTo(enemyPos, ballPos);
    return;
  }else{
    //cover defensive: cover line between goal and enemy
    static const Vector2D &goalMid = worldModel->field.getMarkerPosition(GOAL_L);
    coverLineFromTo(enemyPos, goalMid);
    return;
  }
}


/**********************************************************************************/
/**
 * Lets the agent communicate with his teammates and inform about his
 * beliefs about the environment. Because of limited communication only
 * the ball position and ball velocity is reported.
 */
/**********************************************************************************/
void SkillPool::reportBallStatus(bool dribbling){
  if(!ball->isPositionKnown()){
    DEBUG1("SkillPool::reportWorldModelStatus(): dont know the ball position!!! Why do I think I am the closest to the ball if I don't even known where that little sucker is?");
    return;
  }

  // This only works with at least 8 bytes
  // 
  if (Configuration::sayMsgSize < 8) return;


  char message[Configuration::sayMsgSize];

  float ballx = ball->getPositionPt()->getX();
  float bally = ball->getPositionPt()->getY();

  const Vector2D *vel;
  if (!dribbling)
    vel = ball->getVelocityPt();
  else
    vel = myself->getVelocityPt();
  float velx = vel->getX();
  float vely = vel->getY();

  int ballx1 = (int)ballx;
  int ballx2 = (int)(((ballx - ballx1) * 100.0)*(74.0/100.0));
  int bally1 = (int)bally;
  int bally2 = (int)(((bally - bally1) * 100.0)*(74.0/100.0));
  
  int velx_fit = (int)(velx * 27.0);
  int vely_fit = (int)(vely * 27.0);

  int posSigns = 0;
  if (ballx1 >= 0) posSigns += 1;
  if (bally1 >= 0) posSigns += 2;
  if (velx   >= 0) posSigns += 4;
  if (vely   >= 0) posSigns += 8;

  char info[7];
  int info_f[7];
  info_f[0] = posSigns;
  info_f[1] = abs(ballx1);
  info_f[2] = abs(ballx2);
  info_f[3] = abs(bally1);
  info_f[4] = abs(bally2);
  info_f[5] = abs(velx_fit);
  info_f[6] = abs(vely_fit);

  /************************************
   *
   *  int-val : char-val  : ascii-val
   * :::::::::::::::::::::::::::::::::
   *   0 -  9 : '0' - '9' : 
   *  10 - 35 : 'A' - 'Z' :
   *  36 - 61 : 'a' - 'z' :
   *  62 - 62 : ' '       : 32
   *  63 - 66 : '(' - '+' : 40 - 43
   *  67 - 69 : '-' - '/' : 45 - 47
   *  70 - 70 : '<'       : 60
   *  71 - 72 : '>' - '?' : 62 - 63
   *  73      : '_'       : 95
   *
   ************************************/

  for (int i = 0; i < 7; ++i){
    int my_msg = min((int)round(info_f[i]), 73);
    if (my_msg <  0){ DEBUG2("SkillPool::reportWorldModelStatus(): wrong value: ", my_msg); } 
    if (my_msg > 74){ DEBUG2("SkillPool::reportWorldModelStatus(): wrong value: ", my_msg); } 
    if      (my_msg < 10) info[i] = '0' +  my_msg;
    else if (my_msg < 36) info[i] = 'A' + (my_msg-10);
    else if (my_msg < 62) info[i] = 'a' + (my_msg-36);
    else if (my_msg < 63) info[i] = ' ';
    else if (my_msg < 67) info[i] = '(' + (my_msg-63);
    else if (my_msg < 70) info[i] = '-' + (my_msg-67);
    else if (my_msg < 71) info[i] = '<';
    else if (my_msg < 73) info[i] = '>' + (my_msg-71);
    else                  info[i] = '_';
  }

  message[0] = info[0];
  message[1] = info[1];
  message[2] = info[2];
  message[3] = info[3];
  message[4] = info[4];
  message[5] = info[5];
  message[6] = info[6];
  message[7] = ' ';
  message[8] = '\0';

  effectors->queueSay(message);
  
}

void SkillPool::reportPassPoint(){
  // This only works with at least 7 bytes
  // 
  if (Configuration::sayMsgSize < 7) return;


  char message[Configuration::sayMsgSize];

  int x1 = (int)worldModel->passPointToSay.getX();
  int x2 = (int)(((worldModel->passPointToSay.getX() - x1) * 100.0)*(74.0/100.0));
  int y1 = (int)worldModel->passPointToSay.getY();
  int y2 = (int)(((worldModel->passPointToSay.getY() - y1) * 100.0)*(74.0/100.0));

  int posSigns = 0;
  if (x1 >= 0) posSigns += 1;
  if (y1 >= 0) posSigns += 2;
  
  char info[7];
  int info_f[7];
  info_f[0] = 72;
  info_f[1] = posSigns;
  info_f[2] = abs(x1);
  info_f[3] = abs(x2);
  info_f[4] = abs(y1);
  info_f[5] = abs(y2);
  info_f[6] = worldModel->passPointUnumToSay;

  /************************************
   *
   *  int-val : char-val  : ascii-val
   * :::::::::::::::::::::::::::::::::
   *   0 -  9 : '0' - '9' : 
   *  10 - 35 : 'A' - 'Z' :
   *  36 - 61 : 'a' - 'z' :
   *  62 - 62 : ' '       : 32
   *  63 - 66 : '(' - '+' : 40 - 43
   *  67 - 69 : '-' - '/' : 45 - 47
   *  70 - 70 : '<'       : 60
   *  71 - 72 : '>' - '?' : 62 - 63
   *  73      : '_'       : 95
   *
   ************************************/

  //DEBUG2_HARLEKIN("[REPORT] Reporting my position: ", myself->getPosition());

  for (int i = 0; i < 7; ++i){
    int my_msg = min((int)round(info_f[i]), 73);
    if (my_msg < 0){ DEBUG2("SkillPool::reportPassPoint: wrong value: ", my_msg); }
    else if (my_msg <  10) info[i] = '0' +  my_msg;
    else if (my_msg <  36) info[i] = 'A' + (my_msg-10);
    else if (my_msg <  62) info[i] = 'a' + (my_msg-36);
    else if (my_msg <  63) info[i] = ' ';
    else if (my_msg <  67) info[i] = '(' + (my_msg-63);
    else if (my_msg <  70) info[i] = '-' + (my_msg-67);
    else if (my_msg <  71) info[i] = '<';
    else if (my_msg <  73) info[i] = '>' + (my_msg-71);
    else if (my_msg == 73) info[i] = '_';
    else { DEBUG2("SkillPool::reportPassPoint: wrong value: ", my_msg); }
    //DEBUG4_HARLEKIN("Converted ", my_msg, " to ", info[i]);
  }

  message[0] = info[0];
  message[1] = info[1];
  message[2] = info[2];
  message[3] = info[3];
  message[4] = info[4];
  message[5] = info[5];
  message[6] = info[6];
  message[7] = ' ';
  message[8] = '\0';
  effectors->queueSay(message);

}

void SkillPool::reportPosition(){
  // This only works with at least 7 bytes
  // 
  if (Configuration::sayMsgSize < 7) return;


  char message[Configuration::sayMsgSize];

  int x1 = (int)myself->getPositionPt()->getX();
  int x2 = (int)(((myself->getPositionPt()->getX() - x1) * 100.0)*(74.0/100.0));
  int y1 = (int)myself->getPositionPt()->getY();
  int y2 = (int)(((myself->getPositionPt()->getY() - y1) * 100.0)*(74.0/100.0));

  int posSigns = 0;
  if (x1 >= 0) posSigns += 1;
  if (y1 >= 0) posSigns += 2;
  
  char info[7];
  int info_f[7];
  info_f[0] = 73;
  info_f[1] = posSigns;
  info_f[2] = abs(x1);
  info_f[3] = abs(x2);
  info_f[4] = abs(y1);
  info_f[5] = abs(y2);
  info_f[6] = myself->getUniformNumber();

  /************************************
   *
   *  int-val : char-val  : ascii-val
   * :::::::::::::::::::::::::::::::::
   *   0 -  9 : '0' - '9' : 
   *  10 - 35 : 'A' - 'Z' :
   *  36 - 61 : 'a' - 'z' :
   *  62 - 62 : ' '       : 32
   *  63 - 66 : '(' - '+' : 40 - 43
   *  67 - 69 : '-' - '/' : 45 - 47
   *  70 - 70 : '<'       : 60
   *  71 - 72 : '>' - '?' : 62 - 63
   *  73      : '_'       : 95
   *
   ************************************/

  //DEBUG2_HARLEKIN("[REPORT] Reporting my position: ", myself->getPosition());

  for (int i = 0; i < 7; ++i){
    int my_msg = min((int)round(info_f[i]), 73);
    if (my_msg < 0){ DEBUG2("SkillPool::reportGoodPointToScore: wrong value: ", my_msg); }
    else if (my_msg <  10) info[i] = '0' +  my_msg;
    else if (my_msg <  36) info[i] = 'A' + (my_msg-10);
    else if (my_msg <  62) info[i] = 'a' + (my_msg-36);
    else if (my_msg <  63) info[i] = ' ';
    else if (my_msg <  67) info[i] = '(' + (my_msg-63);
    else if (my_msg <  70) info[i] = '-' + (my_msg-67);
    else if (my_msg <  71) info[i] = '<';
    else if (my_msg <  73) info[i] = '>' + (my_msg-71);
    else if (my_msg == 73) info[i] = '_';
    else { DEBUG2("SkillPool::reportGoodPointToScore: wrong value: ", my_msg); }
    //DEBUG4_HARLEKIN("Converted ", my_msg, " to ", info[i]);
  }

  message[0] = info[0];
  message[1] = info[1];
  message[2] = info[2];
  message[3] = info[3];
  message[4] = info[4];
  message[5] = info[5];
  message[6] = info[6];
  message[7] = ' ';
  message[8] = '\0';
  effectors->queueSay(message);
  
}


/**********************************************************************************/
/**
 * This is for goalie use only - it simple calculates the catchangle and queued the
 * catch command, even if the ball is uncatchable (the worldmodel could be wrong).
 * For determination the function shouldTryToCatchBall() should be used.
 * @return true  if server catch command is queued (always)
 *         
 */
/**********************************************************************************/
bool SkillPool::catchBall() {
  
  float catchAngle = worldModel->angleToBallRel2Body();
  effectors->queueGoalieCatch(catchAngle);
  return true;
  //}
}


/**********************************************************************************/
/**
 * synchronise the agent SEEMessageIncomingTime by switching the viewangle
 * at the right time 
 */
/**********************************************************************************/
void SkillPool::doSSProfiling(){
  DEBUG1_SYNCHRO("[REASONING] \tprofiling called");

  float AT_0_TOO_LATE   = 0.15 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);
  float AT_50_TOO_EARLY = 0.50 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);
  float AT_50_TOO_LATE  = 0.65 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);

  ViewQuality vq = myself->bodyInfo.getViewQuality();
  ViewWidth   vw = myself->bodyInfo.getViewWidth();
  // Number of sees of the current cycle
  int seeCount   = worldModel->getSeeCount();

  int timeDiff   = timerDiffMsec(worldModel->getTimeLastSee(),
                                 worldModel->getTimeLastSb());
  DEBUG2_SYNCHRO("[REASONING] \t timeDiff = ", timeDiff);

  if(vq == VQ_HIGH && vw == VW_NORMAL){
    effectors->sendChangeView(VW_NARROW, VQ_LOW);
    //effectors->queueChangeView(VW_NARROW, VQ_LOW);
    DEBUG1_SYNCHRO("[REASONING] \t Changing view to Narrow/Low");
    return;
  }

  if(ASSERT_DEBUG){
    assert(vq == VQ_LOW && vw == VW_NARROW);
  }

  // The see should be received during those milliseconds
  if(seeCount == 1 &&
     ( timeDiff > 0               && timeDiff <= AT_0_TOO_LATE  ) ||
     ( timeDiff > AT_50_TOO_EARLY && timeDiff <= AT_50_TOO_LATE )    ){
    effectors->sendChangeView(VW_NORMAL, VQ_HIGH);
    worldModel->setProfilingPerformed();
    DEBUG1_SYNCHRO("[REASONING] \t Profiling performed");
  }
}


/**********************************************************************************/
/**
 * switching the viewangle at the right time from normal to narrow and back
 */
/**********************************************************************************/
void SkillPool::doSwitchView(){
  ViewQuality vq = myself->bodyInfo.getViewQuality();
  ViewWidth   vw = myself->bodyInfo.getViewWidth();
  if(vw == VW_NARROW && vq == VQ_HIGH){
    effectors->sendChangeView(VW_NORMAL, VQ_HIGH);
    DEBUG1_SYNCHRO("[REASONING] \t Changing view to Normal/High");
    return;
  }else if(vw == VW_NORMAL && vq == VQ_HIGH){
    effectors->sendChangeView(VW_NARROW, VQ_HIGH);
    DEBUG1_SYNCHRO("[REASONING] \t Changing view to Narrow/High");
    return;
  }
}


/**********************************************************************************/
/**
 * simply performs a tackle in order to kick away the ball in the body direction 
 * with given power - in the behaviour it is decided when to tackle
 * @param power tackle power, affect the ball speed in the next cycle
 */
/**********************************************************************************/
void SkillPool::tackle(int power){
  effectors->queueTackle(power);
  return;
}


/**********************************************************************************/
/**
 * points the arm of the agent to a given point on the field
 * @param pos point on the field to point 
 * @param turnOff finishing the pointing 
 */
/**********************************************************************************/
void SkillPool::pointToPos(const Vector2D &pos, bool turnOff){
  if(turnOff){
    //10 for distance and 90 for direction is set, but wouldnt
    //be looked at in the effectors, simply turn off point to
    effectors->queuePointTo(10, 90, true);
    return;
  }

  //point to pos
  float distance  = worldModel->distanceToPoint(pos);
  float direction = worldModel->angleToPointRel2Head(pos);
  effectors->queuePointTo(distance, direction, false);
  return;
}


/**********************************************************************************/
/**
 * paying attention to a certain player in order to ignore the shit from
 * all other players on the field
 * @param speaker, player to pay attention to - for a while
 * @param turnOff finish paying attention to a certain player
 */
/**********************************************************************************/
void SkillPool::attentionTo(const Player* speaker, bool turnOff){
  //turn off attention 
  if(turnOff){
    //true for ownTeam and number 1 is set, but wouldnt
    //be looked at in the effectors, simply turn off attention to
    effectors->queueAttentionTo(true, 1, true);
    return;
  }

  //turn on attention to given player
  bool ourTeam = !speaker->isOpponent();
  int unum     = speaker->getUniformNumber();
  effectors->queueAttentionTo(ourTeam, unum, false);
  return;
}


/**********************************************************************************/
/**
 * simply gets the current score of the game from the server
 */
/**********************************************************************************/
void SkillPool::getScore(){
  effectors->queueScore();
  return;
}
//------------------ other-Skills Start ------------------------------------------//
void SkillPool::moveLikeAnIdiot(){
#define NUM_OURGOAL   9
#define NUM_THEIRGOAL 1

  int cycle = worldModel->getTime()->getTime();
  Playmode pm = worldModel->getPlaymode();

  if (pm == PM_OurGoal){
    int mode = cycle % NUM_OURGOAL;
  
    // OurGoal - Moves
    // Move 1: Sinus to left
    if (mode == 0){
      float x = worldModel->getTime()->getStoppedCount()/1.2 + 
	(12 - myself->getUniformNumber());
      while(x > 50){ 
	x -= 50; 
      }
      float y = sin(x);
      effectors->queueMove(-x,y);
      return;
    }
    // Move 2: Sinus to top
    if (mode == 1){
      float y = worldModel->getTime()->getStoppedCount()/1.2 + 
	(12 - myself->getUniformNumber()) - 30;
      while(y > 30)
	y -= 30;
      float x = 25 + sin(y);
      effectors->queueMove(-x,-y);
      return;
    }
    // Move 3: Circle
    if (mode == 2){
      float dist  = 15; //worldModel->getTime()->getStoppedCount()/1.2;
      float angle = ((int)(worldModel->getTime()->getStoppedCount() * 16.0) + 
	(12 - myself->getUniformNumber())*4)%360;
      angle = normalizeTo180Deg(angle);
      Vector2D pos = Vector2D(dist, angle, POLAR);
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }
    // Move 4: Double Circle
    if (mode == 3){
      float dist;
      float angle;
      if (myself->getUniformNumber() < 6){
	dist =  7; //worldModel->getTime()->getStoppedCount()/1.2;
	angle = -((int)(worldModel->getTime()->getStoppedCount() * 32.0) +
		 (12 - myself->getUniformNumber())*4)%360;
      }else{
	dist = 13;
	angle = ((int)(worldModel->getTime()->getStoppedCount() * 16.0) + 
		 (24 - myself->getUniformNumber())*4)%360;
      }
     angle = normalizeTo180Deg(angle);
      Vector2D pos = Vector2D(dist, angle, POLAR);
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }
    // Move 5: Rotating Line
    if (mode == 4){
      float dist  = (myself->getUniformNumber()-1)*2;
      float angle = ((int)(worldModel->getTime()->getStoppedCount() * 16.0))%360;
     angle = normalizeTo180Deg(angle);
      Vector2D pos = Vector2D(dist, angle, POLAR);
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }
    // Move 6: Rotating Line (mid)
    if (mode == 5){
      int unum = myself->getUniformNumber();
      float dist;
      if (unum == 1){
	dist = 0;
      }else{
	if (unum < 7){
	  dist = (unum - 1) *  2;
	}else{
	  dist = (unum - 6) * -2;
	}
      }
      float angle = ((int)(worldModel->getTime()->getStoppedCount() * 16.0))%360;
     angle = normalizeTo180Deg(angle);
      Vector2D pos = Vector2D(dist, angle, POLAR);
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }
    // Move 7: Going from mid to homePos
    if (mode == 6){
      int unum = myself->getUniformNumber();
      static Vector2D homePos = Formations::getCurrentFormation()->getInitPosition(unum);
      static Vector2D mid = Vector2D(-Field::LENGTH/4.0, 0);
      float fract = ((float)worldModel->getTime()->getStoppedCount())/42.0;
      static Vector2D way = homePos - mid;
      Vector2D actPos = mid + (way * fract);
      effectors->queueMove(actPos.getX(), actPos.getY());
      return;
    }
    // Move 8: Running around the field borders
    if (mode == 7){
      int unum = myself->getUniformNumber();
      int cycle = worldModel->getTime()->getCycles();
      int pos = unum + cycle;
      static int topPos = 25;
      static int botPos = 25;
      static int lefPos = 10;
      static int rigPos = 10;
      int maxPos = topPos + botPos + lefPos + rigPos;
      pos = pos % maxPos;
      if (pos < topPos){
	effectors->queueMove(Field::LENGTH/2.0 - pos * (Field::LENGTH/(float)topPos),-Field::WIDTH/2.0);
	return;
      }
      if (pos < topPos + lefPos){
	pos -= topPos;
	effectors->queueMove(-Field::LENGTH/2.0, -Field::WIDTH/2.0 + pos * (Field::WIDTH/(float)lefPos));
	return;
      }
      if (pos < topPos + lefPos + botPos){
	pos -= (topPos + lefPos);
	effectors->queueMove(-Field::LENGTH/2.0 + pos * (Field::LENGTH/(float)botPos),Field::WIDTH/2.0);
	return;
      }
      if (pos < topPos + lefPos + botPos + rigPos){
	pos -= (topPos + lefPos + botPos);
	effectors->queueMove(Field::LENGTH/2.0, Field::WIDTH/2.0 - pos * (Field::WIDTH/(float)rigPos));
	return;
      }
      return;
    }
    
    if (mode == 8){//rotating Werder W
      int sc = worldModel->getTime()->getStoppedCount();
      int unum = myself->getUniformNumber();
      float angle;
        
      angle =(int)(unum*28.0);//+(int)(sc * 0.0))%360;	
      Vector2D pos = Vector2D(7, angle, POLAR);      

      if      (unum ==  1) pos.setVector2D(  -3,  -5);
      else if (unum ==  2) pos.setVector2D(  -4, -3);
      else if (unum ==  3) pos.setVector2D(  -4.5, -0);
      else if (unum ==  4) pos.setVector2D(  -3.5,  +2.5);
      else if (unum ==  5) pos.setVector2D(  -2, +5);
      else if (unum ==  6) pos.setVector2D(  0, 3);
      else if (unum ==  7) pos.setVector2D(  +2,+5);
      else if (unum ==  8) pos.setVector2D(  +3, +3);
      else if (unum ==  9) pos.setVector2D(  +4,0.5);
      else if (unum == 10) pos.setVector2D( +3.5, -2);
      else if (unum == 11) pos.setVector2D( +2, -3.5);

      pos.rotate( (sc*18) %360);
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }


    // 3 ... 2 ... 1, doesn't work yet
    if (mode == 9){
      int sc = worldModel->getTime()->getStoppedCount();
      int submode = 4 - (int)ceil(sc / 14.0f);
      int unum = myself->getUniformNumber();
      Vector2D pos;
      if (submode == 3){
	if      (unum ==  1) pos.setVector2D(  0,  0);
	else if (unum ==  2) pos.setVector2D(  5,  3);
	else if (unum ==  3) pos.setVector2D(  5, -3);
	else if (unum ==  4) pos.setVector2D(  7,  7);
	else if (unum ==  5) pos.setVector2D(  7, -7);
	else if (unum ==  6) pos.setVector2D(  5, 11);
	else if (unum ==  7) pos.setVector2D(  5,-11);
	else if (unum ==  8) pos.setVector2D(  2, 14);
	else if (unum ==  9) pos.setVector2D(  2,-14);
	else if (unum == 10) pos.setVector2D( -1, 13);
	else if (unum == 11) pos.setVector2D( -1,-13);
      }
      if (submode == 2){
	if      (unum ==  1) pos.setVector2D(  0,  0);
	else if (unum ==  2) pos.setVector2D(  4, -4);
	else if (unum ==  3) pos.setVector2D( -4,  4);
	else if (unum ==  4) pos.setVector2D(  7, -7);
	else if (unum ==  5) pos.setVector2D(  0, 11);
	else if (unum ==  6) pos.setVector2D( -7, -7);
	else if (unum ==  7) pos.setVector2D( -6,  8);
	else if (unum ==  8) pos.setVector2D( -7, 10);
	else if (unum ==  9) pos.setVector2D( -3, 10);
	else if (unum == 10) pos.setVector2D(  2, 10);
	else if (unum == 11) pos.setVector2D(  7, 10);		
      }
      if (submode == 1){
	if      (unum ==  1) pos.setVector2D(  0,  0);
	else if (unum ==  2) pos.setVector2D(  0, -4);
	else if (unum ==  3) pos.setVector2D(  0, -8);
	else if (unum ==  4) pos.setVector2D(  0,-12);
	else if (unum ==  5) pos.setVector2D(  0,  4);
	else if (unum ==  6) pos.setVector2D(  0,  8);
	else if (unum ==  7) pos.setVector2D(  0, 12);
	else if (unum ==  8) pos.setVector2D( -2,-12);
	else if (unum ==  9) pos.setVector2D( -3,-10);
	else if (unum == 10) pos.setVector2D( -4,- 8);
	else if (unum == 11) pos.setVector2D( -5,- 6);		       	
      }
      pos.moveBy(-25,0);
      effectors->queueMove(pos.getX(), pos.getY());
      return;
    }
  }
  
  if (pm == PM_TheirGoal){
    moveToInitialPosition();
    return;

    // sad Smilie
    /*
      Vector2D pos[11] = { Vector2D( 0, 0), Vector2D(-5,-6), Vector2D( 5,-6),
      Vector2D(-1, 4), Vector2D( 1, 4), Vector2D(-2, 4), 
      Vector2D( 2, 4), Vector2D(-3.5, 5),
      Vector2D( 3.5, 5), Vector2D(-5,6), Vector2D( 5, 6),
      };
      Vector2D myPos = pos[myself->getUniformNumber() - 1] + Vector2D(-25,0);
      effectors->queueMove(myPos.getX(), myPos.getY());
      return;
    */
  }
}



















