/******************************************************************************
 * Ball.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Ball
 *
 * requires:
 *
 *
 * Created on: 15.11.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 ******************************************************************************/

#ifndef BALL_HH
#define BALL_HH

#include "common.hh"
#include "MovingObject.hh"
#include "ConfigFile.hh"
#include "Configuration.hh"
#include "debug.hh"

/**
 * Contains all the available Robocup information for the ball and some
 * ball-related calculations
 */
class Ball : virtual public MovingObject{
private:
  
public:
  Ball();
  virtual ~Ball();

  /**
   * calculates the kick rate: a factor between 0 und 1, expressing, how
   * many of the kick power would be effective transformed into ball
   * acceleration. This factor is dependent from relative angle to ball and
   * ball distance, it decreases with increasing of this values.
   * For further details, look in the soccer server manual, kick model
   * @param distance distance to ball
   * @param angle    relative angle to ball
   * @return kick rate, float [0.50..1]
   */
  float calculateKickRate(const float distance, const float angle) const;
    
  /**
   * calculates whether the ball data is good confident
   * @return true, if the ball has a good confidence, false otherwise
   */
  bool isGoodConfident(const Time *time) const {
    return getConfidence(time) > ConfigFile::goodConfidenceBall;
  }

  /**
   * calculates whether the ball data is high confident
   * @return true, if the ball data has a high confidence,
   * false otherwise
   */
  bool isHighConfident(const Time *time) const {
    return getConfidence(time) > ConfigFile::highConfidenceBall;
  }

  /**
   * calculates whether the ball velocity is good confident
   * @return true, if the ball velocity has a good confidence,
   * false otherwise
   */
  bool isVelGoodConfident(const Time *time) const {
    return getVelConfidence(time) > ConfigFile::goodConfidenceBall;
  }

  /**
   * calculates whether the ball velocity is high confident
   * @return true, if the ball velocity has a high confidence,
   * false otherwise
   */
  bool isVelHighConfident(const Time *time) const {
    return getVelConfidence(time) > ConfigFile::highConfidenceBall;
  }

  /**
   * Calculates the estimated velocity of the ball
   * @param cycles how many cycles to look ahead
   * @return the predicted velocity of the ball
   */
  float getPredictedVelocity(int cycles){
    return fabs(getVelocity().getLength()) * pow(Configuration::ballDecay, cycles);
  }
    
  /**
   * Calculates the estimated position of the ball
   * @param cycles how many cycles to look ahead
   * @param kickDir the direction of a possible kick
   * @param distance the distance from agent to the ball
   * @param angle the relative angle from the agent to the ball
   * @return the predicted position of the ball
   */
  Vector2D predictedPosition(int cycles=1, float kickDir = UNDEFINED_FLOAT,
			     float kickPower = UNDEFINED_FLOAT,
			     float distance = UNDEFINED_FLOAT,
			     float angle    = UNDEFINED_FLOAT) const;
  
  const MovingObject *getLastKickedBy() const { return lastKickedBy; }
  void setLastKickedBy(const MovingObject *p){ lastKickedBy = p; }

  const MovingObject *lastKickedBy;


};

#endif //BALL_HH



