/******************************************************************************
 * MovingObject.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class MovingObject
 *
 * requires:
 *
 *
 * Created on: 16.10.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 ******************************************************************************/

#ifndef MOVINGOBJECT_HH
#define MOVINGOBJECT_HH

#include "Geometry.hh"
#include "Time.hh"
#include "debug.hh"

/**
 * Class MovingObject contains all Robocup information that is available for
 * objects that can move and some calculations on this data.
 */
class MovingObject{
private:

  /**
   * is used by the method 'isInMotion'. Objects with velocity amount
   * (correlates to ro in polar representation) of lesser than this constant
   * value are considered as not moving. This technique is chosen because of
   * noise in the data
   */
  static const float VEL_THRESHOLD = 0.15;

protected:
  /** the absolute position of this object on the field */
  Vector2D position;

  /**
   * the absolute directed velocity of the object,
   * in cartesian representation
   * ( but can easily be converted in polar koordinates )
   * scale unit: meter per cycle
   */
  Vector2D velocity;

  /** the soccerserver time the object was last seen */
  Time timeSeen;
  
  /** the soccerserver time the position of the object was last sensed */
  Time timeOfPosition;
  
  /** the soccerserver time the velocity of the object was last sensed */
  Time timeOfVelocity;
  
public:
  MovingObject();

  /**
   * builds a MovingObject with the specified data stored in.
   * @param pos the absolute position of the object on the field
   * @param vel the absolute velocity of the object, represented cartesian,
   * measured in meter per simulation cycle
   */
  MovingObject(const Vector2D &pos, const Vector2D &vel);

  virtual ~MovingObject();

  /** \name Accessors */
  //@{
  void setPosition(const Vector2D &p) { position = p; }
  void setPosition(float nx, float ny) { 
    position.setVector2D(nx,ny);
  };
  void setVelocity(const Vector2D &v) { velocity = v; }
  void setVelocity(float nx, float ny) {
    velocity.setVector2D(nx,ny);
  };
  void setTimeSeen(const Time &time) { timeSeen = time; }
  void setTimeOfPosition(const Time &time) { timeOfPosition = time; }
  void setTimeOfVelocity(const Time &time) { timeOfVelocity = time; }
  const Vector2D &getPosition()   const { return position; }
  const Vector2D *getPositionPt() const { return &position; }
  const Vector2D *getVelocityPt() const { return &velocity; }
  const Vector2D &getVelocity()   const { 
    /*
    if (velocity.isInvalid()){
      DEBUG2_HARLEKIN("MovingObject::getVelocity(): returning invalid Velocity: ", velocity);
      return Vector2D(0,0);
    }
    */
    return velocity; 
  }
  const Time &getTimeSeen()       const { return timeSeen; }
  const Time &getTimeOfPosition() const { return timeOfPosition; }
  const Time &getTimeOfVelocity() const { return timeOfVelocity; }
  //@}

  /**
   * calculates a confidence value on the base of the current soccer server
   * cycle and the cycle the object was last seen. This value is a float from
   * [0..1].
   * @param time: the current soccer server time
   * @return a confidence value for the object
   */
  float getConfidence(const Time *time) const;

  /**
   * calculates the confidence value of the velocity on the base of the
   * current soccer server time and the time the object was last seen.
   * This value is a float from [0..1].
   * @param time: the current soccer server time
   * @return the velocity confidence value for the object
   */
  float getVelConfidence(const Time *time) const;
  
  /**
   * calculates whether the object data is good confident
   * @return true, if the object has a good confidence, false otherwise
   */
  virtual bool isGoodConfident(const Time *time) const = 0;

  /**
   * calculates whether the object data is high confident
   * @return true, if the object has a high confidence, false otherwise
   */
  virtual bool isHighConfident(const Time *time) const = 0;

  /**
   * calculates whether the object velocity is good confident
   * @return true, if the velocity of the object has a good confidence,
   * false otherwise
   */
  virtual bool isVelGoodConfident(const Time *time) const = 0;

  /**
   * calculates whether the object velocity is high confident
   * @return true, if the velocity of the object has a high confidence,
   * false otherwise
   */
  virtual bool isVelHighConfident(const Time *time) const = 0;
  
  
  /**
   * determines if the object is probably moving. It cannot be determined
   * exactly because of noise in sensor data
   * @return true, if the object is moving, false otherwise
   */
  bool isInMotion() const {
    return (velocity.getLength() >= VEL_THRESHOLD // because of noise
            && !velocity.isInvalid()); }

  /**
   * this method determines, if the position of this object is known.
   * @return true, if the position is known, false otherwise
   */
  bool isPositionKnown() const { return !position.isInvalid(); }

  /**
   * calculates the expected position of this object on the supposition that
   * the object wont change his velocity for the period the calculation looks
   * ahead.
   * @param cycles the position of the object in so many simulation cycles
   * should be predicted.
   * @return the predicted position of this object
   */
  virtual Vector2D predictedPosition(int cycles=1) const;

  /** outputs the data members of this object on a string steam */
  friend ostream& operator<< (ostream &os, const MovingObject &mo){
    os << " position: " << mo.position
       << " velocity: " << mo.velocity << endl;
    return os;
  }
};

#endif //MOVINGOBJECT_HH

