/******************************************************************************
 * Myself.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Myself
 *
 * requires:
 *
 *
 * Created on: 22.08.2001 by wenns
 * Maintainer: cniehaus, harlekin
 * Group:
 * Co-authors: wenns
 *
 * Changelog:
 *
 ******************************************************************************/

#include <algorithm>

#include "Myself.hh"
#include "Configuration.hh"
#include "Geometry.hh"
#include "debug.hh"


Myself::Myself()
  : bodyInfo(), velRelToBody(UNDEFINED_FLOAT, UNDEFINED_FLOAT)
{}

Myself::~Myself(){}


float Myself::getViewAngle() const{
  float angle;
  switch(bodyInfo.getViewWidth()){
  case VW_NARROW:
    angle = Configuration::visibleAngle * 0.5; 
    break;
  case VW_NORMAL: 
    angle = Configuration::visibleAngle ; 
    break;
  case VW_WIDE: 
    angle = Configuration::visibleAngle * 2; 
    break;
  case VW_UNKNOWN:
  default:
    DEBUG1_CERR("Myself::viewWidthToAngle(): error: view width unknown");
    angle = UNDEFINED_FLOAT;
    break;
  }
  return angle;
}


float Myself::correctDashPowerForStamina(const float dashPower,
					 const float stamina) const {
  // Needed power is doubled when running backwards  
  float power = (dashPower < 0.0) ? dashPower * -2.0 : dashPower;
  float reserve = (Configuration::getStaminaMax(type) * Configuration::effortDecThr) + Player::TIRED_BUFFER;
  power = max(min(stamina - reserve, power), 0.0f); 
  // set power to stamina if not enough stamina left
  power = (dashPower < 0.0) ? power / -2.0 : power; // set power back if running backwards
  return power; // * effort;
}


void Myself::updatePredictedStamina(float &stamina, float &effort,
                                    float &recovery, float dashPower) const {
  float maxStamina = Configuration::getStaminaMax(type);


  if(dashPower > 0)
    stamina -= dashPower;
  else
    //substact the doubled amount in case of running backward
    stamina -= 2 * dashPower;
  if(stamina < 0) stamina = 0.0;

  //handle recovery decrement
  if(stamina <= Configuration::recoverDecThr * maxStamina && 
     recovery > Configuration::recoverMin ){
    recovery -= Configuration::recoverDec ;
  }

  //handle effort decrement
  if(stamina <= Configuration::effortDecThr * maxStamina && 
     effort > Configuration::effortMin){
    effort -= Configuration::effortDec;
    effort  = max(effort, Configuration::effortMin);
  }

  //handle effort increment
  if(stamina >= Configuration::effortIncThr * maxStamina && 
     effort < Configuration::effortInit){
    effort += Configuration::effortInc;
    effort  = min(effort, Configuration::effortInit);
  }

  //increase the stamina amount (do regenerate)
  stamina += recovery * Configuration::getStaminaIncMax(type);
  stamina  = min(stamina, maxStamina);
}


Vector2D Myself::predictedPosition(int cycles, float dashPower,
                                   float turnAngle) const {
  if(ASSERT_DEBUG) assert(cycles >= 0);
  if(ASSERT_DEBUG) assert(dashPower <= 100 && dashPower >= -100);
  if(ASSERT_DEBUG) assert(turnAngle <= 180 && turnAngle >= -180);
  
  if (position.isInvalid() || velocity.isInvalid()){
    DEBUG1("Myself::predictedPosition(): Can't estimate future if don't know present");
    return position;
  }

  float currTurnAng   = turnAngle;
  float currDashPower = dashPower;
  float effPower, thisTurn;
  float predEffort    = bodyInfo.getEffort();
  float predStamina   = bodyInfo.getStamina();
  float predRecovery  = 1;  //REVISIT, always 1
  //float predRecovery = *RECOVER_INIT;
  float speedMax      = Configuration::getMaximumPlayerSpeed(type);
  float speedMaxSq    = speedMax * speedMax;
  float playerDecay   = Configuration::getPlayerDecay(type);
  float dashPowerRate = Configuration::getDashPowerRate(type);
  float inertiaMoment = Configuration::getInertiaMoment(type);

  float currBodyDir   = bodyDir;
  static Vector2D currPosition;
  currPosition = position;
  static Vector2D currVelocity;
  currVelocity = velocity;
  static Vector2D acceleration;

  for (int i=0; i<cycles; i++){
    currDashPower = correctDashPowerForStamina(dashPower, predStamina);

    //if a turn necessary
    if (currTurnAng != 0.0) {
      float actualTurnAngle = Configuration::maxmoment /
	(/*const*/ 1.0 + inertiaMoment * currVelocity.getLength() );

      thisTurn = min(max(-actualTurnAngle, currTurnAng), actualTurnAngle);
      //add the turn angle to the body direction
      currBodyDir = normalizeTo180Deg(currBodyDir + thisTurn);
      currTurnAng -= thisTurn;
      // Dont return the velocity!
      //currVelocity.rotate(thisTurn);
    }else{
      if(fabs(currDashPower) > predStamina) effPower = sign(currDashPower) * predStamina;
      else                                  effPower = currDashPower;
      
      //consider effort
      effPower *= (predEffort * dashPowerRate);
      acceleration.setVector2D(effPower, currBodyDir, POLAR);
      currVelocity += acceleration;

      //if the calculated speed exceds the maximum speed 
      if ( currVelocity.getLengthSq() > speedMaxSq )
	currVelocity.setLength( speedMax );       
    }
    // update position
    currPosition += currVelocity;
    //do decelerate
    currVelocity *= playerDecay;
    updatePredictedStamina(predStamina, predEffort, predRecovery, currDashPower);
  }
  return currPosition;
}

int Myself::predictedCyclesToPoint(const Vector2D &point, bool goalie) const {
  float actualSpeedMax   = Configuration::getMaximumPlayerSpeed(type);
  float actualSpeedMaxSq = actualSpeedMax * actualSpeedMax;
  float inertiaMoment    = Configuration::getInertiaMoment(type);
  float dashPowerRate    = Configuration::getDashPowerRate(type);
  float playerDecay      = Configuration::getPlayerDecay(type);

  int predCycles = 0;
  float leftStamina  = bodyInfo.getStamina();
  float leftEffort   = bodyInfo.getEffort();
  // recovery never set? ( = 999.333)
  float leftRecovery = 1.0; // getRecovery();

  int turnCycles = 0;
  static Vector2D pos, vel, acc;
  pos = position;
  vel = velocity;

  // not considering turning error
  float pointDistSq = pos.distanceToSq(point);
  float angleTurn, angleToPoint, velAmount, bD;
  bD = bodyDir;

  float tolerance = 0;
  if(goalie) tolerance = Configuration::catchableAreaL - ConfigFile::toleranceBuffer;
  else       tolerance = Configuration::getKickableMargin(type) + 
	       Configuration::ballSize + 
	       Configuration::getPlayerSize(type) - ConfigFile::toleranceBuffer;
  // use square values to save some time
  tolerance *= tolerance;
  
  static Line path = Line(0,0,0);

  while((pointDistSq - tolerance) > 0){
    // We should always know what our bodyDir is! - Ah, nice speech. ???
    bool shouldTurn = false;
    bool backwards  = false;
    if(!isinf(bodyDir) && !isnan(bodyDir)){
      angleToPoint = normalizeTo180Deg((point - (pos + vel) ).getDirection() - bD);
      
      path.updateLineFromPositionAndAngle((pos + vel), bD); // test wether pos+vel is better
      float endDist = path.distanceToPointSq(point);
      
      // determine if we should turn towards our target or even run backwards
      if (angleToPoint && endDist > tolerance){
	shouldTurn = true;
      }else if ((pointDistSq < 25 /*5*/ && fabs(angleToPoint) >= 170 /*170*/ /*&& vel.getLength() > 0.1*/)){
	shouldTurn = false;
	backwards  = true;
      }else if ((fabs(angleToPoint) > 80)){
	shouldTurn = true;
      }

      if (shouldTurn){
	turnCycles++; // needs another turn for turning ;)
	velAmount = vel.getLength(); // remaining velocity
	// maximum possible turn (minimum of needed and possible)
	if (angleToPoint >= 0){
	  angleTurn = min ( ( Configuration::maxmoment / (1.0f + inertiaMoment * velAmount) ), angleToPoint );
	}else{
	  angleTurn = max ( ( Configuration::minmoment / (1.0f + inertiaMoment * velAmount) ), angleToPoint );
	}
	// move
	pos += vel; // apply velocity
	// do not rotate the velocity!
	//vel.rotate( angleTurn ); // turn direction/velocity
	bD = normalizeTo180Deg(angleTurn + bD); // turn body (vel.getDirection()?)
	vel *= playerDecay;  // slow down   
	angleToPoint = normalizeTo180Deg((point - (pos + vel) ).getDirection() - bD); // new angle to point
	updatePredictedStamina(leftStamina, leftEffort, leftRecovery, 0); // renew stamina
	pointDistSq = pos.distanceToSq(point);
	continue;
      }//endif
    }//endif

    predCycles++;
    // Update dash power (we run slower if our stamina is low)
    float corDash = correctDashPowerForStamina((backwards ? -100 : 100), 
					       leftStamina);
    // Update Velocity (acceleration)
    acc.setVector2D(corDash * leftEffort * dashPowerRate, bD, POLAR);
    vel += acc;
    if (vel.getLengthSq() > actualSpeedMaxSq)
      vel.setLength(actualSpeedMax);
    // Update Position (velocity)
    pos += vel;
    // Update Velocity (decay)
    vel *= playerDecay;
    // Update stamina (recovery + strain)
    updatePredictedStamina(leftStamina, leftEffort, leftRecovery, corDash);
    pointDistSq = pos.distanceToSq(point);
  }

  return predCycles + turnCycles;
}


float Myself::getPowerForDash( const Vector2D &posRelTo, float angBody,
			       const Vector2D &vel, double dEffort, 
			       float maxPower ){
  float maxPlayerSpeed = Configuration::getMaximumPlayerSpeed(type);

  // get distance in direction
  double dAcc = posRelTo.rotated(-angBody).getX();  
  if( dAcc > maxPlayerSpeed )// if too far away
    dAcc = maxPlayerSpeed;   // set maximum speed
  dAcc -= vel.rotated(-angBody).getX(); // subtract current velocity

  // acceleration = dash_power * dash_power_rate * effort ->
  // dash_power = acceleration / (dash_power_rate * effort )
  float dDashPower = dAcc/(Configuration::getDashPowerRate(type) * dEffort );
  //running backwards double up the needed power for dash 
  if( dDashPower < 0 )
    dDashPower *= 2;
  
  // only apply the maximumPower if it is given
  if (maxPower) dDashPower = min(dDashPower, maxPower);

  if( dDashPower > Configuration::maxpower )
    return Configuration::maxpower;
  else if( dDashPower < Configuration::minpower )
    return Configuration::minpower;
  else
    return dDashPower;
}
