/******************************************************************************
 * Player.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Player
 *
 * requires:
 *
 * Created on: 22.08.2001 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 *****************************************************************************/

#ifndef PLAYER_HH
#define PLAYER_HH

//#include "WorldModel.hh"

#include <functional>
using namespace std;

#include "MovingObject.hh"
#include "Ball.hh"
#include "Formations.hh"


/**
 * Class Player contains all available Robocup information about a simulation
 * league player, some positioning information and methods for calculation
 * the predicted positions of a player
 */
class Player: public MovingObject{
protected:
  float bodyDir;     /**< the body direction of the player in degree */
  float headDir;     /**< the head direction of the player in degree,
			not relative to the body direction */
  
  Time timeOfDirs;    /**< the soccer server cycle directions above were
			 last updated */
  
  int uniformNumber; /**< the uniform number of the player */
  bool goalKeeper;   /**< true, if the player is the goalie */
  bool opponent;     /**< true, if the player is an opponent */
  
  bool close;        /**< true, if the player is in the "feeling area" and
			cannot be seen */
  bool  tackling;    /**< true, if the player is tackling */
  float pointDir;    /**< absolute direction the player is pointing with
			his arm */
  
  /** /name Additional data available by using of fullstate message */
  //@{
  int   type;     /**< 0: homo, heteros: 1-6 */
  float stamina;  /**< max 4000 by default */
  float effort;   /**< [0.6 .. 1) by default */
  float recovery; /**< [?? .. 1) by default */

  //@}
  
  /** the positioning of the player is calculated on the basis on this
      position */
  //Vector2D homePosition;
  /** the position the player has to take by a kickoff (begin of a halftime
      and after every goal) */
  //Vector2D initPosition;
  /** The Position where an Attacker shall run to if he is intersected from
      a Player with ther Ball*/
  Vector2D freePosition;
  
  //add accessors for this new member also

  
public:

 
  /** maximum number of cycles to look ahead by interceptions */
  // static const int    MAX_LOOKAHEAD = 25;
  static const int    MAX_LOOKAHEAD = 25;
  
  /**
   * if a stamina level of ( STAMINA_MAX * EFFORT_DEC_THR ) + TIRED_BUFFER
   * is reached, the actions wont be executed with full power. The buffer
   * assures, that neither effort nor recovery will be decreased.
   */
  static const int    TIRED_BUFFER  = 100;
  
  Player();
  

  /**
   * constructs a player object with the specified data stored in.
   * @param pos the absolute position of the player on the field
   * @param vel the absolute velocity of the player
   * @param bDir the body direction of the player in degree
   * @param hDir the head direction of the player in degree, absolute, not
   * relative to the body direction
   * @param unum the uniform number of the player
   * @param goalie true in the case of goalie, false otherwise
   * @param opponent true in the case of an opponent, false otherwise
   * @param close true in the case the player is in the "feeling area"
   * @param tackling true if the player is tackling
   * @param pointDir the absolute direction the player is pointing
   */
  Player(const Vector2D &pos, const Vector2D &vel, const float bDir,
	 const float hDir,    const int  unum,     const bool goalie,
	 const bool opponent, const bool close,    const bool tackling,
	 const float pointDir);
  
  virtual ~Player();
  
  /** \name Accessors */
  //@{
  Vector2D getInitPosition() const { 
    return Formations::currentFormation->getInitPosition(uniformNumber); 
  }
  Vector2D getHomePosition() const { 
    return Formations::currentFormation->getHomePosition(uniformNumber); 
  }
  Vector2D getFreePosition() const { return freePosition; }
  virtual float getBodyDir() const { return bodyDir; }
  virtual float getHeadDir() const { return headDir; }
  Time  getTimeOfDirs() const      { return timeOfDirs; }
  int   getUniformNumber() const   { return uniformNumber; }
  int   getUnum() const            { return uniformNumber; }
  bool  isGoalKeeper() const       { return goalKeeper; }
  bool  isOpponent() const         { return opponent; }
  bool  isClose() const            { return close; }
  bool  isTackling() const         { return tackling; }
  float getPointDir() const        { return pointDir; }
  virtual float getStamina() const { return stamina; }
  virtual float getEffort() const  { return 1.0; }
  virtual float getRecovery() const{ return recovery; }
  int   getType() const            { return type; }
  
  // Not needed anymore, since those informations are stored in
  // the Formation(s)-class
  //void setInitPosition(const Vector2D pos)  { initPosition=pos; }
  //void setHomePosition(const Vector2D home) { homePosition=home; }
  void setFreePosition(const Vector2D &free){ freePosition=free; }
  void setGoalKeeper(const bool g)          { goalKeeper = g; }
  void setOpponent(const bool o)            { opponent = o; }
  void setUniformNumber(const int uNum)     { uniformNumber = uNum; }
  void setHeadDir(const float hDir)         { headDir = hDir; }
  void setBodyDir(const float bDir)         { bodyDir = bDir; }
  void setTimeOfDirs(const Time &time)      { timeOfDirs = time; }
  void setClose(const bool c)               { close = c; }
  void setTackling(const bool t)            { tackling = t; }
  void setPointDir(const float pd)          { pointDir = pd; }
  void setType(const int t)                 { type = t; }
  
  void setAdditionalData(const int t, const float s, const float e,
			 const float r ){
    type = t, stamina = s, effort = e, recovery = r;
  }
  //@}

  void estimateState(Time *time);

  //comparison operators
  bool operator == (const Player &player) const;
  bool operator != (const Player &player) const { return !(*this == player); }
  
  /**
   * determines whether the player data is good confident
   * @param time must be set to current ss cycle
   * @return true, if the player has a good confidence, false otherwise
   */
  bool isGoodConfident(const Time *time) const {
    return getConfidence(time) > ConfigFile::goodConfidencePlayer;
  }
  
  /**
   * determines whether the player data is high confident
   * @param time must be set to current ss cycle
   * @return true, if the player has a high confidence, false otherwise
   */
  bool isHighConfident(const Time *time) const {
    return getConfidence(time) > ConfigFile::highConfidencePlayer;
  }
  
  /**
   * calculates whether the player velocity is good confident
   * @return true, if the player velocity has a good confidence,
   * false otherwise
   */
  bool isVelGoodConfident(const Time *time) const {
    return getVelConfidence(time) > ConfigFile::goodConfidencePlayer;
  }
  
  /**
   * calculates whether the player velocity is high confident
   * @return true, if the player velocity has a high confidence,
   * false otherwise
   */
  bool isVelHighConfident(const Time *time) const {
    return (getVelConfidence(time) > ConfigFile::highConfidencePlayer);
  }
  
  /**
   * determines whether the player is visible in the current cycle
   * @param time must be set to current ss cycle
   * @return true, if the player is visible, false otherwise
   */
  bool isVisible(const Time *time) const {
    return (timeSeen == *time); 
  }
  
  /**
   * determines whether the uniform number of the player is known
   * @return true, if the uniform number is known, false otherwise
   */
  bool isKnownByUnum() const{
    return (uniformNumber != UNDEFINED_INT);
  }
  
  /**
   * calculates the expected position of the player on the supposition that
   * the player be dashing the specified number of cycles with
   * the specified dash power. If a turn angle is specified, a turn is
   * assumed to be executed first.
   * @param cycles the position of the player in so many cycles should be
   * predicted
   * @param dashPower the dash power, that is assumed the player will be
   * dashing while the considered time period
   * @param turnAngle the turn angle, that is assumed the agent will be
   * turning before dashing.
   * @return predicted position of the player
   */
  virtual Vector2D predictedPosition(int cycles = 1, float dashPower = 0,
				     float turnAngle = 0) const;
  
  /**
   * calculates the number of cycles the player needed to reach the
   * specified position on the field.
   * @param point the position that should be reached
   * @param goalie if true, goalies greater reach is considered
   * @return number of cycles needed to reach the position
   */
  virtual int predictedCyclesToPoint(const Vector2D &point, bool goalie = false) const;
  virtual void updatePredictedStamina(float &stamina, float &effort,
				      float &recovery, float dashPower) const;
  float correctDashPowerForStamina(const float dashPower,
				   const float stamina) const;
  



  /** prints the values of the data members on a string stream */
  friend ostream& operator<< (ostream &os, const Player &p){
    os << "unum: " << p.uniformNumber << " position: " << p.position
       << " velocity: " << p.velocity << " bodyDir: " << p.bodyDir
       << " headDir: " << p.headDir
       << " goalie: " << p.goalKeeper << " opponent: " << p.opponent << endl;
    return os;
  }
};


////////////    STL Predicates for searching in STL containers  //////////////

/**
 * A Predicate for searching for a Player with a cetain uniform number in
 * a standard container
 */
class unumEquals : public unary_function<Player, bool> {
  int unum;
  public:
    unumEquals (const int u) : unum(u) {}
    bool operator() (const Player &p) const { return p.getUniformNumber() == unum; }
};

/**
 * A Predicate for searching for the goalie in a standard container
 */
class isGoalie : public unary_function<Player, bool> {
  public:
    bool operator() (const Player &p) const { return p.isGoalKeeper(); }
};

/**
 * A Predicate for searching for a Player with a cetain uniform number in
 * a standard container
 */
class isNotConfident : public unary_function<Player, bool> {
  const Time *_time;
  public:
    isNotConfident (const Time *time) : _time(time) {}
    bool operator() (const Player &p) const { return !p.isGoodConfident(_time);}};



#endif  //PLAYER_HH
