/******************************************************************************
 * Time.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Time
 *
 * requires:
 *
 *
 * Created on: 03.02.2003 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 *****************************************************************************/

#ifndef TIME_HH
#define TIME_HH

#include <iostream>
using namespace std;

/**
 * Contains the time representation of the soccerserver
 * The soccer server time cannot (satisfying) be represented by a single
 * int because in certain situations (playmodes, e.g FreeKickFault) the time
 * is stopped, but the cycles continue to run and actions are possible. 
 * Therefore it is represented by a tuple of (time, stoppCounter)
 * where time is the time stamp of the last sense-body message and stoppCounter
 * contains the number of cycles passed since the time was stopped by the
 * server (i.e. 0 the most time). 
*/
class Time{
private:
  static const int TIME_UNDEFINED = -1;
  
  int time;          /**< the current soccer-server cycle */
  int stoppedCount;  /**< number of cycles, since the time has been stopped
			(possible in many "dead ball" situations */
  
  void setTime(const int i)           { time = i; }
  void setStoppedCount(const int sc ) { stoppedCount = sc;  }
  Time getUndefinedTime() const       { return Time(TIME_UNDEFINED, 0); }

public:
  Time::Time();
  Time(const int time, const int stoppedCount);
  ~Time();
  
  void update(const int timestamp);
  void setUndefined();
  bool isUndefined()     const;
  int  getTime()         const { return time; }
  int  getStoppedCount() const { return stoppedCount; }
  int getCycles() const;

  void setTime(const int a, const int b){
    time = a; stoppedCount = b;
  }
  
  //operators
  Time operator + ( const int i )   const;
  Time operator + ( const Time &t ) const;
  Time operator - ( const Time &t ) const;
  Time operator - ( const int i )   const { return *this + (-i); }

  //the assign-operator for an int isn't provided. use update() instead
  void operator =  ( const Time &t );
  
  void operator += ( const int i )         { *this = *this + i; }
  void operator += ( const Time &t )       { *this = *this + t; }
  void operator -= ( const int i )         { *this = *this - i; }
  void operator -= ( const Time &t )       { *this = *this - t; }
  
  bool operator == ( const Time &t ) const;
  bool operator == ( const int i )   const;
  bool operator != ( const int i )   const { return !(*this == i); }
  bool operator != ( const Time &t ) const { return !(*this == t); }
  
  bool operator <  ( const Time &t ) const;
  bool operator <  ( const int i )   const;
  bool operator <= ( const Time &t ) const;
  bool operator <= ( const int i )   const;
  bool operator >  ( const Time &t ) const { return !( *this <= t ); }
  bool operator >  ( const int i )   const { return !( *this <= i ); }
  bool operator >= ( const Time &t ) const { return !( *this < t );  }
  bool operator >= ( const int i )   const { return !( *this < i );  }
  
  
  // dumps the contents of the time object on a output stream 
  friend ostream& operator << ( ostream &os, const Time &t ) {
    os << "(" << t.getTime() << ", " << t.getStoppedCount() << ")";
    return os;
  }
};

inline bool Time::isUndefined()     const{ 
  return getTime() == TIME_UNDEFINED; 
}

inline int Time::getCycles()       const{
  return max(time + stoppedCount, 0);
}

#endif //TIME_HH

