/******************************************************************************
 * WorldModel.hh
 * Copyright (c) 2001 Virtual Werder
 *
 *
 * implements:
 *   class WorldModel
 *
 * requires:
 *
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors:janw,wollny,asm,wenns,dbernau
 *
 * Changelog:
 *
 * $Id: WorldModel.hh,v 1.32 2004/06/28 16:53:07 mosh Exp $
 *
 *****************************************************************************/

#ifndef WORLDMODEL_HH
#define WORLDMODEL_HH


#include "Geometry.hh"
#include "Configuration.hh"
#include "Field.hh"
#include "Myself.hh"
#include "Team.hh"
#include "PotentialField.hh"
#include "Formations.hh"

#include "Sensors.hh"

/**
 * The class WorldModel represents all the knowledge of the agent about his
 * world. This knowledge can be sure or unsure, dynamic changing or static.
 * An hierarchic structure of the contained data has been chosen: e. g. the
 * class Worldmodel physically contains a part representing the current game
 * situation: "situation", which on his part contains parts representing the
 * both participating teams etc.
 * The methods of the class WorldModel can be logically splitted into
 * different groups:
 * - methods, that deal with updating the world model and synchronizing
 * - methods, that extracts and/or calculates useful information
 * By the latter an attempt of better dealing the unsureness of the knowledge
 * has been done, resulting in methods returning floats instead of booleans.
 * Yet no way was found to implement them in uniform and meaningfull manner.
 */
class WorldModel{
public:
  WorldModel();
  virtual ~WorldModel();

  /** \name Accessors */
  //@{
  Playmode getPlaymode()              const { return playmode; }
  FieldSide getFieldSide()            const { return fieldSide; }
  int getOurScore()                   const { return ourScore; }
  int getTheirScore()                 const { return theirScore; }
  const Time *getTime()               const { return &time; }
  Time getLastSeeSSTime()             const { return lastSeeSSTime;}
  vector<Player> & getUnknownPlayers()      { return unknownPlayers; }
  ServerCommandHistory & getSentCommands()  { return lastSentCommands; }
  ServerCommandHistory & getSentCommands_BackUp() { return lastSentCommands_BackUp; }
  ServerCommandCounters &
  getQueuedCommandCounters()                { return queuedCommands; }
  const timeval * getTimeLastSee()    const { return &lastSeeTime;   }
  const timeval * getTimeLastSb()     const { return &lastSbTime;    }
  int  getSeeCount()                  const { return seeCount;      }
  void setPlaymode(const Playmode p)        { playmode = p; }
  void setFieldSide(const FieldSide fs)     { fieldSide = fs; }
  void setOurScore(const int i)             { ourScore = i; }
  void setTheirScore(const int i)           { theirScore = i; }
  void updateTime(const int timeStamp)      { time.update(timeStamp); }
  void setProfilingPerformed()              { profilingPerformed = true; }
  //@}

  PenaltySide getPenaltySide()              { return penaltySide; }
  void setPenaltySide(PenaltySide ps)       { penaltySide = ps; }

  const Vector2D findBestPassCandOfferPoint(const Vector2D &tmPos,float freeAngle) const;

  void debugSeeTimes(){
    DEBUG2_SYNCHRO("[SENSE] timeDiff : ", (timerDiffMsec(&lastSeeTime, &lastSbTime)));
  }

  /**
   * Searches the angle with the most free space between
   * enemy players in the given borders.
   * @param start point from where to look
   * @param leftBorderPoint left limit point
   * @param rightBorderPoint right limit point
   * @return absolute angle to free space.
   */
  float getFreeAngle(const Vector2D &start, 
		     const Vector2D &leftBorderPoint, 
		     const Vector2D &rightBorderPoint);


  /**
   * Checks if our team posseses the ball
   * @param player to check
   */
  bool hasBallPossession();
  

  /**
   * Checks if the specified player possesses the ball
   * @param player to check
   */
  bool hasBallPlayer(const Player&) const;

  /**
   * Checks if the specified player has the ball in his kickableMargin
   * @param player to check
   */
  bool isBallKickablePlayer(const Player&) const;

  /**
   * Returns the teammate with the ball.
   * @return temmate with ball or NULL if no temmate with ball was found
   */
  const Player * getTMWithBall();

  /**
   * tests whether a new cycle, in which no acting was taken place, has began
   * and sets the flag indicating such a cycle to "false"
   * @return true, if acting is needed, false otherwise
   */
  bool testAndSetNewCycle() {
    bool b = newCycle;
    newCycle = false;
    return b;
  }

  /** sets the time of last see and signals the condition "newinfo" */
  void setTimeLastSee();

  /** sets the time of last sense body and signals the condition "newinfo" */
  void setTimeLastSB();

  /** signals the condition "new info is there" */
  void signalNewInfo();

  /**
   * This method blocks till new information has arrived. Information is
   * either a sense body or a see message. If there isn't received
   * information from the server for longer than 'a constant amount of
   * seconds' (currently: 3) the server as assumed dead and false is
   * returned.
   * @return true, if new information has arrived, false if server is assumed
   * dead
   */
  bool waitForNewInfo();

  /**
   * This method determines, if the agent is receiving the visual information
   * at optimal times relative to the beginning of the simulation cycles.
   * @return true, if the see times are optimal, false if not
   */
  bool haveOptimalSeeTimes() const;

  bool switchViewNeeded() const;

  /**
   * This method returns, if the agent is in the think phase, that means,
   * no further message will appear this cycle. The agent has to send (done)
   * when he has finished with thinking and sending. 
   * Only used in snychronized mode!
   * @return true, if the agent should be thinking and sending right now.
   */
  bool isNowThinkSegment();
  
  /**
   * Resets the isNowThinkSegment variable. Should be called by the
   * sense-thread, when the server sent a (think) message.
   */
  void setThinkSegment();

  /**
   *
   * Checks if the player collides with the ball.
   * This method is called after receiving a senseBody or/and a see-Message
   * (when the worldModel is updated)
   *
   */
  void checkForCollisions();
  
  /**
   * integrates the sensor information into the worldmodel. During integration
   * of visual information selflocalisation is performed using performSelfLoc()
   * At present, all information about the player is overriden each time
   * visual information is "integrated". The information to integrate is
   * gained using the public interface of Sensors
   * @param sensors the reference to the sensors
   * @return true, if fresh and significant information is there -> command
   * generation needed
   */
  bool integrateSensorInfo(Sensors *sensors);
  void integrateSubstitution(const Sense::Substitution &s);

  //////////////// The WorldModel-HighLevel-Interface //////////////////
  /**
   * tests the ball possession for agent's team
   * @return true, if the team has the ball, false otherwise
   */
  bool hasBallTeam() const;
  
  
  /**
   * tests the ball possession for the opponent's team
   * @return true, if the opponent team has the ball, false otherwise
   */
  bool hasBallOppTeam() const;
  bool hasBallPossessionOpp();
  /**
   * tests the ball kickable possession for the agent's team
   * @return true, if one player of the agent's team has the ball 
   * in his kickable area, false otherwise
   */
  bool isBallKickableTeam() const;

  /**
   * tests the ball kickable possession for the opponent's team
   * @return true, if one player of the opponent team has the ball 
   * in his kickable area, false otherwise
   */
  bool isBallKickableOppTeam() const;
  const Player * getAnEnemyWithBallKickable() const;

  /** \name DISTANCE FROM ... TO ... */
  //@{
  /**
   * returns the distance from a player to the opponent goal-out-line
   * @param p the player, distance from which should be returned
   * @return the distance to the own goal-out-line
   */
  float distanceToOwnGoalOutLine(const Player &p) const;

  /**
   * returns the distance from a player to the opponent goal-out-line
   * @param p the player, distance from which should be returned
   * @return the distance to the opponent goal-out-line
   */
  float distanceToOppGoalOutLine(const Player &p) const;

  /**
   * calculates the distance between two points on the field
   * @param v1 the first point
   * @param v2 the second point
   * @return the distance
   */
  float distanceFromPointToPoint(const Vector2D &v1, const Vector2D &v2) const;

  /**
   * calculates the distance from a point on the field and the ball
   * @param the point, from which the distance should be calculated
   * @return distance from point to ball
   */
  float distanceFromPointToBall(const Vector2D &v1) const;
  
  /**
   * calculates the distance from the agent to a point on the field
   * @param v the aforesaid point
   * @return the distance to a point
   */
  float distanceToPoint(const Vector2D &v) const;
  float distanceToPointSq(const Vector2D &v) const;

  /** Returns the Player specified by his unum or a NULL Pointer if the Player was not found
      @param unum the Uniform Number after which is searched
      @return a Player Object of the Player specified by the overgiven Unum
  */
  const Player* getPlayerByUnum(int unum);

  /** Returns the Player specified by his unum or a NULL Pointer if the Player was not found
      @param unum the Uniform Number after which is searched
      @return a Player Object of the Player specified by the overgiven Unum
  */
  const Player* getOtherTeamPlayerByUnum(int unum);

  /**
   * calculates the distance from the agent to the ball
   * @return the distance to the ball
   */
  float distanceToBall() const;

  /**
   * calculates the distance from the agent to a player
   * @param p player, distance to which should be calculated
   * @return distance to a player
   */
  float distanceToPlayer(const Player &p) const;

  /**
   * calculates the distance from a player to a point on the field
   * @param p the aforesaid player
   * @param v the aforesaid point on the field
   * @return distance from player to point
   */
  float distanceFromPlayerToPoint(const Player &p, const Vector2D &v) const;

  /**
   * calculates the distance from a player to the ball
   * @param p the aforesaid player
   * @return distance from player to the ball
   */
  float distanceFromPlayerToBall(const Player &p) const;
  //@}


  /** \name ANGLES TO ... (RELATIVE TO ...) */
  //@{
  /**
   * Calculates the angle between the own position and pos,
   * relative to the own bodyDir
   * @param pos the position to which the angle should be calculated
   * @return an angle of [-180..180] degree
   */
  float angleToPointRel2Body(const Vector2D &pos) const;

  /**
   * Calculates the angle between the own position and pos,
   * relative to the own headDir
   * @param pos the position to which the angle should be calculated
   * @return an angle of [-180..180] degree
   */
  float angleToPointRel2Head(const Vector2D &pos) const;
  
  /**
   * Calculates the global angle between the own position and pos
   * @param pos the position to which the angle should be calculated
   * @return an angle of [-180..180] degree
   */
  float angleToPoint(const Vector2D &pos) const;

  /**
   * Calculates the global angle between pos1 and pos2
   * @param from the begin of the vector, which direction should be calculated
   * @param to the end of the vector, which direction should be calculated
   * @return an angle of [-180..180] degree
   */
  float angleFromPointToPoint(const Vector2D &from,
			      const Vector2D &to) const;

  /**
   * Calculates the angle between the own position and the ball position,
   * relative to the own bodyDir
   * @return an angle of [-180..180] degree
   */
  float angleToBallRel2Body() const;
  
  /**
   * Calculates the angle between the own position and the ball position,
   * relative to the own headDir
   * @return an angle of [-180..180] degree
   */
  float angleToBallRel2Head() const;
  
  /**
   * Calculates the global angle between the own position and the ball
   * position
   * @return an angle of [-180..180] degree
   */
  float angleToBall() const;
  //@}
  
  
  /** \name IS SOMEBODY CLOSE(ST) TO SOMETHING/SOMEBODY */
  //@{
  /**
   * tests whether the given player is the closest player to the given point
   * on the field
   * @param player the player to test
   * @param point the aforesaid point
   * @param excludePlayer an optional player that should not be taken into account
   * @return true, if the player is closest, false otherwise
   */
  bool isClosestToPoint(const Player &player, 
			const Vector2D &point,
			float conf = -1,
			const Player *excludePlayer = NULL) const;
  /**
   * tests whether the first given player is the closest player to the 
   * second given player
   * @param from the first player
   * @param to the second player
   * @return true, if the first player is the closest to the second player
   */
  bool isClosestToPlayer(const Player &from, 
			 const Player &to,
			 float conf = -1) const;
  /**
   * tests whether the given player is closest to the ball
   * @param player the player to test
   * @return true, if he closest to the ball is, false otherwise
   */
  bool isClosestToBall(const Player &player, float conf=-1) const;
  
  bool isClosestTeammateToPoint(const Player &teammate,
				const Vector2D &p,
				float conf=-1,
				const Player *excludePlayer = NULL, 
				bool considerMayKickBall = false) const;
  

  bool isClosestTeammateToPlayer(const Player &teammate,
				 const Player &player,
				 float conf=-1) const;
  
  bool isClosestTeammateToBall(const Player &teammate,
			       float conf=-1, 
			       bool considerMayKickBall = false);
  
  bool isSecondClosestToPoint(const Player &player,
			      const Vector2D &point,
			      float conf=-1,
			      const Player *excludePlayer = NULL) const;
  
  bool isSecondClosestToPlayer(const Player &from,
			       const Player &to,
			       float conf=-1) const;
  
  bool isSecondClosestToBall(const Player &player,
			     float conf=-1) const;
  
  bool isSecondClosestTeammateToPoint(const Player &teammate,
				      const Vector2D &point,
				      float conf=-1,
				      const Player *excludePlayer = NULL) const;
  
  bool isSecondClosestTeammateToPlayer(const Player &teammate,
				       const Player &player,
				       float conf=-1) const;
  
  bool isSecondClosestTeammateToBall(const Player &teammate,
				     float conf=-1) const;
  
  /**
   * tests whether the given player is close to the opponent goal out line
   * @param player the player to test
   * @return true if he is, false otherwise
   */
  bool isCloseToOpponentGoalOutLine(const Player &player) const;

  /**
   * tests whether the given player is close to the own goal out line
   * @param player the player to test
   * @return true if he is, false otherwise
   */
  bool isCloseToOwnGoalOutLine(const Player &player) const;
  //@}


  /** \name NUMBER OF PLAYERS IN AREAS
      
      the methods below calculate how much of the players in a certain
      player set are inside a certain area. The player set is either
      "all players" or "teammates" or "opponents" or "unknown players".
      The areas are arbitrary Rectagles, Circles or Cones on the field.
      The parameter conf is used to filter the players: the players with
      confidence values below are not considered by the calculation. If
      the last parameter is skipped, a confidence value from the configfile
      is taken
  */
  //@{
  int numInSetInRectangle(const PlayerSet ps, const Rectangle &rect,
			  float conf=-1) const;

  int numInSetInCircle(const PlayerSet ps, const Circle &circle,
		       float conf=-1) const;

  int numInSetInCone(const PlayerSet ps,
		     const Cone &cone,
		     float conf=-1) const;
  
  /*TEST*/ int numInSetInCone(const PlayerSet ps,
			      const Vector2D &start,  // Start of Cone
			      const Vector2D &end,    // End of Cone
			      double width,    // Width of Cone after a distance of 1.0
			      float conf=-1) const;

  //@}


  /** \name BEING IN SOME AREAS 
      the methods below determine whether a point on the field or a player
      are inside some area and return true if they are. The areas are either
      the whole field, the own penalty room or the opponents penalty room. 
  */
  //@{
  bool isOnField(const Vector2D &p) const;

  bool isOnField(const Player &p) const;
  
  bool isInOwnPenaltyArea(const Vector2D &p) const;
  
  bool isInOwnPenaltyArea(const Player &player) const;

  bool isInOppPenaltyArea(const Vector2D &p) const;
  
  bool isInOppPenaltyArea(const Player &player) const;
  //@}


  /** \name OFFSIDE RELATED */
  //@{
  /**
   * finds the offside line. The offside line is the X-Coordinate of the 
   * second nearest opponent player to the opponent Goal Outline or the
   * X-Coordinate of the ball, if it is closer to the Goal Outline than the
   * second nearest opponent. 
   * If no opponents are seen, OPP_GOAL_OUTLINE is returned as OffsideX.
   * @return the X-Coordinate of the offside line
   */
  float getOffsideX(float conf = -1) const;
  float getOppOffsideX(float conf = -1) const;
  /**
   * tests whether the given player is offside. If no opponents are seen,
   * "false" is assumed.
   * @param player the player to test
   * @return true, if the player is offside, false otherwise
   */
  bool isOffside(const Player &player, float conf = -1) const;

  /**
   * tests whether the given point is offside. If no opponents are seen,
   * "false" is assumed.
   * @param p the point to test
   * @return true, if the point is offside, false otherwise
   */
  bool isOffside(const Vector2D &p, float conf = -1) const;
  //@}

  /** Returns if the current Agent is the closest Teammate in front of
   *  the Player specified as Parameter
   *  @param player test player
   *  @return if the current Agent is the closest to the overgtiven Player
   */
  bool closestTMInFrontOfPlayer(const Player& player);
  

  /** \name BEING ATTACKED */
  //@{
  /**
   * tests whether the given player attacked/pressed by the his opponents
   * @param player the player to test
   * @return true if he is, false otherwise
   */
  bool isAttacked(const Player &player) const;
  
  /**
   * tests whether the given player is frontal attacked/pressed by the opponents
   * @param player the player to test
   * @return true if he is, false otherwise
   */
  bool isFrontalAttacked(const Player &player) const;

  /** Returns the Point to which it is best to dribble because it is
      the direction in which the next opponent is farest Away
      @param player the Player who shall perform the dribble
  */
  void getBestDribblePoint(const Player &player, Vector2D &drPt);
  
  /** \name MISC PREDICATES */
  //@{
  /**
   * tests whether the ball is catchable for the goalie
   * @return true, if the ball is catchable, false otherwise
   */
  bool isBallCatchable() const;

  /**
   * tests whether the ball is kickable for the agent at the present
   * @return true, if the ball is kickable, false otherwise
   */
  bool isBallKickable() const;

  /**
   * tests whether the ball will be kickable in the given number of cycles
   * on the supposition that the player will be dashing with the given
   * dash power and the ball wont change his velocity
   * @param cycles specifies, how far in the future should be looked
   * @param dashPower this dash power is assumed by the estimation
   * @return true, if the ball will be kickable, false otherwise
   */
  /*TEST*/ bool willBallBeKickable(int cycles=1, float dashPower=0) const;

  /**
   * tests, if a shot is a goal shot, so the goalie has to react on it.
   * @return true, if the shot is a goal shot, false otherwise
   */
  /*TEST*/ bool isGoalShot() const;

  /**
   * tests, if a point on the field is a good position to try a goal shoot
   * @param p the position to test
   * @return true, if the position is promising for a goal shoot, false otherwise
   */
  /*REIMPLEMENT*/ bool isAGoodPointToScore(const Vector2D &p, bool oppGoal=false);

  /**
   * tests whether the given player is at a promising position to try a goal shoot
   * @param player the player to test
   * @return true if he is, false otherwise
   */
  /*TEST*/ bool isAtGoodPointToScore(const Player &player, bool oppGoal=false);
  //@}
  

  //-----------------------  closest to point ---------------------------
  /** \name CLOSEST TEAMMATE, OPPONENT, ... TO POINT, PLAYER, BALL, ...
      
      the methods below calculate the closest player in a player set
      (which can be PS_Players, PS_Teammates, PS_Opponents or PS_Unknown)
      to some object (which can be a point on the field, a player, the
      ball, myself, or one of the goal outlines. The parameter conf is used
      to filter the players: the players with confidence values below are
      not considered in the calculation. If conf is skipped, a confidence
      value from the configfile is taken.
  */
  //@{
  const Player* closestInSetToPoint(const PlayerSet ps,
				    const Vector2D &point,
				    float conf = -1,
				    const Player *excludePlayer = NULL,
					  bool considerMayKickBall = false) const;
  
  const Player* secondClosestInSetToPoint(const PlayerSet ps,
					  const Vector2D &point,
					  float conf = -1,
					  const Player *excludePlayer = NULL) const;
  
  const Player* closestInSetToPlayer(const PlayerSet ps,
				     const Player &player,
				     float conf = -1) const;
  
  const Player* closestInSetToMe(const PlayerSet ps,
				 float conf = -1) const;
  
  const Player* closestInSetToBall(const PlayerSet ps,
				   float conf = -1) const;
  
  const Player* secondClosestInSetToPlayer(const PlayerSet ps,
					   const Player &player,
					   float conf = -1) const;
  
  const Player* secondClosestInSetToMe(const PlayerSet ps,
				       float conf = -1) const;
  
  const Player* secondClosestInSetToBall(const PlayerSet ps,
					 float conf = -1) const;
  
  const Player* closestInSetToOwnGoalOutline(PlayerSet ps,
					     float conf=-1) const;
  
  const Player* closestInSetToOppGoalOutline(PlayerSet ps,
					     float conf=-1) const;
  //@}

  
  /*TEST*/ const Player* fastestInSetToPoint(const PlayerSet ps,
					     Vector2D pos,
					     int minCycles,
					     float conf,
					     bool considerMayKickBall) const;
  
  /*TEST*/ const Player* fastestInSetToBall(const PlayerSet ps,
					    float conf = -1,
					    bool considerMayKickBall = false) const;
  const Player* fastestInSetToBall(const PlayerSet ps, bool considerMayKickBall) const {
    return fastestInSetToBall(ps, -1, considerMayKickBall);
  }
 
  /**
   * This method determines the power with which the ball must be kicked in
   * order to travel a given distance and still have a speed after that
   * distance
   * @param dist distance ball should travel
   * @param endSpeed speed ball should have at target position
   * @return power value for kick command
   */
  float calcKickSpeedToTravel( const float dist, const float endSpeed ) const;

  /**
   * This method determines the kick power which is needed to speed up
   * the ball to a given speed. 
   * @param speed speed ball should have
   * @return power value for kick command
   */
  float calcKickPowerForSpeed( const float speed ) const;

  /**
   * This method returns the speed that has to be given to the ball when
   * it should have an endspeed of 'dEndSpeed' after 'dCycles' number of cycles.
   * This can be calculated using a geometric series.
   * @param endSpeed desired end speed for the ball
   * @param cycles nr of cycles after which ball should have speed 'endSpeed'
   * @return initial speed given to the ball to have speed 'endSpeed' after
   * 'cycles' cycles.
   */
  float calcInitSpeedFromEndSpeed( const float endSpeed,
				   const float cycles ) const;

  /**
   * This method returns the speed that has to be given to the ball when
   * it should travel a given distance in a given numbers of cycles.
   * @param distance distance to travel
   * @param cycles nr of cycles after which the ball should have travel the distance
   * @return initial speed given to the ball to travel the given distance
   * in the given number of cycles.
   */
  float calcInitSpeedForDistInCycles( const float distance,
				      const int cycles ) const;
  
  /**
   * This method determines the actual angle that is used when 'angle' is
   * sent to the SoccerServer.
   * @param angle angle send with turn command
   * @param speed current speed of the player
   * @return actual angle that player is turned
   */
  float calcActualTurnAngle( const float angle, const float speed ) const;
  
  /**
   * This method determines the angle that should be sent to the soccerserver
   * when the player wants to turn desiredAngle.
   * @param desiredAngle angle that player wants to turn
   * @param speed current speed of the player
   * @return angle that can be sent with turn command
   */
  float calcAngleForTurn( float const desiredAngle, const float speed ) const;


  /**
   * This method calculates the point, in which the ball on the shortest
   * way can be intercepted by the player p. Moreover it calculates the number
   * of cycles needed to get to this point. The calculation assumes that the
   * ball wont change his velocity.
   * @param p the player for which the calculation has to be done
   * @param intPoint by this reference the calculated point is returned
   * @param noCycles by this reference the number of cycles to the
   * interception point is returned
   * @return true, if interception is possible, false otherwise
   */
  bool interceptionPoint(const Player& p, Vector2D& intPoint,
                         int &noCycles, bool goalie=false);

  /**
   * This method calculates the point, in which the ball on the easiest
   * way can be intercepted by the player p. Moreover it calculates the number
   * of cycles needed to get to the point, where the agent needs the lowest 
   * number of cycles to. The calculation assumes that the
   * ball wont change his velocity.
   * @param p the player for which the calculation has to be done
   * @param safetyPoint by this reference the calculated point is returned
   * @param noCycles by this reference the number of cycles to the
   * safety point is returned
   * @return true, if interception is possible, false otherwise
   */
  bool safetyPoint(const Player& p, Vector2D &safetyPoint,
                   int &noCycles, bool goalie = false);

  void resetPrecalculatedValues();

  //----------------------------------------------------------------------
  //kick it as soon the behavior doesnt need it 
  //----------------------------------------------------------------------
  /**
   * determines, if the home position (the position that is taken as the base
   * for the positioning calculations) of the agent is the closest (among all
   * the other home positions) to a given position. The goalie is not
   * considered by this calculation.
   * @param pos the regarded position
   * @return true, if the home positions is closest, false otherwise
   */
  bool isMyHomePosClosestToPos(const Vector2D &pos) const;
  

  //----------------------------------------------------------------------
  //kick it as soon the behavior doesnt need it 
  //----------------------------------------------------------------------
  /**
   * This method calculates the stategic position of a certain teammate
   * @param unum the uniformnumber of the regarded teammate
   * @return strategic position of a teammate
   */
  /*REPLACE*/ Vector2D calcSBSPPosition(int unum);
    /*TEST*/
  float getDashPowerForPositioning();

  //----------------------------------------------------------------------
  //pass related routines, must be replaced by something because they are
  //bullshit
  //----------------------------------------------------------------------
 /* intersection method used in pass in open space
   *@param ballVelocity - the Velocity of the ball after pass in open space
   *@param start - current ball position
   *@param goal - Point where we want our player to intercept the ball
   *return true if pass line is free, false otherwise
   */
  bool isNotIntersected_PIOS(Vector2D &ballVelocity, 
			     const Vector2D &start, 
			     const Vector2D &goal, 
			     const Vector2D &ourPos);
  // stores return-values into those references
  bool getPassPoint(Vector2D& passPoint, Vector2D& wishVelo); 
 
  // stores return-values into those references
  bool canDribbleBehindDefense(Vector2D& dribblePoint, Vector2D& wishVelo); 
  
  /**
   * calculates the risk of a pass from one position to another position
   * based on the info which players will NOT intercept the pass
   * @param from position from which the pass starts
   * @param to position to which the pass goes
   * @param ignore vector of pointers to players who will not intercept
   * @return a (positive) value to represent how risky the pass is
   */
  float intersectionRisk(const Vector2D &from, const Vector2D &to, vector<Player*> &ignore);

  /**
   * tests whether the pass from player p1 to player p2 could be
   * intercepted by any player except p1, p2 (and those in ignore)
   * @param p1 player who should pass
   * @param p2 player who should receive the pass
   * @param ignore vector of pointers to players who will not intercept
   * @return true, if the pass could be intercepted, false otherwise
   */
  bool WorldModel::isIntersected(Player &p1, Player &p2, vector<Player*>* ignore = NULL);

  /**
   * tests whether the pass from the current ballposition to player p1
   * could be intercepted by any player except p1 (and those in ignore)
   * @param p1 player who should receive the pass
   * @param ignore vector of pointers to players who will not intercept
   * @return true, if the pass could be intercepted, false otherwise
   */
  bool WorldModel::isIntersected(Player &p1, vector<Player*>* ignore = NULL);

  /** Tests whether a a Point is intersected
      @param p1 the Point to which is looked
      @param p2 the Point from which is looked
      @param ballSpeed the Speed of the Ball
      @param reachable Area, normally the kickable Area (1), but for the Goalie other Values can
      be given as parameter
      @return true if intersected, false otherwise
  */
  bool  WorldModel::isIntersected_NH(const Vector2D &p1, const Vector2D &p2, float ballSpeed);
 double  WorldModel::isIntersected_NH_Risk(const Vector2D &p1, const Vector2D &p2, float ballSpeed);
  float WorldModel::isNotIntersected_NH(const Vector2D &p1, const Vector2D &p2, float ballSpeed);

  /**
   * tests whether the pass to the given position could have been
   * intersepted by the enemy
   * @param p a position on the field, the target for the virtual pass
   * @return true, if the pass could have been intercepted, false otherwise
   */
  bool isIntersected(const Vector2D &p);

  /**
   * tests whether the pass from a position to an other position could have
   * been intersepted by the enemy
   * @param p1 the position of the virtual pass-emitter
   * @param p2 the position of the virtual pass-receiver
   * @return true, if the pass could have been intercepted, false otherwise
   */
  bool isIntersected(const Vector2D &p1, const Vector2D &p2);

  /**
   * these calculate the tolerated min distance to a given pass line
   */
  float enemyIntersectionTolerance(float distance);
  float teammateIntersectionTolerance(float distance);

  /**
   * tests whether the the given teammate is marked by the opponents
   * @param player the teammate to test
   * @return true if not marked, false otherwise
   */
  bool isNotMarked(const Player &player);

  /**
   * tests whether the given player is well positioned with regard to the
   * passing
   * @param player the player to test
   * @return true, if he is, false otherwise
   */
  bool isWellPositioned(const Player &player);

  /**
   * tests whether the given player is the teammate with the biggest
   * opponent-free-area
   * @player the player to test
   * @return true, if he is, false otherwise
   */
  bool isUnvulnerable(const Player &player);
  
  /*
   * gibt einen gut postiertenPlayer zurueck, jenachdem ob
   * param==true --> offensiv
   * param==false --> sicher
   */
  const Player* wellPositionedPlayer();

  /**
   * returns the teammate with the biggest opponent-free-zone
   * @return the teammate
   */
  const Player* mostUnvulnerablePlayer();

  /*
   * Gibt den Player zurueck, der nach geltenden Kriterien anzuspielen ist
   * wenn NULL zurckgegeben wird, sollte der Ballbesitzende nicht passen
   */
  const Player* getPassCand(bool offensive, bool considerOffside = false, bool considerbackPass = true, PassType pt = PT_DEFENSIVE);

  //------------------------- tackle-Methods -------

  const double WorldModel::getTackleFailProb();
  const bool WorldModel::isBallTackable();

  // --------------------- /tackle-Methods -------------
  
  /******************************************************************************/
  /** This function positions the Goalie according to a found Pass Partner.
      It searches for a Pass Partner and gets the closest Goalie Kick Corner in
      relation to the Pass Partner. The Goalie will be moved to this point.
      @return a Vector2D which is the Point where the Goalie shall position itself.
      If no Pass Partner is found he positions himself to the left side (very random).
      @todo this is just a temporary design, it is still much too easy. But still better
      than before, because no passes in front line of the goal will be made.
  */
  Vector2D findGoalieKickOffPoint();
  
private:
  /**
   * This method tries to estimate the state of the worldmodel. There is a need
   * to do it because sometimes the agent has to do a decision in absense of any
   * visual information. The estimation is based on the state of the worldmodel
   * in the cycle before and the commands sent to the server in last cycle. At the
   * present, only the most important data is estimated: the position and the
   * velocity of the ball and the position, velocity, head direction and body
   * direction of the agent.
   */
  void estimateCurrentState();

// If all players are on their initial positions (before kick off, 
// after goal and such) put them their in this worldModel...
  void resetPlayerPositions();
  
  /**
   * Determines and sets the own absolute position on the field.
   * Uses thereto the seen flags and a kind of MonteCarlo-Localization.
   * @param flags the flags seen in the last See-Msg
   */
  void performSelfLoc(vector < Sense::Flag > &flags);

  /**
   * needed to determine the type of the message that invokes the update
   * of the player particles.
   */
  enum MsgTypeForParticlesUpdate {
    SenseBody,
    See
  };
  
  /**
   * Updates the particles by adding the odometrie-data. 
   * @param msgType because of what sensor information is this method called
   */
  void updatePlayerParticles( MsgTypeForParticlesUpdate msgType );

  /**
   * Deletes all particles which have an impossible relative 
   * position to the seen flags. Resamples the deleted particles afterwards.
   * @param flags the flags sensed in a see message
   * @param faceDirection the absolute angle the player is facing
   */
  void kickImpossiblePlayerParticlesAndResample( vector < Sense::Flag > &flags,
						 float faceDirection );

  /**
   * Adds Samples until ConfigFile::nrParticlesPlayer is reached.
   * @param flags the flags that are used to estimate the position
   * @param faceDirection the absolute angle the player is facing
   */
  void resamplePlayerParticles( vector < Sense::Flag > &flags,
				float faceDirection );

  /**
   * Calculates our position. 
   * @return the average of the position of the particles
   */
  Vector2D calcPosition();

  /* Methods needed by performSelfLoc. They caculate the noise-range of quantized noise.
   * Copy from Team UVA-Trilearn: WorldModelUpdate.C
   */  
  bool getMinMaxDistQuantizeValue( double dOutput, double *dMin,
				   double *dMax,   double x1, double x2 );
  bool getMinMaxDirChange( double dOutput, double *dMin,
			   double *dMax,   double x1 );
  bool getMinMaxDistChange( double dOutput, double dDist,
			    double *dMin, double *dMax, double x1, 
			    double xDist1, double xDist2 );
  double invQuantizeMin( double dOutput, double dQuantizeStep );
  double invQuantizeMax( double dOutput, double dQuantizeStep );
  double quantize( double input, double step ){
    return (rint(input/step))*step;
  };

  /**
   * assigns the new sensed players to stored players on the base of the
   * uniform number of the sensed players. The assigned players are then
   * removed from the player vector.
   * @param players the new sensed players
   */
  void assignPlayersByUnum(vector<Sense::Player> &players, bool ours);

  /**
   * trys to map the new sensed player to the stored player on the base of
   * their positions. 
   * @param teamPlayers players with known team affiliation
   * @param teamlessPlayers players with unknown team affiliation
   */
  void mapPartialKnownPlayers(vector<Sense::Player> &teamPlayers, bool ours);
  void mapUnknownPlayers(vector<Sense::Player> &teamlessPlayers);
  

  /**
   * assigns the data of a fresh sensed player to a stored player, assumes
   * perfect (not noised) data.
   * @param p the stored Player to assign the data to
   * @param sp the fresh sensed Player
   */
  void assignPerfectData(Player &p, const Sense::Player &sp);

  /**
   * assigns the data of a fresh sensed player to a stored player, assumes
   * noised (not perfect) data.
   * @param p the stored Player to assign the data to
   * @param sp the fresh sensed Player
   */
  void assignNoisedData(Player &p, const Sense::Player &sp);

  /**
   * this method searches the first player in team, which is not assigned,
   * that means, which is not seen yet or is not good confident. That place
   * can then be used to store the fresh sensed information
   * @param team the team to search an empty spot into
   * @return pointer to a player, which can be overwritten
   */
  Player* getFirstUnassignedPlayerInTeam(vector<Player> &team) const;

  /**
   * updates the stored ball with the information of the fresh sensed ball
   * @param balls a vector holding 1 or 0 balls
   */
  void updateBall(vector<Sense::Ball> &balls);

  /**
   * removes ghost from the worldmodel. Ghosts := Objects that should have
   * been seen in the last see message, but arent. Only the mobile object
   * (the players and the ball)
   */
  void removeGhosts();

  /** 
   * Stores still unmapped players in the vector unknownPlayers
   * @param guys the vector possibly containing unmapped players
   */
  void storeStillUnmapped(vector<Sense::Player> &guys);
  
  /**
   * the own messages are ignored at present
   * @param a the heared and parsed message from myself
   */
  void integrateSelfAudio (const Sense::SelfAudio &a);
  void integrateCoachFreeform (const Sense::CoachFreeform &freeform);
  
  /**
   * Integrates the messages from the other players into the worldmodel. The
   * messages from opponents are ignored. From teammates' messages the ball
   * position and ball velocity is obtained
   * @param a the heared and parsed message from some player
   */
  void integratePlayerAudio (const Sense::PlayerAudio &a);

  /**
   * The messages from the coach are ignored at the present
   * @param a the heared and parse message from the coach
   */
  void integrateCoachAudio (const Sense::CoachAudio &a);

  /**
   * Integrates the messages from the referee into the worldmodel.
   * The most important information here is the playmode
   * @param a the heared and parsed message from the referee
   */
  void integrateRefAudio (const Sense::RefAudio &a);

  
public:
  /**
   * This method is a helper routine for the estimation of the worldmodel
   * state, it estimates the velocity of the ball.
   * @param kickPow the kick power used in the kick command
   * @param kickDir the kick direction used in the kick command
   * @return the estimated ball velocity
   */
  Vector2D estimateBallVelocity(float kickPow = 0, float kickDir = 0)  const;

private:
  /**
   * With the limiting of the say message length, its now nessasary to be
   * able to store as much information as possible within 10 bytes. Only few
   * characters can be used in the say message. For coding this information,
   * we use a numbering system with the base "54". This function assigns a digit
   * in this numerical system to given character.
   * @param token the charakter to transform
   * @return a float, which is a digit in the "54"-System, that correlates to
   * the given character
   */
  float doAsciiToFloat(const char token);

  

  /** \name Helper Functions for performing calculations
      The following are private methods which are called by the
      high-level methods.
  */
  //@{
  void setUpPlayerSets(const PlayerSet ps,
		       const vector<Player> *&mates,
		       const vector<Player> *&opps,
		       const vector<Player> *&unknowns) const;

  void searchClosest(const Vector2D &point,
		     const vector<Player> *guys,
		     float conf, float *minDist,
		     const Player *&closest,
		     const Player *excludePlayer = NULL,
		     bool considerMayKickBall = false) const;
  
  void searchSecondClosest(const Vector2D &point,
			   const vector<Player> *guys,
			   float conf, float *minDist,
			   float *sndMinDist,
			   const Player *&closest,
			   const Player *&sndClosest,
			   const Player * excludePlayer = NULL) const;

  void searchClosestToGoalOutLine(const bool ownGoalOutLine,
				  const vector<Player> *guys,
				  float conf, float *minDist,
				  const Player *&closest) const;

public:
  /*TEST*/ void searchFastest(const vector<Player> *guys,
			      const Vector2D &ballPos,
			      const Player *&fastest, int &minCycles,
			      float conf,
			      bool considerMayKickBall = false) const;
private:
  void addUpInsidePlayers(const Rectangle &rect,
			  const vector<Player> *guys,
			  float conf, int *number) const;

  void addUpInsidePlayers(const Circle &rect,
			  const vector<Player> *guys,
			  float conf, int *number) const;

  void addUpInsidePlayers(const Cone &cone,
			  const vector<Player> *guys,
			  float conf, int *number) const;

  void addUpInsidePlayers(const Vector2D &start,
			  const Vector2D &end,
			  double width,
			  const vector<Player> *guys,
			  float conf, 
			  int *number) const;
  //@}


  /** \name Debugging-Routines */
  //@{
  void printUnknownPlayers();
  void printInit();
  void checkMovObject(const MovingObject&) const;
  void checkBall();
  void checkPlayer(const Player&) const;
  void checkTeammate(const Player&) const;
  void checkOpponent(const Player&) const;
  void checkMyself();
  void checkMyTeam();
  void checkTheirTeam();
  void checkUnknownPlayers();

  //@}

public:
  // Testing functions for new gaussian methods
  double getDeviationTo(const Vector2D &from, const Vector2D &to);
  // b: distance to goalie, a: angle from goalie to way of ball
  static double getLinearDiscriminant(double a, double b){
    return (a - 26.1) * 0.043 + (b - 9.0) * 0.09 - 0.2;
  };
  // calculate the gaussian ... well... calculate it 
  // d = derivation
  static double gaussian(double y, double d){
    return (1.0f/(d*sqrt(2*M_PI)) * exp(-((y*y)/(2.0f*d*d))));
  };
  // calculate if the goalie can be passed
  // pos: point to shoot to
  double getProbForPassingGoalie(const Vector2D &from, const Vector2D &to, float minSpeed = -1);
  // returns the probability for hitting the goal without concerning any other players
  double probToVector(const Vector2D &from, const Vector2D &to);
  // returns the probability for hitting the given point
  double getProbabilityForScoringGoal(const Vector2D &from, const Vector2D &to, float minSpeed = -1);
  // Go throught ConfigFile::scoringProbabilitySteps points and check which one 
  // has the highest scoring probability
  // return this point or an undefined vector
  Vector2D getBestPointToShootTo(const Vector2D &from, float minSpeed = -1);
  Vector2D getClosestPointToShootTo(const Vector2D &from, float minSpeed = -1);

  // http://www.wer-weiss-was.de/theme158/article1770084.html
  double simpsonFormel(double l, double r, int d, double dev, double (*f)(double x, double dev));

  double taylorFormel(const Vector2D &from, double t1, double t2, int d, const Vector2D &p1, const Vector2D &p2, double (*f)(double x, double dev));


  bool mayKickBall(const Player* p) const;
  bool mayKickBall();
  void updateFreeKickSituation();

public:
  Configuration configuration; /**< config parameters of the soccer server */
  Ball ball;                   /**< information about the ball */
  Myself myself;               /**< Informationen ueber mich selber */
  Team myTeam;                 /**< information about my team */
  Team theirTeam;              /**< information about the opp. team */
  Field field;                 /**< all information concerning the field */
  PotentialField potField;     /**< information about the potential field */
  float directionHistory[3];   /**< last three Angles seen by player */

  // Crazy stuff
  // Player who should receives should go there
  Vector2D passPointToGo;
  int passPointToGoCycles;
  // Passing players saves values and communicates those
  int passPointToSayCycles;
  int passPointUnumToSay;
  Vector2D passPointToSay;

private:
  //only for convenience
  vector<Player> *teammates;   //= &myTeam.getPlayers();
  vector<Player> *opponents;   //= &theirTeam.getPlayers();

  Playmode playmode;             /**< the current play mode */
  FieldSide fieldSide;           /**< the own field side */
  PenaltySide penaltySide;       /**< which goal to shoot at */
  int ourScore;                  /**< goals scored by my team von */
  int theirScore;                /**< goals scored by the opponents */
  Time time;                     /**< the current server time */
  vector<Player> unknownPlayers; /**< info about unrecognized players  */
  ServerCommandHistory lastSentCommands;  /**< info about the commands sent to
					     the server in the last cycle */
  ServerCommandHistory lastSentCommands_BackUp; 
  ServerCommandCounters queuedCommands;  /**< counters for all performed
					    commands */
  

  // particles for selfloc
  vector<Vector2D> particlesPlayer;           /**< particles used for the player-Selfloc */
  Time timeUpdatePlayerParticlesSee;          /**< time of last update because of see msg */
  Time timeUpdatePlayerParticlesSenseBody;    /**< time of last update because of sense msg */
  Vector2D velUpdatePlayerParticlesSenseBody; /**< velocity added because of sense data */

  /** \name synchronisation-related */
  //@{
  pthread_mutex_t mutex_newInfo;   /**< mutex to protect newInfo   */
  pthread_cond_t  cond_newInfo;    /**< cond variable for new info */
  bool            newInfo;         /**< indicates new info to integrate */
  bool            newCycle;        /**< indicates new soccer server cycle,
				      in which we have not acted yet */

    // Values communicated by the coach
  float coach_OffsideX;
  int coach_Formation;
  
  // Precalculated Values
  bool     iP_intable;
  Vector2D iP_intPoint;
  int      iP_noCycles;
  bool     iP_calculated;

  bool     sP_intable;
  Vector2D sP_intPoint;
  int      sP_noCycles;
  bool     sP_calculated;

  bool pre_isClosestTeammateToBall;
  bool pre_isClosestTeammateToBallReturn;

  bool            isFreeKickSituation;

  Time lastSeeSSTime;          /**< ss time of last receiving a see msg */
  int seeCount;                /**< number of sees in the current cycle so far */
  bool   profilingPerformed;   /**< true, if ss-profiling has been done */
  struct timeval  lastSeeTime; /**< time of last receiving a see msg */
  struct timeval  lastSbTime;  /**< time of last receiving a sense body msg */

  bool thinkSegment; /**< should the player think now (in synchMode) */
  //@}

  /** Threshold to decide if an object (player or ball) is close to
      the opponents goal outline (measured in meter) */
  static const float IS_CLOSE_TO_OPPONENT_GOAL_OUTLINE = 20.0;
  /** Threshold to decide if an object (player or ball) is close to
      the own goal outline (measured in meter) */
  static const float IS_CLOSE_TO_OWN_GOAL_OUTLINE = 20.0;
  /** Threshold to decide if a player is attacked by opponents */
  static const float IS_ATTACKED_SQUARE = 3.0;
  /** Threshold to decide if a player is attacked by opponents */
  static const float IS_ATTACKED_RADIUS = 5.0;
  /** Used in the implementation of (being|is)FrontalAttacked() */
  static const float FRONTAL_ATTACKED_HEIGHT = 10.0;
  /** Used in the implementation of (being|is)FrontalAttacked() */
  static const float FRONTAL_ATTACKED_WIDTH = 10.0;
  /** Threshold for ball possession */    
  static const double BALL_POSSESSION_MARGIN = 3.0;




};


//---------------- distance to something/somebody ------------------
inline float WorldModel::distanceToOwnGoalOutLine(const Player &p) const{
  return p.getPosition().getX() - field.OWN_GOAL_OUTLINE;
}

inline float WorldModel::distanceToOppGoalOutLine(const Player &p) const{
  return field.OPP_GOAL_OUTLINE - p.getPosition().getX();
}

inline float WorldModel::distanceFromPointToPoint(const Vector2D &v1,
						  const Vector2D &v2) const {
  return v1.distanceTo(v2);
}

inline float WorldModel::distanceFromPointToBall(const Vector2D &v1) const {
  return v1.distanceTo(ball.getPosition());
}

inline float WorldModel::distanceToPoint(const Vector2D &v) const {
  return distanceFromPointToPoint(myself.getPosition(), v);
}

inline float WorldModel::distanceToPointSq(const Vector2D &v) const {
  return myself.getPositionPt()->distanceToSq(v);
}

inline float WorldModel::distanceToBall() const {
  return distanceToPoint(ball.getPosition());
}

inline float WorldModel::distanceToPlayer(const Player &p) const {
  return distanceToPoint(p.getPosition());
}

inline float WorldModel::distanceFromPlayerToPoint(const Player &p,
						   const Vector2D &v) const {
  return distanceFromPointToPoint(p.getPosition(), v);
}

inline float WorldModel::distanceFromPlayerToBall(const Player &p) const {
  return distanceFromPointToPoint(p.getPosition(), ball.getPosition());
}
//-------------- /distance to something/somebody ------------------


//-------------- relative/global angles to point/ball ------------------
inline float WorldModel::angleFromPointToPoint(const Vector2D &from,
					const Vector2D &to) const{
  return (to - from).getDirection();
}

inline float WorldModel::angleToPoint(const Vector2D &pos) const{
  return angleFromPointToPoint(myself.getPosition(), pos);
}

inline float WorldModel::angleToPointRel2Body(const Vector2D &pos) const{
  return normalizeTo180Deg(angleToPoint(pos) - myself.getBodyDir());
}

inline float WorldModel::angleToPointRel2Head(const Vector2D &pos) const {
  return normalizeTo180Deg(angleToPoint(pos) - myself.getHeadDir());
}

inline float WorldModel::angleToBall() const{
  return angleToPoint(ball.getPosition());
}

inline float WorldModel::angleToBallRel2Body() const{
  return normalizeTo180Deg(angleToBall() - myself.getBodyDir());
}

inline float WorldModel::angleToBallRel2Head() const{
  return normalizeTo180Deg(angleToBall() - myself.getHeadDir());
}
//-------------- /relative/global angles to point/ball ------------------


//---------------- is close(st) to something/somebody --------------
inline bool WorldModel::isClosestToPlayer(const Player &from,
					  const Player &to,
					  float conf) const {
  return isClosestToPoint(from, to.getPosition(), conf, &to);
}

inline bool WorldModel::isClosestToBall(const Player& player,
					float conf) const {
  return isClosestToPoint(player, ball.getPosition(), conf);
}


inline bool WorldModel::isClosestTeammateToPlayer(const Player &teammate,
						  const Player &player,
						  float conf) const {
  return isClosestTeammateToPoint(teammate, player.getPosition(), conf, &player);
}

inline bool WorldModel::isClosestTeammateToBall(const Player &teammate,
						float conf,
						bool considerMayKickBall) {
  if (!pre_isClosestTeammateToBall){
    pre_isClosestTeammateToBall = true;
    pre_isClosestTeammateToBallReturn = isClosestTeammateToPoint(teammate, ball.getPosition(), conf, NULL, considerMayKickBall);
    return pre_isClosestTeammateToBallReturn;
  }
  return pre_isClosestTeammateToBallReturn;
}


inline bool WorldModel::isSecondClosestToPlayer(const Player &from,
						const Player &to,
						float conf) const {
  return isSecondClosestToPoint(from, to.getPosition(), conf, &to);
}

inline bool WorldModel::isSecondClosestToBall(const Player &player,
					      float conf) const {
  return isSecondClosestToPoint(player, ball.getPosition(), conf);
}


inline bool WorldModel::isSecondClosestTeammateToPlayer(const Player &teammate,
							const Player &player,
							float conf) const {
  return isSecondClosestTeammateToPoint(teammate, player.getPosition(), conf, &player);
}

inline bool WorldModel::isSecondClosestTeammateToBall(const Player &teammate,
						      float conf) const {
  return isSecondClosestTeammateToPoint(teammate, ball.getPosition(), conf);
}


inline bool WorldModel::isCloseToOwnGoalOutLine(const Player& player) const {
  return distanceToOwnGoalOutLine(player) < IS_CLOSE_TO_OWN_GOAL_OUTLINE;
}
inline bool WorldModel::isCloseToOpponentGoalOutLine(const Player& player) const {
  return distanceToOppGoalOutLine(player) < IS_CLOSE_TO_OPPONENT_GOAL_OUTLINE;
}
//---------------- /is close(st) to something/somebody --------------


//-----------------------  being in some area ------------------------
inline bool WorldModel::isOnField(const Vector2D &point) const {
  return field.FIELD.isInside(point);
}
inline bool WorldModel::isOnField(const Player &player) const {
  return isOnField(player.getPosition());
}

inline bool WorldModel::isInOwnPenaltyArea(const Vector2D &point) const {
  return field.OWN_PENALTY_AREA.isInside(point); 
}
inline bool WorldModel::isInOwnPenaltyArea(const Player &player) const {
  return isInOwnPenaltyArea(player.getPosition());
}
inline bool WorldModel::isInOppPenaltyArea(const Vector2D &point) const {
  return field.OPP_PENALTY_AREA.isInside(point);
}
inline bool WorldModel::isInOppPenaltyArea(const Player &player) const {
  return isInOppPenaltyArea(player.getPosition());
} 
//-----------------------  /being in some area ------------------------


//--------------------------- offside related -------------------------
inline bool WorldModel::isOffside(const Vector2D& point,
				  float conf) const {
  return (point.getX() > 0  &&  point.getX() > getOffsideX(conf));
}

inline bool WorldModel::isOffside(const Player& player,
				  float conf) const {
  return isOffside(player.getPosition(), conf);
}
//--------------------------- /offside related ------------------------


//------------------------  closest ... to ...  -----------------------
inline const Player* WorldModel::closestInSetToPlayer(const PlayerSet ps,
						      const Player &player,
						      float conf) const {
  return closestInSetToPoint(ps, player.getPosition(), conf, &player);
}

inline const Player* WorldModel::closestInSetToMe(const PlayerSet ps,
						  float conf) const {
  return closestInSetToPoint(ps, myself.getPosition(), conf, &myself);
}

inline const Player* WorldModel::closestInSetToBall(const PlayerSet ps,
						    float conf) const {
  return closestInSetToPoint(ps, ball.getPosition(), conf);
}
//------------------------  /closest ... to ...  ----------------------
  

//-------------------------  second closest ... to ...  -------------------
inline const Player* WorldModel::secondClosestInSetToPlayer(const PlayerSet ps,
							    const Player &player,
							    float conf) const {
  return secondClosestInSetToPoint(ps, player.getPosition(), conf, &player);
}

inline const Player* WorldModel::secondClosestInSetToMe(const PlayerSet ps,
							float conf) const {
  return secondClosestInSetToPoint(ps, myself.getPosition(), conf, &myself);
}

inline const Player* WorldModel::secondClosestInSetToBall(const PlayerSet ps,
							  float conf) const {
  return secondClosestInSetToPoint(ps, ball.getPosition(), conf);
}
//-------------------------  /second closest ... to ...  ------------------



// $Source: /home/robocup/virtualwerder/CVSRepository/VirtualWerder2004/worldmodel/WorldModel.hh,v $
#endif //WORLDMODEL_HH
