/******************************************************************************
 * WorldModel_LowLevel.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class WorldModel
 *
 * requires:
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: harlekin 
 * Group:
 * Co-authors: asm, mag, wollny, janw, cyco, dbernau, pagra, cniehaus
 *
 * Changelog:
 *
 * $Id: WorldModel_LowLevel.cc,v 1.59 2004/05/16 11:29:05 aw Exp $
 *
 *****************************************************************************/

#include <pthread.h>
#include <errno.h>
#include <algorithm>

#include "WorldModel.hh"
#include "Sensors.hh"
#include "debug.hh"
#include "common.hh"
#include "ConfigFile.hh"
#include "PotentialField.hh"

WorldModel::WorldModel()
  : configuration(),
    ball(),
    myself(),
    myTeam(),
    theirTeam(),
    field(),
    potField(),
    playmode(PM_BeforeKickOff),
    fieldSide(FS_Unknown),
    ourScore(0),
    theirScore(0),
    time(),
    unknownPlayers(),
    lastSentCommands(),
    particlesPlayer(),
    newInfo(false),
    newCycle(false),
    lastSeeSSTime(),
    seeCount(0),
    profilingPerformed(false){
  // initialize the synchronisation primitives
  pthread_mutex_init( &mutex_newInfo, NULL );
  pthread_cond_init ( &cond_newInfo, NULL );

  vector<Player>::iterator i;
  for(i = myTeam.getPlayers().begin(); i != myTeam.getPlayers().end(); ++i)
    i->setOpponent(false);
  for(i = theirTeam.getPlayers().begin(); i != theirTeam.getPlayers().end(); ++i)
    i->setOpponent(true);
  
  queuedCommands.clear();

  teammates = &(myTeam.getPlayers());
  opponents = &(theirTeam.getPlayers());
  // Should be clear to everyone
  (*teammates)[0].setGoalKeeper(true);
  // Just to be sure and avoid mapping errors
  (*opponents)[0].setGoalKeeper(true);
  (*opponents)[0].setPosition(Vector2D(50,0));
  (*opponents)[0].setUniformNumber(1);

  isFreeKickSituation = false;

  for(int j = 0; j < 3; j++)
    directionHistory[j] = UNDEFINED_FLOAT;

  passPointToGo.setVector2D( NAN, NAN );
  passPointToSayCycles = 0;
  passPointToGoCycles  = 0;

  ball.setPosition(Vector2D( 0, 0));
  ball.setVelocity(Vector2D( 0, 0));
  ball.setTimeSeen(Time(0,0));
  ball.setTimeOfPosition(Time(0,0));
  ball.setTimeOfVelocity(Time(0,0));

  coach_OffsideX = 0;
}

WorldModel::~WorldModel(){
}

bool WorldModel::waitForNewInfo() {
  bool returnValue = true;
  static int  timeoutSec = ConfigFile::serverDeadThr;

  if( newInfo == false ) { // there hasn't arrived any information yet
    static struct timeval now;
    static struct timespec timeout;
    gettimeofday(&now, NULL);
    timeout.tv_sec  = now.tv_sec + timeoutSec;
    timeout.tv_nsec = 0;

    pthread_mutex_lock( &mutex_newInfo );

    // this call atomically unlocks the associated mutex and waits for a
    // signal on the condition variable. The execution of the calling thread
    // is suspended until the cond variable is signaled. Before returning,
    // it locks the mutex again.
    int rc;

    //cond_timedwait can be interrupted by SIGALRM. In this case, it returns
    //with EINTR
    while( (rc = pthread_cond_timedwait( &cond_newInfo,
                                         &mutex_newInfo, &timeout) ) == EINTR )
      ;

    if( rc == ETIMEDOUT ){  //timeout event occured
      returnValue = false;
      DEBUG1_SYNCHRO("[REASONING] \twaitForNewInfo() waiting exited on timeout ");
    }

    pthread_mutex_unlock ( &mutex_newInfo );
  }

  // reset the indication of new information
  newInfo = false;

  return returnValue;
}

void WorldModel::setTimeLastSee(){
  DEBUG1_SYNCHRO("[SENSE] \tgot see");
  gettimeofday(&lastSeeTime, NULL);
  seeCount++;
  signalNewInfo();
}

void WorldModel::setTimeLastSB() {
  DEBUG1_SYNCHRO("[SENSE] \tgot sb");
  DEBUG1_SYNCHRO("[SENSE] \t   ------------  new server cycle   ------------" );
  gettimeofday(&lastSbTime, NULL);
  newCycle = true;
  seeCount = 0;
}


void WorldModel::signalNewInfo(){
  pthread_mutex_lock( &mutex_newInfo );
  newInfo = true;
  pthread_cond_signal ( &cond_newInfo );
  pthread_mutex_unlock( &mutex_newInfo );
}


bool WorldModel::haveOptimalSeeTimes() const {
  // if in synchmode we don't do profiling any more, because it doesn't work
  if ( Configuration::synchMode ) {
    // Yes, we have optimal see times. At least we pretend to have.
    DEBUG1_SYNCHRO("[REASONING] WorldModel::haveOptimalSeeTimes: SYNCH_MODE");
    return true;
  }
 
  if(profilingPerformed){   //prevents from executing the 2. time after returning false
    DEBUG1_SYNCHRO("[REASONING] profiling was already done..."); 
    return true;
  }
  
  float AT_0_TOO_LATE   = 0.15 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);
  float AT_50_TOO_EARLY = 0.50 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);
  float AT_50_TOO_LATE  = 0.65 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);
  float AT_0_TOO_EARLY  = 1.00 * (float)(Configuration::simulatorStep * Configuration::slowDownFactor);

  ViewQuality vq = myself.bodyInfo.getViewQuality();
  ViewWidth   vw = myself.bodyInfo.getViewWidth();

  if (vq != VQ_HIGH || vw != VW_NORMAL){
    DEBUG1_SYNCHRO("[REASONING] is already profiling");
    return false;
  }

  int timeDiff = timerDiffMsec(&lastSeeTime, &lastSbTime);

  if(/*(timeDiff < 0) 
       ||*/ 
     (timeDiff >  AT_0_TOO_LATE  && timeDiff < AT_50_TOO_EARLY) // see not at 0 ms
     ||                                                         // and
     (timeDiff >= AT_50_TOO_LATE && timeDiff <= AT_0_TOO_EARLY)  // see not at 50 ms
     ){
    DEBUG2_SYNCHRO("[REASONING] profiling needed. timeDiff: ", timeDiff); 
    return false;
  }

  DEBUG2_SYNCHRO("[REASONING] profiling not needed. timeDiff: ", timeDiff); 
  return true;
}


bool WorldModel::switchViewNeeded() const{
  int timeDiff = timerDiffMsec(&lastSeeTime, &lastSbTime);
  DEBUG4("[", time, "] timeDiff: ", timeDiff);
  // Got a See this cycle (and are using it now)
  if(timeDiff >= 0){
    ViewWidth   vw = myself.bodyInfo.getViewWidth();
    int nextSEEms = Configuration::sendStep;
    if(vw == VW_NARROW) nextSEEms /= 2;
    int diffNextSEE = timeDiff + nextSEEms;
    DEBUG4("got a See, viewWidth: ", vw, " next: ", diffNextSEE);
    if( diffNextSEE < 125 /*100*/ // next see will arrive this cycle
	|| 
	diffNextSEE > 190 /*200*/ // next see will arrive in t+2 cycles
	){
      //DEBUG2("diffNextSEE: ", diffNextSEE);
      // we need to change the view
      return true;
    }
  }
  // No see this cycle...
  return false;
}


// If all players are on their initial positions (before kick off, 
// after goal and such) put them their in this worldModel...
// 
// Problems: If a player is seen without his unum it can happen 
// that another player (with (wrong) unum) is mapped on that position
// Perhaps it's better to clear the player vector and fill it with
// correct unums and correct (intial) position. But we would loose
// all other informations: eg. playerType
/*TEST*/
void WorldModel::resetPlayerPositions(){
  if (playmode != PM_BeforeKickOff &&
      playmode != PM_OurKickOff &&
      playmode != PM_TheirKickOff
      ) return;

  // Resest information about ball
  ball.setPosition(Vector2D(0,0));
  ball.setVelocity(Vector2D(0,0));
  ball.setTimeSeen(time);
  ball.setTimeOfPosition(time);
  ball.setTimeOfVelocity(time);

  static vector<Player>::iterator i;
  int unum;

  // Reset the players of the own team to their normal position
  if (playmode == PM_BeforeKickOff){
    for(int j = 1; j < 12; j++){
      if (j == myself.getUniformNumber()){
	(*teammates)[myself.getUniformNumber()-1] = myself;
	// Shouldn't set the own position. Otherwise the agent
	// will stay at the side of the field cause they think they
	// already reached their intial positions.
	//myself.setPosition(Formations::getCurrentFormation().getInitPos(j));
	if (time.getStoppedCount() < 2) 
	  myself.setPosition(0, -35);
	myself.setVelocity(0, 0);
      }else{
	((*teammates)[j-1]).setUniformNumber(j);
	((*teammates)[j-1]).setPosition(Formations::getCurrentFormation()->getInitPos(j));	
	((*teammates)[j-1]).setVelocity(0,0); // cant move during this playmode	
	((*teammates)[j-1]).setTimeSeen(time); // now
      }
    }
  }else{
    for(i = myTeam.getPlayers().begin(); i != myTeam.getPlayers().end(); ++i){
      unum = i->getUniformNumber();
      // avoid problems with moveTo
      if (unum == myself.getUniformNumber()) continue;
      if(unum != UNDEFINED_INT){
	const Vector2D &newPos = Formations::getCurrentFormation()->getInitPos(unum);
	if ((playmode != PM_TheirKickOff || !(newPos.getX() == 0 && newPos.getY() == 0)) &&
	    (playmode != PM_BeforeKickOff || myself.getUnum() != unum)){
	  i->setPosition(newPos);
	  i->setTimeSeen(time);
	  //DEBUG4_HARLEKIN("Neue Position von Spieler Nr. ", unum, ": ", newPos);
	}
      }
    }
  }
  for(i = theirTeam.getPlayers().begin(); i != theirTeam.getPlayers().end(); ++i){
    // "delete" enemy players which are on our side
    if (i->getPositionPt()->getX() < 0){
      i->setTimeSeen(Time());
    }
    // "delete" enemy players who are too close to the middle circle
    if(i->getPositionPt()->distanceToSq(Vector2D(0,0)) <= 83.7225 && playmode == PM_OurKickOff){
      i->setTimeSeen(Time());
    }
  }
}

bool WorldModel::isNowThinkSegment() {
  bool result = thinkSegment;
  // Reset think-message
  thinkSegment = false;
  return result;
}

void WorldModel::setThinkSegment() {
  thinkSegment = true;
}

void WorldModel::checkForCollisions(){
  if (playmode == PM_BeforeKickOff) return;
  float ballSize = Configuration::ballSize;
  float playerSize = Configuration::getPlayerSize(myself.getType());
  const Vector2D &myPos = myself.getPosition();
  const Vector2D &ballPos = ball.getPosition();
  
  // Use square to save time
  double minDist = (ballSize + playerSize) * (ballSize + playerSize);
  minDist *= minDist;
  double dist = myPos.distanceToSq(ballPos);

  if (dist < minDist){
    myself.setVelocity(myself.getVelocity() * -0.1);
    ball.setVelocity(ball.getVelocity() * -0.1);
    DEBUG1_HARLEKIN("WorldModel::checkForCollisions: collided with Ball");
  }
}

void WorldModel::integrateCoachFreeform(const Sense::CoachFreeform &freeform){
  const int *types = freeform.getPlayerTypes();
  for (int m = 0; m < 11; m++){	  
    (*opponents)[m].setType(types[m]);
    DEBUG4_HARLEKIN("  Player ", (m+1), " has type ", types[m]);
    DEBUG4_AW("  Player ", (m+1), " has type ", types[m]);
  }  
  coach_OffsideX = freeform.getOffsideX();
  DEBUG2_HARLEKIN("  New OffsideX received: ", coach_OffsideX);
  DEBUG2_AW("  New OffsideX received: ", coach_OffsideX);
  coach_Formation = freeform.getFormation();
  if (coach_Formation == -1) coach_Formation = ConfigFile::standardFormation;
  Formations::setCurrentFormation(coach_Formation);
  DEBUG2_AW(" New FormationId received: ", coach_Formation);
}

/*TEST*/ bool WorldModel::integrateSensorInfo( Sensors *sensors){
  static bool fsUsed = ConfigFile::useFullstate;
  bool rc = false;

  if(ASSERT_DEBUG && getPlaymode() != PM_BeforeKickOff){
    checkBall();
    checkMyself();
  }

  //------------------ integrate SenseBody  -------------------------
  static Sense::SenseBody *sb;
  if(sensors->getSenseBody(&sb)){
    //DEBUG1_HARLEKIN("SenseBody");
    DEBUG1_SYNCHRO("[REASONING] \tintegrate sb");

    rc = true;
    float speedAmount, speedDir, headAngle;
    headAngle   = sb->getHeadAngle();
    speedAmount = sb->getSpeedAmount();
    speedDir    = sb->getSpeedDirection();

    static Vector2D velocity;
    velocity.setVector2D(speedAmount, normalizeTo180Deg(speedDir + headAngle), POLAR);
    myself.setVelRelToBody(velocity);

    myself.bodyInfo.
      setAll(sb->getViewQuality(), sb->getViewWidth(),
	     sb->getStamina(),     sb->getEffort(),
             sb->getHeadAngle(),
             sb->getKickCount(),  sb->getDashCount(), sb->getTurnCount(),
             sb->getSayCount(),   sb->getTurnNeckCount(),
             sb->getCatchCount(), sb->getMoveCount(), sb->getChangeViewCount());

    myself.bodyInfo.
      setTackleInfo(sb->tackleInfo.getExpires(), sb->tackleInfo.getCount());

    myself.bodyInfo.
      setFocusInfo(sb->focusInfo.isTargetSet(),   sb->focusInfo.getTargetSide(),
                   sb->focusInfo.getTargetUnum(), sb->focusInfo.getCount());

    myself.bodyInfo.
      setArmInfo(sb->armInfo.getMovable(), sb->armInfo.getExpires(),
                 sb->armInfo.getTarget(),  sb->armInfo.getCount());

    updateTime(sb->getTime());

    //estimate the most important data (own position, velocity, headangle and bodydirection,
    //ball position and ball velocity)
    estimateCurrentState();

    // update the player particles by our odometry-data (velocity)
    // this happens every cycle since each cycle a sensebody-msg occurs
    updatePlayerParticles( SenseBody );
  }

  //------------------integrate visual info  --------------------------
  static vector < Sense::Line >    lines;
  static vector < Sense::Flag >    flags;
  static vector < Sense::Goal >    goals;
  static vector < Sense::Player >  allies, enemies, neuts;
  static vector < Sense::Ball >    balls;

  //as a side effect of this call, the vectors are filled with data, if there
  //are some. otherwise they keep empty + return false
  if( !fsUsed //do not use see infos in the case fullstate is available
      &&
      sensors->getVisual(&lines, &flags, &goals, &allies, &enemies, &neuts, &balls)
    ){
    //DEBUG1_HARLEKIN("See");
    DEBUG1_SYNCHRO("[REASONING] \tintegrate see");

 
    rc = true;
    lastSeeSSTime = time;
    
    static vector<Sense::Flag>::iterator f;    
    for (f = flags.begin(); f != flags.end(); ++f){
      //DEBUG2_HARLEKIN("TST: Have seen flag: ", f->getId());
      //HACK!
      int id = f->getId();
      if (fieldSide == FS_Right){
	if      (id == FLAG_GLT) id = FLAG_GRB;
	else if (id == FLAG_GLB) id = FLAG_GRT;
	else if (id == FLAG_GRT) id = FLAG_GLB;
	else if (id == FLAG_GRB) id = FLAG_GLT;
	// Ignore other flags cause we dont use their confidence
      }
      field.setMarkerConfidence(id, &time);
    }

    // perform self localisation

    // calculate and set the face direction and the body direction for the agent
    float faceDirection = field.computeFaceDirection(lines, fieldSide);
    // Only update this if we get good values
    if (faceDirection != BIG_FLOAT){
      myself.setHeadDir(faceDirection);
      myself.setBodyDir(normalizeTo180Deg(faceDirection
					  - myself.bodyInfo.getHeadAngle()));
    }
    
    //update direction History -- needed for passInOpenSpace
    for(int i = 0; i <= 2; ++i){
      if(directionHistory[i] == UNDEFINED_FLOAT){
	directionHistory[i] = faceDirection;
	break;
      }
      if(i == 2){
	directionHistory[0] = directionHistory[1];
	directionHistory[1] = directionHistory[2];
	directionHistory[2] = faceDirection;
      }
    }

    // now the body direction is known and the velocity relative to the body
    // from the sense body message can be converted into global velocity
    static Vector2D myVel;
    myVel.setVector2D(myself.getVelRelToBody().rotated(myself.getBodyDir()));
    float speedMax = Configuration::getMaximumPlayerSpeed(myself.getType());
    speedMax *= speedMax;
    if (myVel.getLengthSq() > speedMax)
      myVel.setLength(Configuration::getMaximumPlayerSpeed(myself.getType()));
    myself.setVelocity(myVel);
    myself.setTimeOfVelocity(time);

    // calculate the global position of the agent
    performSelfLoc(flags);
    myself.setTimeOfPosition(time);
    myself.setTimeOfDirs(time);
    myself.setTimeSeen(time);


    //transmit the myself-data into the teammates vector (redundant, but much
    //easier for many calculations)
    (*teammates)[myself.getUniformNumber()-1] = myself;

    // First assign those, whose unum and team we know
    assignPlayersByUnum(allies,  true);
    assignPlayersByUnum(enemies, false);
    // Still remaining:
    //  - players without team
    //  - players without unum

    // Assign all allied players (without unum) to not confident players of our   team
    mapPartialKnownPlayers(allies,  true);
    mapPartialKnownPlayers(enemies, false);
    mapUnknownPlayers(neuts);
    // Assign all enemy  players (without unum) to not confident players of their team
    //mapUnknownPlayers(enemies, neuts, false);

    // Still remaining:
    //  - some unknown players who cant be mapped cause the information from some
    //    not seen players are too new to be replaced

    // unknownPlayers are from the last cycle
    unknownPlayers.clear();
    // put all of them into the unknownPlayers-Vector
    storeStillUnmapped(allies);
    storeStillUnmapped(enemies);
    storeStillUnmapped(neuts);
    
    //p.push_back(Player); assign(*(++i), *(++j)
    
    updateBall(balls);

    removeGhosts();
    
    //DEBUG1_WENNS("!!!!!!!!!!!!!!!!    checking Section    !!!!!!!!!!!!!!!");
    //DUMPVAR(myself);
    //DUMPVAR(myTeam);
    //DUMPVAR(theirTeam);
    //printUnknownPlayers();
    //DUMPVAR(ball);
  }


  //--------------------  integrate fullstate -----------------------------
  static Sense::Fullstate *fs;
  if(fsUsed && sensors->getFullstate(&fs, &goals, &allies, &enemies, &balls)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate fullstate msg");
    rc = true;

    //storing the time, viewmode, stamina, effort, headangle, counts, and arminfo
    //redundant to sb and therefore skipped

    //ignore recovery for now

    int playerCount;

    setOurScore(fs->getOurScore());
    setTheirScore(fs->getOppScore());
    setPlaymode(fs->getPlaymode());
    // setPenaltySide(fs->getPenaltySide());

    playerCount = allies.size();
    for(int i=0; i < playerCount; ++i)
      assignPerfectData((*teammates)[i], allies[i]);

    playerCount = enemies.size();
    for(int i=0; i<playerCount; ++i)
      assignPerfectData((*opponents)[i], enemies[i]);

    //store data in myself also
    assignPerfectData(myself, allies[myself.getUniformNumber()-1]);

    if(!balls.empty()){
      ball.setPosition(balls[0].getRelPosPt());
      ball.setVelocity(balls[0].getRelVelPt());
      ball.setTimeOfPosition(time);
      ball.setTimeOfVelocity(time);
      ball.setTimeSeen(time);
    }
  }

  //--------------------  integrate RefAudio -----------------------------
  static Sense::RefAudio *refAudio;
  if(sensors->getRefAudio(&refAudio)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate ref audio");
    integrateRefAudio(*refAudio);
  }

  //--------------------  integrate SelfAudio ----------------------------
  static Sense::SelfAudio *selfAudio;
  if(sensors->getSelfAudio(&selfAudio)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate self audio");
    //ignore own messages
    //integrateSelfAudio(*selfAudio);
  }

  //--------------------  integrate PlayerAudio --------------------------
  static Sense::PlayerAudio *playerAudio;
  if(sensors->getPlayerAudio(&playerAudio)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate player audio");
    integratePlayerAudio(*playerAudio);
  }

  //--------------------  integrate CoachAudio ---------------------------
  static Sense::CoachAudio *coachAudio;
  if(sensors->getCoachAudio(&coachAudio)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate coach audio");
    integrateCoachAudio(*coachAudio);
  }
  
  //--------------------- integrate some sort of CLANG -------------------
  static Sense::CoachFreeform *coachFreeform;
  if (sensors->getCoachFreeform(&coachFreeform)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate coach freeform");
    integrateCoachFreeform(*coachFreeform);
  }

  //--------------------  integrate Score --------------------------------
  static Sense::Score *score;
  if(sensors->getScore(&score)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate score");
    setOurScore(score->getOurScore());
    setTheirScore(score->getOppScore());
  }

  //--------------------  integrate Substitution -------------------------
  static Sense::Substitution *substitution;
  if(sensors->getSubstitution(&substitution)){
    integrateSubstitution(*substitution);
    DEBUG1_SYNCHRO("[REASONING] \tintegrate subst msg");
  }

  //--------------------  integrate Error --------------------------------
  static Sense::Error *error;
  if(sensors->getError(&error)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate error msg");
    DEBUG3("Error : <", error->getType(),">");
  }

  //--------------------  integrate server init reply  --------------------
  static Sense::InitMsg *initMsg;
  if(sensors->getInitMsg(&initMsg)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate init msg");

    myself.setUniformNumber(initMsg->getUnum());
    setFieldSide((FieldSide)initMsg->getSide());
    setPlaymode(initMsg->getPlaymode()); //Before_KickOff
    // setPenaltySide(fs->getPenaltySide());
  }

  //--------------------  integrate Server Parameter  --------------------
  static Sense::ServerParam *serverParam;
  if(sensors->getServerParam (&serverParam)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate serverParam");
    configuration.integrateServerParam(*serverParam);
  }

  //--------------------  integrate Player Parameter  --------------------
  static Sense::PlayerParam *playerParam;
  if(sensors->getPlayerParam(&playerParam)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate playerParam");
    configuration.integratePlayerParam(*playerParam);
  }

  //--------------------  integrate Player Type --------------------------
  static Sense::PlayerType *playerType;
  while(sensors->getPlayerType (&playerType)){
    DEBUG1_SYNCHRO("[REASONING] \tintegrate playerType");
    configuration.addPlayerType(*playerType);
  }

  if(ASSERT_DEBUG && getPlaymode() != PM_BeforeKickOff){
    checkBall();
    checkMyself();
    //checkMyTeam();
    //checkTheirTeam();
    //checkUnknownPlayers();
  }

  // only do this if we have new information
  // either a sb (new cycle and therefore new estimated values)
  // or a new see
  if (rc){
    resetPlayerPositions();
    updateFreeKickSituation();
    checkForCollisions();
    resetPrecalculatedValues();
  }


  lines.clear();
  flags.clear();
  goals.clear();
  allies.clear();
  enemies.clear();
  neuts.clear();
  balls.clear();

  return rc;
}


void WorldModel::assignPlayersByUnum(vector<Sense::Player> &players, bool ours){
  static vector<Sense::Player>::iterator i;
  static int  unum;
  static bool erased;

  static vector<Player> *team;
  if(ours)  team = teammates;
  else      team = opponents;

  for(i = players.begin(); i != players.end();  ){
    erased = false;
    unum = i->getUnum();
    if(unum != UNDEFINED_INT){
      assignNoisedData((*team)[unum-1], *i);
      // DEBUG4_WENNS("assigning by unum: ", (*team)[unum-1].getUniformNumber(),
      // 		   " pos: ", (*team)[unum-1].getPosition());
      players.erase(i);
      erased = true;
    }
    if(!erased) ++i;
  }
}

void WorldModel::mapPartialKnownPlayers(vector<Sense::Player> &teamPlayers, bool ours){
  static float distance;
  static const float distTolerance = ConfigFile::playerMapThreshold * ConfigFile::playerMapThreshold;
  static vector<Sense::Player>::iterator i;
  static vector<Player>::iterator j;
  Player *candidate = NULL;
  //bool erased;

  static vector<Player> *team;
  if(ours)  team = teammates;
  else      team = opponents;

  float dist = BIG_FLOAT, minDist = BIG_FLOAT;
  static Vector2D pos;
  const Vector2D &myPos = myself.getPosition();
  float myFaceDir = myself.getHeadDir();

  //map the players with known team affiliation
  //DEBUG1_WENNS("mapping team");
  for(i = teamPlayers.begin(); i != teamPlayers.end(); ++i){
    distance = i->getRelPosPt()->getLengthSq();
    
    candidate = NULL, minDist = BIG_FLOAT;
    // getRelPos() used because Sense::Player has no global Position
    pos = i->getRelPosPt()->relativeToGlobal(myPos, myFaceDir);

    //try to map only to the players that are not seen in this cycle
    //and that are still confident
    for(j = team->begin(); j != team->end(); ++j){
      // cant see yourself
      if ((Player*)&(*j) == &myself) continue;
      if(!j->isVisible(&time) && j->isGoodConfident(&time)){
	dist = j->getPositionPt()->distanceToSq(pos);
	if(dist < minDist)
          minDist = dist, candidate = (Player *)&(*j);
            //DEBUG6_WENNS("candidate: ", candidate->getUniformNumber(),
	    //    	 ", minDist = ", minDist, ", index: ", in);
      }
    }

    if( candidate != NULL && /*candidate->isKnownByUnum() && */
	minDist < distTolerance )
      //if a player found and the distance to that player is lesser than
      //a threshold, assign the new sensed info to that player
      assignNoisedData(*candidate, *i);
	//DEBUG4_WENNS("mapped to: ", candidate->getUniformNumber(),
	//	     ", minDist = ", minDist);
    else if((candidate = getFirstUnassignedPlayerInTeam(*team)) != NULL)
      //if no player found, search the first empty place in the players
      //vector and store the new sensed info in that place
      // This only overwrites old information.
      assignNoisedData(*candidate, *i);
      //DEBUG4_WENNS("assigned to: ", candidate->getUniformNumber(),
      //             ", minDist = ", minDist);
    else
      ;
    //DEBUG1_WENNS("team member not mapped");
  }
}

/*TEST*/ void WorldModel::mapUnknownPlayers(vector<Sense::Player> &teamlessPlayers){
  static const float distTolerance = ConfigFile::playerMapThreshold * 
                                     ConfigFile::playerMapThreshold;
  //float distTolerance;
  static float distance;
  static vector<Sense::Player>::iterator i;
  static vector<Player>::iterator j;
  Player *candidate = NULL;
  bool erased;

  float dist = BIG_FLOAT, minDist = BIG_FLOAT;
  static Vector2D pos;
  const Vector2D &myPos = myself.getPosition();
  float myFaceDir = myself.getHeadDir();

  //map the players with unknown team affiliation
  //DEBUG1_WENNS("mapping teamless");
  for(i = teamlessPlayers.begin(); i != teamlessPlayers.end();  ){
    erased = false; candidate = NULL; minDist = BIG_FLOAT;
    distance = i->getRelPosPt()->getLengthSq();
    pos = i->getRelPosPt()->relativeToGlobal(myPos, myFaceDir);

    //try to map only to the players that are not seen in this cycle
    //and that are still confident
    for(j = teammates->begin(); j != teammates->end(); ++j){
      // cant be seen by myself
      if((Player*)&(*j) == &myself) { continue; }
      if(!j->isVisible(&time) && j->isGoodConfident(&time)){
	dist = j->getPositionPt()->distanceToSq(pos);
	if(dist < minDist)
	  minDist = dist, candidate = (Player *)&(*j);
	    //DEBUG6_WENNS("candidate: ", candidate->getUniformNumber(),
	    //    	 ", minDist = ", minDist, ", index: ", in);
      }
    }

    for(j = opponents->begin(); j != opponents->end(); ++j){
      // cant be seen by myself
      if((Player*)&(*j) == &myself) { continue; }
      if(!j->isVisible(&time) && j->isGoodConfident(&time)){
	dist = j->getPositionPt()->distanceToSq(pos);
	if(dist < minDist)
	  minDist = dist, candidate = (Player *)&(*j);
	    //DEBUG6_WENNS("candidate: ", candidate->getUniformNumber(),
	    //    	 ", minDist = ", minDist, ", index: ", in);
      }
    }

    if( candidate != NULL /*&& candidate->isKnownByUnum()*/ &&
	minDist < distTolerance ){
      //if a player found and the distance to that player is lesser than
      //a threshold, assign the new sensed info to that player
      assignNoisedData(*candidate, *i);
      // DEBUG4_WENNS("mapped to: ", candidate->getUniformNumber(),
      // 		   ", minDist = ", minDist);
      //needed, becase the teamless player will be tried to map to both teams
      teamlessPlayers.erase(i);
      erased = true;
    }else if( i->isClose() && 
	     (candidate = getFirstUnassignedPlayerInTeam(*opponents)) != NULL ){
      //if there is an player with unknown team affiliation in the
      //"feeling area" assume it is an opponent getting the ball from me
      //or marking me
      assignNoisedData(*candidate, *i);
      // DEBUG4_WENNS("mapped close player to: ", candidate->getUniformNumber(),
      // 		   ", minDist = ", minDist);
      teamlessPlayers.erase(i);
      erased = true;
    }else
      ;
    // DEBUG1_WENNS("teamless not mapped");

    if(!erased) ++i;
  }

}


void WorldModel::assignPerfectData(Player &p, const Sense::Player &sp){
  p.setPosition(sp.getRelPos());    p.setTimeOfPosition(time);
  p.setVelocity(sp.getRelVel());    p.setTimeOfVelocity(time);

  p.setBodyDir(sp.getBodyDir());
  // p.setHeadDir(sp.getHeadDir()); 
  //headDir not is global 
  p.setHeadDir(normalizeTo180Deg(sp.getHeadDir() + sp.getBodyDir()));
  p.setTimeOfDirs(time);

  p.setGoalKeeper(sp.isGoalie());
  p.setClose(sp.isClose());
  p.setTackling(sp.isTackling());
  p.setPointDir(sp.getPointDir());
  p.setAdditionalData(sp.getType(), sp.getStamina(), sp.getEffort(), sp.getRecovery());

  p.setTimeSeen(time);
}


/*TEST*/ void WorldModel::assignNoisedData(Player &p, const Sense::Player &sp){
  float myFaceDir = myself.getHeadDir();

  if(!sp.getRelPos().isInvalid()){
    p.setPosition(sp.getRelPosPt()->relativeToGlobal(myself.getPosition(), myFaceDir));
    p.setTimeOfPosition(time);
  }
  
  if(!sp.getRelVel().isInvalid()){
    p.setVelocity(sp.getRelVelPt()->relativeToGlobal(myself.getVelocity(), myFaceDir));
    p.setTimeOfVelocity(time);
  }

  if(sp.getUnum() != UNDEFINED_INT)
    p.setUniformNumber(sp.getUnum());

  if(sp.getBodyDir() != UNDEFINED_FLOAT)
    p.setBodyDir(normalizeTo180Deg(sp.getBodyDir() + myFaceDir)), p.setTimeOfDirs(time);
  
  if(sp.getHeadDir() != UNDEFINED_FLOAT)
    p.setHeadDir(normalizeTo180Deg(sp.getHeadDir() + myFaceDir)), p.setTimeOfDirs(time);

  //TODO: because we dont have an undefined bool,
  //      the goalie attribute cannot be merged
  p.setGoalKeeper(p.isGoalKeeper() || sp.isGoalie() || sp.getUnum() == 1);  
  p.setClose(sp.isClose());        //close is always sure

  //TODO: the model for tackling is not complete now: the time of first assigning
  //      must be stored and later the flag must be unset if "takling unknown" and
  //      tackle_ban_cycle passed since the time
  p.setTackling(sp.isTackling());

  if(sp.getPointDir() != UNDEFINED_FLOAT)
    p.setPointDir(sp.getPointDir());

  // do not transfer the attributes type, stamina effort and recovery,
  // because by noised info (from a see) they are set to unknown, in any case

  p.setTimeSeen(time);
}

// HACK: is really getLast...
Player* WorldModel::getFirstUnassignedPlayerInTeam(vector<Player> &players) const {
  static vector<Player>::reverse_iterator i;
  for (i = players.rbegin(); i != players.rend(); ++i){
    if (!i->isGoodConfident(&time)){
      return &(*i);
    }
  }
  //i = find_if(players.begin(), players.end(), isNotConfident(&time));
  return NULL;
  //if(i == players.end()) return NULL;   //not found
  //else                   return (Player*)&(*i);  //isnt nonsense: * is no dereferencing
}


void WorldModel::updateBall(vector<Sense::Ball> &balls){
  //if the ball wasnt seen ... return and rely of the data estimation to the begin of
  //the cycle (after the sense body message)
  if(balls.empty()) 
    return;      
  
  static Vector2D position, velocity;
  float myFaceDir = myself.getHeadDir();
  
  position = ((balls[0].getRelPosPt()->relativeToGlobal(myself.getPosition(), myFaceDir)));
  
  //DEBUG4_HARLEKIN("Ball-Position, estimated: ", ball.getPosition(), " sensed: ", position);
  
  if(!balls[0].getRelVelPt()->isInvalid()){
    //the velocity is sensed
    velocity = ((balls[0].getRelVelPt()->relativeToGlobal(myself.getVelocity(), myFaceDir)));
    
    //DEBUG4_HARLEKIN("Ball-Velocity, estimated: ", ball.getVelocity(), " sensed: ", velocity);
    
    // takes the average from the sensed-velocity and the estimated velocity,
    // but only if the different between the both velocities is not too big, 
    // similar to the different is unrealistic. 
    if( fabs(velocity.getX() - ball.getVelocityPt()->getX()) <=
	2 * Configuration::ballRand * Configuration::ballSpeedMax && 
	fabs(velocity.getY() - ball.getVelocityPt()->getY()) <=
	2 * Configuration::ballRand * Configuration::ballSpeedMax ){
      velocity += ball.getVelocity();
      velocity /= 2;
    }
    //DEBUG2_HARLEKIN("Ball-Velocity, average: ", velocity);
    ball.setTimeOfVelocity(time);
  }
  
  //special case 1: if the ball is in the agents close area
  //-> the velocity of the ball is not sensed
  //and therefore the velocity must be calculated by the positions from 
  //the last and the current cycle
  else if( balls[0].isClose()){
    // too long ago, cant do anything...
    if( time - ball.getTimeOfPosition() > 2 ){
      // Should this be done? We still have the estimated values
      // which can be good enough
      velocity.setVector2D(UNDEFINED_FLOAT, UNDEFINED_FLOAT);
      //DEBUG1_HARLEKIN("Ball-Velocity: Information too old");
    }
    // saw it last cycle
    else if( time - ball.getTimeOfPosition() == 1 ){
      if( !ball.getVelocityPt()->isInvalid() ){
	// Shouldnt we consider the ball decay during this "back step" 
	velocity = ((position - (ball.getPosition() - ball.getVelocity())));
	
	DEBUG2_HARLEKIN("Ball-Velocity, seen last cycle w/ vel: ", velocity);
      }else{
	// Shouldnt we use an older position instead of the estimated one?
	// Otherwise we can get a velocity of zero. 
	// eg: estimated position and seen/sensed position are the same 
	velocity.setVector2D(&position);
	velocity -= ball.getPosition();       
      }
    }
    else if( time - ball.getTimeOfPosition() == 2 ){
      if( !ball.getVelocityPt()->isInvalid() ){
	velocity = ((position - (ball.getPosition() - ball.getVelocity())));
      }else{
	// Calculate velocity ac_cord_ing to the given positions (this cycle (p2), before 2 cycles (p1)) and 
	// the known ballDecay. p2 = p1 + vel + vel*decay. Therefore this formulae arises to calculate the
	// new velocity.
	velocity = ((((position - ball.getPosition()) / 
		      (1.0f + Configuration::ballDecay)) * 
		     Configuration::ballDecay));
      }
    }
    ball.setTimeOfVelocity(time); 
  }

  //special case 2: if the velocity amount is greater than max velocity
  //amount with maximum ball velocity noise: set the velocity amount to max,
  //since the difference is cause by the noise in the see data
  if( !velocity.isInvalid() && 
      velocity.getLength() > ((1.0 + Configuration::ballRand) * Configuration::ballSpeedMax) ){
    velocity.setLength(Configuration::ballSpeedMax);
    ball.setTimeOfVelocity(time); 
  }
  
  /*
  //special case 3: check for collision between the ball and the agent
  //if the positions of ball and player overlap at the end of the cycle
  //the position of the ball is set back a small distance and 
  //it's velocity is multiplied by -0.1!
  //
  // Why change the position of the ball? Nothing ac_cord_ing found in the manual
  if( (position - myself.getPosition()).getLength() < 
      (Configuration::getPlayerSize(myself.getType()) + Configuration::ballSize) ){
    DEBUG1_HARLEKIN("WorldModel::updateBalls(): collision with ball");
    //get the direction the ball was coming from
    float ang = (position - myself.getPosition()).getDirection();
    //correction of ball position and velocity
    position  = myself.getPosition() + Vector2D(0.2, ang, POLAR);
    velocity *= -0.1;
    ball.setTimeOfVelocity(time); 
    DEBUG2_HARLEKIN("Ball-Velocity, collision: ", velocity);
  }
  */

  //special case 4: the ball velocity is known to be (0,0) in non-play-on-modes
  if ((getPlaymode() != PM_PlayOn) &&
      (getPlaymode() != PM_TheirPenaltyTaken) &&
      (getPlaymode() != PM_OurPenaltyTaken)){
    velocity.setVector2D(0, 0);
    ball.setTimeOfVelocity(time);
    //DEBUG2_HARLEKIN("Ball-Velocity, non-playmode: ", velocity);
  }

  //special case 5: the ball position is known to be (0,0) before the kickoff
  if (getPlaymode() == PM_BeforeKickOff ||
      getPlaymode() == PM_OurKickOff    ||
      getPlaymode() == PM_TheirKickOff){
    position.setVector2D(0, 0);
  }

  //special case 6: if the ball is in the kickable area of an opponent
  //and not in my kickable area
  //-> set the vel to (0,0) because we don't know the actions of the opponent
  //and to avoid the fooling through smart dribblings
  if (getPlaymode() == PM_PlayOn            ||
      getPlaymode() == PM_TheirPenaltyTaken ||
      getPlaymode() == PM_OurPenaltyTaken   ){
    const Player* enemy = getAnEnemyWithBallKickable();
    if( enemy!= NULL & !isBallKickable()){
      if (enemy->getVelocityPt()->isInvalid()){
	velocity.setVector2D(Configuration::getPlayerSpeedMax(enemy->getType()), enemy->getBodyDir(), POLAR);
      }else{
	velocity.setVector2D(enemy->getVelocityPt());
      }
      //if (velocity.getLength() > Configuration::getPlayerSpeedMax())
      //velocity.setLength(Configuration::playerSpeedMax);
      ball.setTimeOfVelocity(time);
      //DEBUG2_HARLEKIN("Ball-Velocity, dribbling Enemy: ", velocity);
    }
  }

  /*** Removed for further testing
  // If we have a freekick-situation and
  // we think the ball is in the penalty area
  // (which is impossible) we should position
  // it on the next corner of the penalty area.
  // Problem: perhaps 'position' is invalid (999.333, 999.333)
  // so we use the last known position as indicator
  if (getPlaymode() == PM_TheirFreeKick || 
      getPlaymode() == PM_TheirIndirectFreeKick){
    // ball.getPosition() has to be valid or it wouldnt be
    // in the penalty area. No further checking needed
    if (isInOwnPenaltyArea(ball.getPosition())){
      if (ball.getPosition().getY() > 0)
	position = field.getMarkerPosition(FLAG_PRB);
      else
	position = field.getMarkerPosition(FLAG_PRT);
    }
  }
  */
  
  ball.setPosition(position);
  ball.setTimeOfPosition(time);
  ball.setVelocity(velocity);
  ball.setTimeSeen(time);
}


/*TEST*/ void WorldModel::removeGhosts(){
  // Don't remove ghosts if the ball is not far away. Otherwise we 
  // get some strange results. This is extremly dangerous
  // for the goalie. But other players could also
  // improve from this sloppiness or whatever this shit is called
  if (myself.getViewAngle() < 90 && 
      myself.getPosition().distanceToSq(ball.getPosition()) < 100 /*10*/)
    return;
  float halfViewAng = myself.getViewAngle() / 2.0;
  halfViewAng -= 0.05 * halfViewAng;  //for error
  
  //do the removing only for the ball for now
  if( fabs (angleToBallRel2Head()) < halfViewAng
      && ball.getTimeSeen() != lastSeeSSTime) {
    //DEBUG1_WENNS("ball ghost removed!");
    ball.setTimeSeen(Time()); //default constructor creates
                              //an undefined Time
    ball.setPosition(UNDEFINED_FLOAT, UNDEFINED_FLOAT);
  }

  static vector<Player>::iterator it;
  for(it = teammates->begin(); it != teammates->end(); ++it){
    if(it->getConfidence(&time) != 0.0) //player was ever seen
      if( fabs(angleToPointRel2Head(it->getPosition())) < halfViewAng
	  && it->getTimeSeen() != lastSeeSSTime ){
	it->setTimeSeen(Time());             //default constructor creates
                                             //an undefined Time
      }
  }

  for(it = opponents->begin(); it != opponents->end(); ++it){
    if(it->getConfidence(&time) != 0.0) //player was ever seen
      if( fabs(angleToPointRel2Head(it->getPosition()) ) < halfViewAng
	  && it->getTimeSeen() != lastSeeSSTime ){
	it->setTimeSeen(Time());             //default constructor creates
                                             //an undefined Time
      }
  }
}


/*TEST*/ void WorldModel::storeStillUnmapped(vector<Sense::Player> &guys){
  static vector<Sense::Player>::iterator i;

  for(i = guys.begin(); i!=guys.end(); ++i){
    unknownPlayers.push_back(Player()); 
    assignNoisedData(*(unknownPlayers.end()-1), *i);
  }
}


void WorldModel::integrateSelfAudio  (const Sense::SelfAudio  &a){
  //DEBUG3("Message from myself ignored: \"", a.getMsg(), "\"");
  if (&a == &a);
}

void WorldModel::integrateSubstitution(const Sense::Substitution &s){
  // I am substituted! AHHHHHHHHHHHHH!
  if (s.getUnum() == myself.getUniformNumber()){
    myself.setType(s.getType());
    //DEBUG2_HARLEKIN("WorldModel::integrateSubstitution: my new PlayerType: ", s.getType());
  }else{
    (*teammates)[s.getUnum()-1].setType(s.getType());
    //DEBUG4_HARLEKIN("WorldModel::integrateSubstitution: ", s.getUnum() ,"'s new PlayerType: ", s.getType());
  }
}

/*TEST*/ void WorldModel::integratePlayerAudio(const Sense::PlayerAudio &a){
  switch(a.getSide()){
  case 3:
  case 4:
  case 5: // Their and unknown messages
    return;
  case 0: // Left side messages
    if (fieldSide == FS_Left) break;
    else return;
  case 1: // Right side messages
    if (fieldSide == FS_Right) break;
    else return;
  case 2: // Our message
    break;
  default:
    return;
  }
   
  const char *message = (a.getMsg());
  int signInt = (int)doAsciiToFloat(message[0]);

  // This messages tells the agent where he should run to.
  // The talking player will try to pass to this position.
  // Because it is not guaranteed that the ball will ever
  // reach this position the agent will only run for 5 values
  // to this point. After this he has to rely on his own
  // interceptBall.
  if (signInt == 72){ // PassPoint ('?')
    float x2 = (doAsciiToFloat(message[3]) * (100/74)) / 100;
    float y2 = (doAsciiToFloat(message[5]) * (100/74)) / 100;
    
    // get signs
    signInt = (int)doAsciiToFloat(message[1]);
    int signInts[2];
    if (signInt >= 2){ signInt -= 2; signInts[1] = 1; } else { signInts[1] = -1; }
    if (signInt >= 1){ signInt -= 1; signInts[0] = 1; } else { signInts[0] = -1; }
    
    // get position
    static Vector2D pPos;
    pPos.setVector2D (signInts[0] * (doAsciiToFloat(message[2]) + x2),
		      signInts[1] * (doAsciiToFloat(message[4]) + y2));

    int unum = (int)doAsciiToFloat(message[6]);
    if (unum == myself.getUniformNumber()){
      //DEBUG2_HARLEKIN("Setting passPointToGoTo to: ", pPos);
      passPointToGo = pPos;
      passPointToGoCycles = 5; // CHECK VALUE!!!
    }else{
      //DEBUG4_HARLEKIN("Ignoring message: ", pPos, " cause it is for: ", unum);
      // If someone else is about to receive a pass we should forget our position
      // because there is only one ball in this game. ;)
      passPointToGo.setVector2D( NAN, NAN );
      passPointToGoCycles = 0;
    }
    return;
  }

  // Those messages were send by players who think they should
  // get the ball. Because there is no consideration of such
  // things in our behaviour only the position of that player
  // is updated. Perhaps we will see that his position is a
  // good one. But because we have another worldmodel than that
  // guy this is very unlikely.
  if (signInt == 73){ // GoodPointToScoreMessage ('_')
    float x2 = (doAsciiToFloat(message[3]) * (100/74)) / 100;
    float y2 = (doAsciiToFloat(message[5]) * (100/74)) / 100;
    
    signInt = (int)doAsciiToFloat(message[1]);
    int signInts[2];
    if (signInt >= 2){ signInt -= 2; signInts[1] = 1; } else { signInts[1] = -1; }
    if (signInt >= 1){ signInt -= 1; signInts[0] = 1; } else { signInts[0] = -1; }
    
    static Vector2D pPos;
    pPos.setVector2D(signInts[0] * (doAsciiToFloat(message[2]) + x2),
		     signInts[1] * (doAsciiToFloat(message[4]) + y2));

    int unum = (int)doAsciiToFloat(message[6]);
    (*teammates)[unum-1].setPosition(pPos);
    (*teammates)[unum-1].setTimeOfPosition(time-1);   
    //DEBUG4_HARLEKIN("[REPORT] Received message from ", unum, " with position: ", pPos);

    return;
  }

  // This message just states where the ball is and what it's velocity is
  int signInts[4];
  if (signInt >= 8){ signInt -= 8; signInts[3] = 1; } else { signInts[3] = -1; } 
  if (signInt >= 4){ signInt -= 4; signInts[2] = 1; } else { signInts[2] = -1; } 
  if (signInt >= 2){ signInt -= 2; signInts[1] = 1; } else { signInts[1] = -1; } 
  if (signInt >= 1){ signInt -= 1; signInts[0] = 1; } else { signInts[0] = -1; } 
  
  float x2 = (doAsciiToFloat(message[2]) * (100/74)) / 100;
  float y2 = (doAsciiToFloat(message[4]) * (100/74)) / 100;

  static Vector2D ballPos, ballVel;
  ballPos.setVector2D (signInts[0] * (doAsciiToFloat(message[1]) + x2),
		       signInts[1] * (doAsciiToFloat(message[3]) + y2));
  ballVel.setVector2D (signInts[2] * doAsciiToFloat(message[5]) / 27,
		       signInts[3] * doAsciiToFloat(message[6]) / 27);

  static Time timeOfSee = Time(0,0);
  if (time.getStoppedCount() > 0){
    timeOfSee.setTime(time.getTime(), time.getStoppedCount()-1);
  }else{
    timeOfSee.setTime(time.getTime()-1, time.getStoppedCount());
  }
  if (!ball.isHighConfident(&time)){
    ball.setTimeSeen(timeOfSee);
    ball.setTimeOfPosition(timeOfSee);
    ball.setPosition(ballPos+ballVel);
  }
  if (!ball.isVelHighConfident(&time)){
    ballVel *= Configuration::ballDecay;
    ball.setVelocity(ballVel);
    ball.setTimeOfVelocity(timeOfSee);
  }
}


void WorldModel::integrateCoachAudio (const Sense::CoachAudio  &a){
  DEBUG3("Message from coach ignored: \"", a.getMsg(), "\"");
}


void WorldModel::integrateRefAudio (const Sense::RefAudio  &a){
  Playmode pmode = a.getPlaymode();
  setPlaymode(a.getPlaymode());
  if(pmode == PM_OurGoal)
    ourScore = a.getGoals();
  if(pmode == PM_TheirGoal)
    theirScore = a.getGoals();

  setPenaltySide(a.getPenaltySide());  
  //unknownRefAudio ignored for now
}


void WorldModel::estimateCurrentState(){
  DEBUG1_WENNS("\n");
  if(ASSERT_DEBUG && getPlaymode() != PM_BeforeKickOff){
    checkBall();
    checkMyself();
  }

  //allocate variables for all possible command  parameters
  static float moveX, moveY;
  static float kickPow, kickDir;
  static float dashPow;
  static float turnAng;
  static float catchDir;
  static float turnNeckAng;
  kickPow = dashPow = turnAng = turnNeckAng = 0;
  moveX = moveY = kickDir = catchDir = UNDEFINED_FLOAT;

  //assign values to these according to the stored cmd types
  switch(lastSentCommands.getFst()){
    case CMD_MOVE: {
      moveX = lastSentCommands.getFstParam1();
      moveY = lastSentCommands.getFstParam2();
      //DEBUG5_HARLEKIN("Last command was: (move ", moveX, " ", moveY, ")");
      break;
    }
    case CMD_KICK: {
      kickPow = lastSentCommands.getFstParam1();
      kickDir = lastSentCommands.getFstParam2();
      //DEBUG5_HARLEKIN("Last command was: (kick ", kickPow, " ", kickDir, ")");
      break;
    }
    case CMD_DASH: { 
      dashPow = lastSentCommands.getFstParam1(); 
      //DEBUG3_HARLEKIN("Last command was: (dash ", dashPow, ")");
      break; 
    }
    case CMD_TURN: { 
      turnAng = lastSentCommands.getFstParam1(); 
      //DEBUG3_HARLEKIN("Last command was: (turn ", turnAng, ")");
      break; 
    }
    case CMD_GOALIE_CATCH: {
      catchDir = lastSentCommands.getFstParam1();
      //DEBUG3_HARLEKIN("Last command was: (catch ", catchDir, ")");
      break;
    }
    default : { 
      //DEBUG1_HARLEKIN("Last command was nothing special");

      ; 
    }
  }

  if(lastSentCommands.getSnd() == CMD_TURN_NECK){
    turnNeckAng = lastSentCommands.getSndParam1();
    //DEBUG3_HARLEKIN("Last command was: (turn_neck ", turnNeckAng, ")");
  }

  //get up-to-date data
  float headDir = myself.getHeadDir();
  float bodyDir = myself.getBodyDir();
  static Vector2D myPos;
  myPos = myself.getPosition();
  const Vector2D &myVelRelToBody = myself.getVelRelToBody();
  static Vector2D myVel;

  static Vector2D ballPos;
  ballPos.setVector2D(ball.getPosition());
  static Vector2D ballVel;
  ballVel.setVector2D(ball.getVelocityPt());

  // ball
  if(kickPow != 0){
    if(isBallKickable() && !ballVel.isInvalid()){  
      //beruht auf den daten aus dem letzten cycle      
      ballVel.setVector2D(estimateBallVelocity(kickPow, kickDir));
    }
  }

  //step 2: update the body dir if necessary
  if(turnAng != 0){
    //INERTIA_MOMENT is already taken into account in Effectors before sending
    //the command in previos cycle
    bodyDir += turnAng;
    //the angle should keep in [-180..180]
    bodyDir = normalizeTo180Deg(bodyDir);
  }
  //update the (abs) headDir by adding the rel headAngle sensed by sb
  headDir = normalizeTo180Deg(bodyDir + myself.bodyInfo.getHeadAngle());

  //calculate the absolute velocity by rotating the relative velocity by
  //the bodyDir
  myVel = myVelRelToBody.rotated(bodyDir);
  float maxSpeed = Configuration::getMaximumPlayerSpeed(myself.getType());
  if (myVel.getLength() > maxSpeed)
    myVel.setLength(maxSpeed);

  //step 3: move
  if(moveX != UNDEFINED_FLOAT && moveY != UNDEFINED_FLOAT)
    myPos.setVector2D(moveX, moveY); // does this affect monte carlo?
  else
  //increase the position by the velocity of the *previous* cycle
    myPos += (myVel / Configuration::getPlayerDecay(myself.getType()));     //me

  //increase the ball position by the ball velocity of the *previous* cycle
  //no ballDecay division needed because the velocity wasn't changed since
  //last cycle
  //use of the confidenceValue to set the ballValues on undefined
  if(!ballPos.isInvalid() && !ballVel.isInvalid()){
    ballPos += ballVel;
  }
  
  //step 4 : decay ball speed
  if(!ballVel.isInvalid()) 
    ballVel *= Configuration::ballDecay; //ball
  
  //step 5 : check for collision between the ball and the agent
  if( (ballPos - myPos).getLength() < 
      (Configuration::getPlayerSize(myself.getType()) + Configuration::ballSize) ){
    //get the direction the ball was coming from
    float ang = (ballPos - myPos).getDirection();
    //correction of ball position and velocity
    ballPos.setVector2D(0.2f, ang, POLAR);
    ballPos += myPos;
    ballVel *= -0.1f;
  }
  
  // write back the estimated data
  myself.setBodyDir(bodyDir);
  myself.setHeadDir(headDir);
  myself.setPosition(myPos);
  myself.setVelocity(myVel);
  
  ball.setPosition(ballPos);
  ball.setVelocity(ballVel);
  
  // transmit the myself-data into the teammates vector (redundant, but much
  // easier for many calculations)
  int myunum = myself.getUniformNumber();
  (*teammates)[myunum-1] = myself;

  // estimate the state of all other players
  static vector<Player>::iterator i;
  for(i = teammates->begin(); i != teammates->end(); ++i){
    if(i->getUniformNumber() != myunum){
      i->estimateState(&time);
    }
  }
  
  for(i = opponents->begin(); i != opponents->end(); ++i){
    i->estimateState(&time);
  }

  if(ASSERT_DEBUG && getPlaymode() != PM_BeforeKickOff){
    //DUMPVAR(myTeam);
    checkBall();
    checkMyself();
    //  checkMyTeam();
    //  checkTheirTeam();
    //  checkUnknownPlayers();
  }
}




Vector2D WorldModel::estimateBallVelocity(float kickPow, float kickDir) const {
  //DEBUG5_HARLEKIN("WorldModel::estimateBallVelocity(", kickPow, ", ", kickDir, ");");
  static Vector2D ballVel;
  ballVel.setVector2D(ball.getVelocity());
  static Vector2D ballRelPos;
  ballRelPos.setVector2D(ball.getPosition() - myself.getPosition());

  //determine the relative angle
  float ballRelAngle = normalizeTo180Deg(ballRelPos.getDirection() -
					 myself.getBodyDir());

  //calculate the effect of the kickPow
  kickPow *= (ball.calculateKickRate(ballRelPos.getLength(), ballRelAngle)
              * Configuration::kickPowerRate);
  //create the according acceleration vector
  static Vector2D accel;
  accel.setVector2D(kickPow, normalizeTo180Deg(kickDir + myself.getBodyDir()),
		    POLAR);
  //add them to the velocity (accelerate)
  ballVel += accel;
  //if "overdriven" take steps against
  if (ballVel.getLength() > Configuration::ballSpeedMax){
    ballVel.setLength(Configuration::ballSpeedMax);
  }
  return ballVel;
}


void WorldModel::performSelfLoc( vector < Sense::Flag > &flags ) {
  // At the start of the agent don't perform the selfloc. 
  // It wouldn't work because the flag infos are not sufficient.
  if ( myself.bodyInfo.getViewQuality() == VQ_LOW  ||
       time < Time(0,10) ) {    
    return;
  }
  // update the player-particles
  updatePlayerParticles( See );
  // kick particles that became impossible and resample them directly
  kickImpossiblePlayerParticlesAndResample( flags, myself.getHeadDir() );
  // add new samples until the wanted amount nrParticlesPlayer is reached
  // resamplePlayerParticles( flags, myself.getHeadDir() );
  // calculate the position of the agent, based on the average of the particles
  myself.setPosition(calcPosition());
}


/*
 * adds the agent-velocity from the last sense-body message to all particle-positions
 */
void WorldModel::updatePlayerParticles( MsgTypeForParticlesUpdate msgType ) {
  static Vector2D vel;
  vel.setVector2D( 0, 0 );

  // sense body message arrived
  if ( msgType == SenseBody ) {
    // in this cycle there was no update yet
    if ( time > timeUpdatePlayerParticlesSee ||
	 time > timeUpdatePlayerParticlesSenseBody ) {
      
      // the velocity was estimated good enough by estimateCurrentState().
      // estimateCurrentState() used the sent commands and the senseBodyMsg to
      // estimate the velocity. 
      vel.setVector2D(myself.getVelocityPt());
      vel = vel / Configuration::getPlayerDecay(myself.getType());
      
      // save, that a senseBodyMsg-Update happened and save 
      // the positionchange of the samples
      timeUpdatePlayerParticlesSenseBody = time;
      velUpdatePlayerParticlesSenseBody = vel;

      // update the particles
      static vector<Vector2D>::iterator i;
      for ( i = particlesPlayer.begin(); i != particlesPlayer.end(); ++i ) {
	i->setVector2D( i->getX() + vel.getX(),
			i->getY() + vel.getY() );
      }
    }
  } 

  // see-message arrived
  if ( msgType == See ) {

    // case 1: we already made a senseBodyMsg-Update this cycle
    if ( time <= timeUpdatePlayerParticlesSenseBody ) {
      // undo the old senseBodyMsg-Update and use the better data of the see message
      vel = (myself.getVelocity() / Configuration::getPlayerDecay(myself.getType())) -
	velUpdatePlayerParticlesSenseBody;

      // save, that a seeMsg-Update happened
      timeUpdatePlayerParticlesSee = time;

      // is a correction necessary?
      if ( vel.getLength() > 0.001 ) {
	// update the particles
	vector<Vector2D>::iterator i;
	for ( i = particlesPlayer.begin(); i != particlesPlayer.end(); ++i ) {
	  i->setVector2D( i->getX() + vel.getX(),
			  i->getY() + vel.getY() );
	}
      }
    } 
    else {
      // case 2: no update has happened in this cycle yet
      if ( time > timeUpdatePlayerParticlesSee ||
	   time > timeUpdatePlayerParticlesSenseBody ) {

	vel.setVector2D(myself.getVelocityPt());
	vel = vel / Configuration::getPlayerDecay(myself.getType());
	
	// save, that a seeMsg-Update happened
	timeUpdatePlayerParticlesSee = time;

	// update the particles
	static vector<Vector2D>::iterator i;
	for ( i = particlesPlayer.begin(); i != particlesPlayer.end(); ++i ) {
	  i->setVector2D( i->getX() + vel.getX(),
			  i->getY() + vel.getY() );
	}
      }
    }
  }
}


void WorldModel::kickImpossiblePlayerParticlesAndResample( vector < Sense::Flag > &flags,
							   float faceDirection ) {

  static Vector2D flagPos;
  static double minDist, maxDist;
  static double globalFlagAngle, flagDist;

   // use all seen flags
  int flagCount = flags.size();
  for ( int flag = 0; flag < flagCount; ++flag ) {

    flagPos.setVector2D(field.getMarkerPosition( flags[ flag ].getId() ));
    // take into account on which side we play
    if ( fieldSide == FS_Right ) {
      flagPos = -flagPos;
    }
    
    // angle of the vector from the player to the flag
    globalFlagAngle = normalizeTo180Deg( flags[ flag ].getRelPosPt()->getDirection() + faceDirection );

    // calculate the maximum and minimum value the real (not with noise polluted)
    // distance to the flag can be
    flagDist = flags[ flag ].getRelPosPt()->getLength();
    // the inner quantize step is 0.01 for flags (see formula in manual)
    getMinMaxDistQuantizeValue( flagDist, &minDist, &maxDist, 0.01, 0.1 );
    minDist *= minDist;
    maxDist *= maxDist;

    // check for every particle
    int actParticle = 0;
    int lastParticle = particlesPlayer.size() - 1;
    while ( actParticle <= lastParticle ) {
      // check if the actual distance of the flag is in the possible range and
      // if the angle doesn't differ more than 1.
      // angles are rounded and since noise is in global neck angle and relative
      // angle flag, maximum error is in range [-0.5,0.5] + [-0.5,0.5] = [-1,1].
      if (! (particlesPlayer[ actParticle ].distanceToSq( flagPos ) >= minDist &&
	     particlesPlayer[ actParticle ].distanceToSq( flagPos ) <= maxDist &&
	     fabs(normalizeTo180Deg(globalFlagAngle - (flagPos-particlesPlayer[ actParticle ]).getDirection())) <= 1) ) {
	// this particle has to be erased
	// overwrite the actual particle with the last particle in the vector
	particlesPlayer[ actParticle ] = particlesPlayer[ lastParticle ];
	// erase the last particle
	particlesPlayer.erase( particlesPlayer.begin() + lastParticle );
	lastParticle--;
      } else {
	// go to the next particle
	actParticle++;
      }
    }
    // TODO: moep?
    // fill the particles until the wanted amount is reached again
    resamplePlayerParticles( flags, faceDirection );
  }
}


void WorldModel::resamplePlayerParticles( vector < Sense::Flag > &flags,
					  float faceDirection ) {

  int nrParticles = particlesPlayer.size();

  // there are no particles at all left
  if (nrParticles == 0) {

    if ( flags.size() == 0 ) {
      // no flags seen -> can't initialize the particles
      return;
    }
    
    // find the closest flag
    int flag = UNDEFINED_INT;
    float flagDist = BIG_FLOAT;
    float distance;
    int flagCount = flags.size();

    for ( int j = 0; j < flagCount; ++j ) {
      if ( flags[j].isClose() ) { 
	//do not take into account "close" flags, i.e.
	//flags, that are in the "feeling area" and thus
	//not fully sensed
	continue;
      }
      distance = flags[j].getRelPosPt()->getLength();
      if ( distance < flagDist ) {
	flagDist = distance; 
	flag = j;
      }
    }

    // get the maximum and minimum distance of the flag (noise)
    double minDist, maxDist;
    getMinMaxDistQuantizeValue( flagDist, &minDist, &maxDist, 0.01, 0.1 );

    // get the angle of the vector from the flag to the player
    double globalFlagAngle = flags[ flag ].getRelPosPt()->getDirection() + faceDirection + 180.0;

    double actDist, actAngle;
    
    // create the particles
    for ( int i = 0; i < ConfigFile::nrParticlesPlayer; ++i ) {

      // choose a possible Value for the angle and the distance
      actDist = minDist + drand48()*(maxDist-minDist);
      actAngle = normalizeTo180Deg( globalFlagAngle - 1.0 + drand48()*2 );
 
      // create a particle with these values
      Vector2D actPos;
      if ( fieldSide == FS_Right ) {
	actPos.setVector2D( actDist, actAngle, POLAR );
	actPos -= field.getMarkerPosition( flags[flag].getId() );
      } else {
	actPos.setVector2D( actDist, actAngle, POLAR );
	actPos += field.getMarkerPosition( flags[flag].getId() );
      }
      // save the particle
      particlesPlayer.push_back( actPos );
    }


  } else {
    // some samples are left
    for ( int i = nrParticles; i < ConfigFile::nrParticlesPlayer; ++i ) {
      int idx = (int)( drand48() * (nrParticles-1) );
      // copy a randomly chosen particle and place the copy very near to it
      particlesPlayer.push_back( Vector2D( particlesPlayer[ idx ].getX() +
					     - 0.02 + drand48() * 0.04,
					   particlesPlayer[ idx ].getY() +
					     - 0.02 + drand48() * 0.04) );
      // other method that makes an exact copy
      // particlesPlayer.push_back( Vector2D( particlesPlayer[ idx ].getX(),
      //	 		              particlesPlayer[ idx ].getY() ) );
    }
  }
}


Vector2D WorldModel::calcPosition() {
  unsigned int size = particlesPlayer.size();
  if ( size != 0 ) {
    // calculate the average of the particles
    float sumX = 0;
    float sumY = 0;
    static vector<Vector2D>::iterator i;
    for ( i = particlesPlayer.begin(); i != particlesPlayer.end(); ++i ) {
      sumX += i->getX();
      sumY += i->getY();
    }
    sumX = sumX / size;
    sumY = sumY / size;
    return Vector2D( sumX, sumY );
  }
  // should never happen...
  // localisation did not succeed -> return old position
  DEBUG1("ERROR: WorldModel::calcPosition(): localisation was not successful!");
  return myself.getPosition();
}



// the following methods are taken from the source of
/*-- UVA Trilearn 2002 WorldModelUpdate.C --*/

/*  This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) distance of 'dOutput'.
    With other words, this is the range of values that will have the same
    quantized value 'dOutput'. The quantized steps are defined by x1 and x2.
    See the soccer server manual for details.
    \param dOutput resulting quantized value
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 value of inner quantize call (0.1 for player/ball, 0.01 for flags)

    \param x2 value of outer quantize call (normally 0.1)
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDistQuantizeValue( double dOutput, double *dMin,
					     double *dMax,   double x1, double x2 ) {
  // change output a little bit to circumvent boundaries
  // q = quantize(e^(quantize(ln(V),x1)),x2) with quantize(V,Q) = rint(V/Q)*Q
  // e^(quantize(ln(V),x1)_min = invQuantize( q, x2 )
  // quantize(ln(V),x1) = ln ( invQuantize( q, x2 ) )
  // ln(V) = invQuantize( ln ( invQuantize( q, x2 ) ), x1 )
  // V_min = e^( invQuantize( ln ( invQuantize( q, x2 ) ), x1 ) )
  // apply inverse quantize twice to get correct value
  dOutput -= 1.0e-10;
  *dMin = exp( invQuantizeMin( log( invQuantizeMin(dOutput,x2) ), x1 ) );
  dOutput += 2.0e-10;
  *dMax = exp( invQuantizeMax( log( invQuantizeMax(dOutput,x2) ), x1 ) );
  return true;
}

/*  This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) for the direction
    change of 'dOutput'.
    See the soccer server manual for details.
    \param dOutput resulting quantized value for direction change
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 value of outer quantize call (normally 0.1)
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDirChange( double dOutput, double *dMin,
                                     double *dMax,   double x1 ) {
  *dMin = invQuantizeMin( dOutput, x1 ) ;
  *dMax = invQuantizeMax( dOutput, x1 ) ;
  return true;
}

/*  This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) for the distance
    change of 'dOutput'.
    See the soccer server manual for details.
    \param dOutput resulting quantized value for direction change
    \param dDist distance to the perceived object
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 quantize step for distance change
    \param xDist1 value of inner quantize call to determine quantize value
                  corresponding to distance
    \param xDist2 value of outer quantize call to determine quantize value
                  corresponding to distance
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDistChange( double dOutput, double dDist,
				      double *dMin, double *dMax, double x1, 
				      double xDist1, double xDist2 ) {
  // Q_dist = quantize(e^(quantize(ln(V),xDist1)),xDist2)
  // q = Q_dist * Quantize( distance_change/distance, x1 )
  // dOutput = q/Q_dist = Quantize( distance_change/distance, x1 )
  // (distance_change/distance)_min = invQmin(q/Q_dist, x1 )
  // real distance is not know so should take into account distance range
  static double dMinDist, dMaxDist;
  getMinMaxDistQuantizeValue( dDist, &dMinDist, &dMaxDist, xDist1, xDist2 );
  dOutput = dOutput/dDist;
  double dMinCh = invQuantizeMin( dOutput, x1 ) ;
  double dMaxCh = invQuantizeMax( dOutput, x1 ) ;
  *dMin = min( dMinDist*dMinCh, dMaxDist*dMinCh );
  *dMax = max( dMinDist*dMaxCh, dMaxDist*dMaxCh );
  return true;
}

/*  This method returns the minimum value that generates dOutput as a quantized
    value when 'dQuantizeStep' is used as the quantized step.
    \param dOutput quantized output
    \param dQuantizeStep quantize step
    \return minimum value that when quantized produces 'dOutput' */
double WorldModel::invQuantizeMin( double dOutput, double dQuantizeStep ) {
  // q = quantize( V, Q ) = rint(V/Q)*Q -> q/Q = rint( V/Q)
  // min = rint(q/Q)-0.5 = V_min/Q -> V_min = (rint(q/Q)-0.5)*Q
  return (rint( dOutput / dQuantizeStep )-0.5 )*dQuantizeStep;
}

/*  This method returns the maximum value that generates dOutput as a quantized
    value when 'dQuantizeStep' is used as the quantized step.
    \param dOutput quantized output
    \param dQuantizeStep quantize step
    \return maximum value that when quantized produces 'dOutput' */
double WorldModel::invQuantizeMax( double dOutput, double dQuantizeStep ) {
  // q = quantize( V, Q ) = rint(V/Q)*Q -> q/Q = rint( V/Q)
  // max = rint(q/Q)+0.5 = V_max/Q -> V_max = (rint(q/Q)-0.5)*Q
  return (rint( dOutput/dQuantizeStep) + 0.5 )*dQuantizeStep;
}





  // int-val : char-val  : ascii-val
  //:::::::::::::::::::::::::::::::::
  //  0 -  9 : '0' - '9' : 48 -  57
  // 10 - 35 : 'A' - 'Z' : 65 -  90
  // 36 - 61 : 'a' - 'z' : 97 - 122
  // 62 - 62 : ' '       : 32
  // 63 - 66 : '(' - '+' : 40 -  43
  // 67 - 69 : '-' - '/' : 45 -  47
  // 70 - 70 : '<'       : 60
  // 71 - 72 : '>' - '?' : 62 -  63
  // 73      : '_'       : 95
float WorldModel::doAsciiToFloat(const char token){
  int my_msg = (int)token;
  if      (my_msg ==  32) my_msg  = 62;
  else if (my_msg <=  43) my_msg += 23;
  else if (my_msg <=  47) my_msg += 12;
  else if (my_msg <=  57) my_msg -= 48;
  else if (my_msg ==  60) my_msg  = 70;
  else if (my_msg <=  63) my_msg +=  9;
  else if (my_msg <=  90) my_msg -= 55;
  else if (my_msg ==  95) my_msg  = 73;
  else if (my_msg <= 122) my_msg -= 61;
  else my_msg = 0;
  return (float)my_msg;
}


//----   / freekick-methods ---

// checks wether the agent is allowed to kick the ball
bool WorldModel::mayKickBall(const Player* p) const{
  if (ball.getLastKickedBy() == NULL) return true;
  return (!(isFreeKickSituation && 
	    ((Player*)ball.getLastKickedBy())->getUniformNumber() == p->getUniformNumber()));
}
bool WorldModel::mayKickBall(){
  if (ball.getLastKickedBy() == NULL) return true;
  return (!(isFreeKickSituation && 
	    ((Player*)ball.getLastKickedBy())->getUniformNumber() == myself.getUniformNumber()));
}


void WorldModel::updateFreeKickSituation(){
  // Those are the modes after which we are not allowed to kick the ball if he executed the last kick
  if (playmode != PM_PlayOn        &&
      playmode != PM_OurFreeKick   && playmode != PM_OurKickIn &&
      playmode != PM_OurCornerKick && playmode != PM_OurIndirectFreeKick &&
      playmode != PM_OurKickOff /* && playmode != PM_DropBall */){ 
    isFreeKickSituation = false;
    ball.setLastKickedBy(NULL);
    return;      
  }
  
  const Player* closest = closestInSetToBall(PS_Players, ConfigFile::goodConfidencePlayer);
  
  // The freeKick was executed 
  if (playmode == PM_PlayOn){
    if (closest != ball.getLastKickedBy()){
      isFreeKickSituation = false;
      ball.setLastKickedBy(NULL);
      //DEBUG1_HARLEKIN("WorldModel::updateFreeKickSituation(): Leaving freeKickSituation");
      return;
    }
  }
    
  // Positioning for freeKick is in progress, save closest player to ball
  // this will be the one who wont be allowed to kick the ball afterwards
  isFreeKickSituation = true;
  //DEBUG1_HARLEKIN("WorldModel::updateFreeKickSituation(): kickedLastBy set");
  if (closest != NULL) ball.setLastKickedBy(closest);
  return;
}

//----

//--------------     bug-chasing    ------------------------
void WorldModel::printUnknownPlayers(){
  DEBUG1("UnknownPlayers:");
  vector<Player>::const_iterator pli;
  for(pli = unknownPlayers.begin(); pli!= unknownPlayers.end(); ++pli){
    DEBUG1(*pli);
  }
}

void WorldModel::printInit(){
  DUMPVAR(myself.getUniformNumber());
  DUMPVAR(getFieldSide());
  DUMPVAR(getPlaymode());
}

#define MAX_X_COORD 60
#define MIN_X_COORD -60
#define MAX_Y_COORD 40
#define MIN_Y_COORD -40
#define MAX_PLAYERVEL_WITHNOISE 0.55
#define MAX_BALLVEL_WITHNOISE   3
#define MAX_VEL_WITHNOISE MAX_BALLVEL_WITHNOISE
#define MAX_UNUM 11
#define MIN_UNUM 1
#define MAX_TEAMSIZE 11
#define MAX_PLAYERS_WITHOUTME 21

void WorldModel::checkMovObject(const MovingObject &mo) const {
  const Vector2D &pos = mo.getPosition();
  const Vector2D &vel = mo.getVelocity();
  assert(pos.isInvalid() ||
         ((pos.getX() < MAX_X_COORD && pos.getX() > MIN_X_COORD)
          && (pos.getY() < MAX_Y_COORD && pos.getY() > MIN_Y_COORD)));
  DUMPVAR_CERR_WENNS(vel);
  assert(vel.isInvalid() ||
         ((vel.getLength() <= MAX_VEL_WITHNOISE)
          && isValidAngle(vel.getDirection())));
}

void WorldModel::checkBall() {
  checkMovObject(ball);
}

void WorldModel::checkPlayer(const Player &p) const {
  checkMovObject(p);
  const Vector2D &vel = p.getVelocity();
  float hDir = p.getHeadDir();
  float bDir = p.getBodyDir();
  int unum = p.getUniformNumber();
  assert( vel.isInvalid() || vel.getLength()<=MAX_PLAYERVEL_WITHNOISE );
  if(!vel.isInvalid() && vel.getLength() > MAX_PLAYERVEL_WITHNOISE){
    DUMPVAR_CERR(vel.getLength());
    DUMPVAR_CERR(vel.getX());
    DUMPVAR_CERR(vel.getY());
  }

  assert( isUndefined(hDir) || isValidAngle(hDir) );
  assert( isUndefined(bDir) || isValidAngle(bDir) );
  assert( isUndefined(unum) || (unum <= MAX_UNUM && unum >= MIN_UNUM) );
}

void WorldModel::checkMyself(){
  checkTeammate(myself);
}

void WorldModel::checkTeammate(const Player &p) const {
  checkPlayer(p);
  assert(!p.isOpponent());
}

void WorldModel::checkOpponent(const Player &p) const {
  checkPlayer(p);
  assert(p.isOpponent());
}

void WorldModel::checkMyTeam(){
  int guysCount = teammates->size();
  assert(guysCount<=MAX_TEAMSIZE);
  for(int i = 0; i < guysCount; ++i)
    checkTeammate((*teammates)[i]);
}

void WorldModel::checkTheirTeam(){
  int guysCount = opponents->size();
  assert(guysCount<=MAX_TEAMSIZE);
  for(int i = 0; i < guysCount; ++i)
    checkOpponent((*opponents)[i]);
}

void WorldModel::checkUnknownPlayers(){
  int guysCount = unknownPlayers.size();
  assert(guysCount<=MAX_PLAYERS_WITHOUTME);
  for(int i = 0; i < guysCount; ++i)
    checkPlayer(unknownPlayers[i]);
}

//--------------     /bug-chasing    ------------------------


// $Source: /home/robocup/virtualwerder/CVSRepository/VirtualWerder2004/worldmodel/WorldModel_LowLevel.cc,v $
